#!/usr/bin/env node

import * as fs from 'fs-extra';
import * as child_process from 'child_process';
import * as glob from 'glob';
import * as path from 'path';
import * as yaml from 'js-yaml';
import * as program from 'commander';

let src: string;
let repoRelativePath: string;
let dest = 'docs-ref-autogen';
let packageWhiteList: string[] = [];

let pjson = require('../package.json');

program
.version(`v${pjson.version}`)
.description('A tool to go through source code and convert it to yaml data model for docfx.')
.option('--whiteList [value]', 'white list some package name for extraction.')
.arguments('<sourceFolder> [repoConfigPath]')
.action(function (sourceFolder: string, repoConfig: string) {
    src = sourceFolder;
    repoRelativePath = repoConfig;
 })
.parse(process.argv);

if (!src) {
    console.log('Error: The source file folder is not specified!');
    program.help();
}

if (program.whiteList) {
    try {
        let tempList = JSON.parse(program.whiteList);
        if (tempList && tempList.length) {
            packageWhiteList = tempList;
        }
    } catch (e) {
    }
}

// 1. prepare
fs.removeSync(dest);
fs.mkdirpSync(dest);

let tsconfigs = glob.sync(path.join(src, '**/tsconfig.json'));
let toc: any[] = [];

tsconfigs.forEach(function (tsconfig) {
    let packagePath = tsconfig.replace('tsconfig.json', 'package.json');
    generatePackageDoc(packagePath, dest);
});

toc = JSON.parse(JSON.stringify(toc));
fs.writeFileSync(`${dest}/toc.yml`, yaml.safeDump(toc));
process.exit(0);

function generatePackageDoc(packagePath: string, dest: string): void {
    let dir = path.dirname(packagePath);
    if (!fs.existsSync(packagePath)) {
        return;
    }

    let packageName = fs.readJsonSync(packagePath).name;
    if (packageWhiteList.length === 0 || packageWhiteList.indexOf(packageName) >= 0) {
        console.log(packageName);
        console.log(dir);
        let sourceCodeBasePath = `${dir}/src`;
        if (fs.existsSync(sourceCodeBasePath)) {
            let packageDest = `${dest}/${packageName}`;
            fs.mkdirpSync(packageDest);
            try {
                // child_process.execSync(`typedoc --json ${dir}/api.json ${dir}/src --module commonjs --ignoreCompilerErrors`);
                child_process.execFileSync('node', ['node_modules/typedoc/bin/typedoc', '--json', `${dir}/api.json`,
                    `${dir}/src`, '--module', 'commonjs', '--ignoreCompilerErrors']);
                let basePath = sourceCodeBasePath.replace(`${src}/`, '');
                child_process.execFileSync('node', ['node_modules/type2docfx/dist/main', `${dir}/api.json`, packageDest]);

                if (fs.existsSync(`${packageDest}/toc.yml`)) {
                    let subToc = yaml.safeLoad(fs.readFileSync(`${packageDest}/toc.yml`).toString());
                    toc.push(subToc[0]);
                    fs.removeSync(`${packageDest}/toc.yml`);
                }
             } catch (e) {
                console.log(`${packageName}: TypeDoc Parse Error.`);
             }
        } else {
            console.log(`No source file for ${packageName}`);
        }
    }
}
