# [sp-client](../../../README.md) › Best Practices › React Style Guide

SharePoint coding conventions for projects that use the [ReactJS](https://facebook.github.io/react/) UI toolkit.

## Component
- Use ES6/ES2015 style classes.
- The name of your component must be PascalCased
- Use `export default` for the component.
- Use an interface for the `props`.
- Use an interface for the `state`.
- State is set up in the `constructor`.
- Order lifecycle methods according to their order immediately after the constructor.
  - `componentWillMount`, `componentDidMount`, `componentWillRecieveProps`,
    `shouldComponentUpdate`, `componentWillUpdate`, `componentDidUpdate`,
    and `componentWillUnmount`.
- `render` method is last.
- Use "_handle" as a prefix for event handlers, e.g. `_handleClick` for `onClick`
  , `_handleDrag` for `onDrag`, etc.
  - Make the handler `private`.
- Use `bind` inside the constructor as far a possible.
  - If you have an event handler which uses `this`, you will need to bind the handler to
    the current context otherwise `this` will be different (window object) than what you expect.
  - Avoid binding in the render() because of performance implications. Create the bound
    function in the constructor. e.g.

```javascript
export default class MyComponent extends React.Component<MyComponentProps> {
  constructor(props: MyComponentProps) {
    super(props);

    // If I do not bind, than 'this' in the _handleClick method is the
    // invoker of the handler and not the component. Placing bind in the render()
    // method also has performance implications because render() gets called repeatedly.
    this._handleClick = this._handleClick.bind(this);
  }
```

## JSX
- Use `( ... )` when returning from the render method. The opening parenthesis is on the same line as `return` and the JSX
  starts on a newline indented one level.
- If properties cannot all fit on one line, align and sort the properties on their own respective
  lines after the tag name.

```javascript
// bad
public render(): JSX.Element {
  return ( <div>
    <div propB='value'
         propA='value'
    >
      <p>Hello World</p>
    </div>
    <div
      propA='value'
      propB='value'/>
  </div>);
}

// good
public render(): JSX.Element {
  return (
    <div>
      <div
        propA='value'
        propB='value'
      >
        <p>Hello World</p>
      </div>
      <div
        propA='value'
        propB='value'
      />
    </div>
  );
}
```

- Use single quotes.
- Do not use spaces between curly brackets.

## Example Component

```javascript
export interface MyComponentProps {
  // ...
}

// If your component does not have props or state you can omit the type.
// state will be of type any.
export default class MyComponent extends React.Component<MyComponentProps> {
  constructor(props: MyComponentProps) {
    super(props);

    // If I do not bind, than 'this' in the _handleClick method is the
    // invoker of the handler and not the component
    this._handleClick = this._handleClick.bind(this);
  }

  // Any lifecycle methods go here

  private foo(): void {
    // ...
  }

  private bar(): string {
    // ...
  }

  private _handleClick(): void {
    this.foo();
  }

  public render(): JSX.Element {
    return <div>
      <button onClick={this._handleClick}/>
      <AnotherComponent>
        A='value'
        B={anotherValue}
      </AnotherComponent>
    </div>;
  }
}
```