# [sp-client](../../../README.md) › Best Practices › Strict Null Checks

The "[strictNullChecks](https://www.typescriptlang.org/docs/handbook/release-notes/typescript-2-0.html)"
option is a new feature supported by the TypeScript compiler.  Across the board, wherever this feature
is enabled, it reveals important bugs lurking in the code.  We want to enable this everywhere, but
unfortunately a nontrivial amount of work is required to fix up a legacy project to compile with this
new option.  Thus, it is currently an "opt in" feature.

## Should I enable strictNullChecks for my project?

Yes, wherever possible.  Specific details:

- If your project has dependencies on libraries that are NOT strictly checked, then you
  may be unable to use this feature.  :-(

- If your project is a popular library, then you MUST enable strictNullChecks; otherwise your
  consumers will complain.

- For this reason, **all SPFx packages must enable strictNullChecks.**

- If your project has a large legacy code base, then it may be too expensive to fix up all
  the errors.

## How do I enable strictNullChecks?

Enable it at a project level by adding `"strictNullChecks": true` to the compiler options in
your `tsconfig.json` file.

## How to convert your code to work with strictNullChecks

### Principle 1: Scrubber functions should accept null and undefined
"Scrubber" functions are defined as functions that take a "dirty" input (from an untrusted source) and return a valid,
well-defined result. The expected typing for the "dirty" input should be `Expected Input Type | undefined | null` in this
order.

```typescript
class Guid  {
  // This function contains logic to handle null/undefined input. It will return undefined if the value isn't a valid
  // Guid string
  public static tryParse(value: string | undefined | null): Guid | undefined;
}
```

### Principle 2: Parameters for exported functions should have strict types, but also validate
All functions exported by the NPM package, regardless of being a "Scrubber" function or not, should contain logic to
appropriately handle null/undefined input. This will prevent consumers of your API from running into unexpected runtime
errors that could be handled within the function. The logic to handle if a value is null/undefined is also important
because the type signature of string is equivalent to string | undefined when strictNullChecks is disabled.

```typescript
class Text  {
  public static format(s: string, ...values: any[]): string {
    Validate.isNotNullOrUndefined(s);
    let formattedString: string;
    ...
    return formattedString;
  }
}
```

### Principle 3: Return types should be trustworthy
This allows consumers of the API to avoid writing code that checks if the return value of a function is undefined. In
strictNullChecks mode, attempting to use undefined as a parameter for a function that accepts only a string object is
a compiler error, which can be worked around by casting an object to type any.(The type any includes null and undefined
even in strictNullChecks mode) This is particularly useful for writing unit tests that test the behavior of an API with
unexpected input.

```typescript
class Guid  {
  // This function also contains logic to handle null/undefined input. However, it will always return a Guid object or
  // throw an error if the value isn't a Guid string.
  public static parse(value: string | undefined | null): Guid;
}
```

## How does strictNullChecks work?

From the [TypeScript release notes](https://www.typescriptlang.org/docs/handbook/release-notes/typescript-2-0.html):

> "TypeScript has two special types, Null and Undefined, that have the values `null` and `undefined` respectively. Previously
> it was not possible to explicitly name these types, but `null` and `undefined` may now be used as type names regardless of
> type checking mode.
>
> The type checker previously considered `null` and `undefined` assignable to anything. Effectively, `null` and `undefined` were
> valid values of every type and it wasn’t possible to specifically exclude them (and therefore not possible to detect
> erroneous use of them).
>
> `--strictNullChecks` switches to a new strict null checking mode.
>
> In strict null checking mode, the `null` and `undefined` values are *not* in the domain of every type and are only assignable
> to themselves and `any` (the one exception being that `undefined` is also assignable to void). So, whereas `T` and `T | undefined`
> are considered synonymous in regular type checking mode (because `undefined` is considered a subtype of any `T`), they are
> different types in strict type checking mode, and only `T | undefined` permits `undefined` values. The same is true for the
> relationship of `T` to `T | null`."
