# [sp-client](../../../README.md) › Best Practices › tslint Rules

These rules are automatically enforced by TSLint.
This document exposes what the rule does and how to properly solve the TSLint conflicts.

## File Structure

### File Names
- File names should be the same as the class of the default export of the file.

### Whitespace
- Do not use tabs.
- Use 2 spaces for indentation.
- Remove trailing whitespace on lines.

> Refer to the workspace set up documentation for instructions to set up different text editors
  to meet the whitespace guidelines.

### Line Length
- Lines must not be longer than 110 characters.

> This rule has been overriden for specific projects with different line lengths.

## Syntax

### Variables
- Do not use `var`.
- Use `const` or `let` for all your references.
> This ensures you do not unintentionally reassign references, which can lead to bugs.

```javascript
// bad
var a: number = 1;
var b: number = 2;

// good
const a: number = 1;
let b: number = 2;
```

- Use `const` when possible if you do not need to reassign the reference.
- Use `let` if you want to reassign references.
> `const` and `let` are both block-scoped rather than function-scoped.

```javascript
// bad
var count: number = 1;
if (true) {
  count += 1;
}

// good
let count: number = 1;
if (true) {
  count += 1;
}
```

- Do not combine declarations with one `const` or `let`.
- Group by `const` and then `let`, respectively.

```javascript
// bad
let stuff: string = 'foo';
const len: number = 2,
    isTrue: boolean = true;

// good
const len: number = 2;
const isTrue: boolean = true;
let stuff: string = 'foo';
```

### Strings
- Use single quotes for string literals.

```javascript
// bad
const str: string = "John said \"Hello\"";

// good
const str: string = 'John said "Hello"';
```

### Braces
- Always use braces for blocks.
- Opening braces do not go on their own line.
- `else/else if/catch` start on the same line as the closing brace of the
  previous statement.

```javascript
// bad
if (true) foo();

// bad
if (true)
  foo();

// bad
if (true)
{
  foo();
}
else
{
  bar();
}

// good
if (true) {
  foo();
} else {
  bar();
}

// good
try {
  foo();
} catch (e) {
  // ...
} finally {
  // ...
}

// bad
let x  = {y:0,z: 1}

// good
let x = { y: 0, z: 1 }
```

### Arrow Functions
- When you must use function expressions (as when passing an anonymous
  function), use arrow function notation.
> The created function executes in the current context of `this`, which is usually what
  we want, and is a more concise syntax.

```javascript
// bad
[1, 2, 3].map(function(x) {
  const y = x + 1;
  return x * y;
});

// good
[1, 2, 3].map((x: number): number => {
  const y = x + 1;
  return x * y;
});
```

### Functions/Methods
- Always define the return type.
> You can omit the return type for the constructor.

```javascript
// bad
public foo() {
  return true;
}

// good
public foo(): boolean {
  return true;
}
```

### Properties
- Use dot notation when accessing properties.

```javascript
const cat = {
  isNinja: true
}

// bad
const isNinjaCat = cat['isNinja'];

// good
const isNinjaCat = cat.isNinja;
```

- Use subscript notation `[]` when accessing properties with a variable.

```javascript
const cat = {
  isNinja: true
}

public testCat(prop: string): bool {
  return cat[prop];
}

const isNinja: bool = testCat('isNinja');
```

### Comparison & Equality
- Use `===` and `!==` instead of `==` and `!=`.
> The only exception is when checking for both null and undefined you can use
  `==`.

```javascript
// bad
if (x == y) {
  // ...
}

// bad
if (x != y) {
  // ...
}

// good
if (x === y) {
  // ...
}

// good
if (x !== y) {
  // ...
}
```

- Use shortcuts.

```javascript
// bad
if (str !== null && str !== undefined && str !== '') {
  // ...
}

// bad
if (collection.length > 0) {
  // ...
}

// good, empty string evaluates to false
if (str) {
  // ...
}

// good
if (collection.length) {
  // ...
}
```

### null vs undefined
- Do not use `null`.
- Use `undefined` when you do not want to assign a value.
> We want to minimize how often we want to check for `null` and `undefined`.

```javascript
let x; // x is undefined
let y = { };

// Missing properties and uninitialized references are undefined by default
console.log(x, y.z) // => undefined, undefined
```

- When you need to check for both use type coerced equality `==`

## References
Some of the guidelines have been adopted and modified from the following resources:
- [Airbnb JavaScript Style Guide](https://github.com/airbnb/javascript)
- [Khan Academy JavaScript Style Guide](https://github.com/Khan/style-guides/blob/master/style/javascript.md)
