# [sp-client](../../../README.md) › Best Practices › TypeScript Syntax Conventions

These rules affect the code, but are not enforced by TSLint, so it's up to the developers and reviewer
to enforce them.

## File Structure

### Copyright Header

- All files must have a copyright declaration header at the top of the file.
  Additionally, there must be an informative description of the file. The
  description should be targeted to someone unfamiliar with the code.

```javascript
/**
 * @copyright Microsoft Corporation. All rights reserved.
 *
 * @file Various components to handle management of lists of coaches for
 * the profile page. These utilities were not written to be a general purpose
 * utility for the entire code base, but has been optimized with the
 * assumption that the Profile namespace is fully loaded.
 */
```

### Imports
- Separate imports by _first_ and _third_ parties.
  - First-party is code we wrote __and__ is in the same repository.
  - Third-party is everything else.
- Use `import ... from '...'` when possible. Otherwise use `import ... = require('....')` syntax.
- Do not use `var ... = require('...')` syntax.

> Import ordering is still in debate. Ideally, the sorted order should be something an editor can generate automatically and not
  something which has to be constantly updated by hand.

```javascript
// bad
import { A, default as Z } from 'foobar';
import { C, B } from '../utility/utilities';
import React = require('react');
import FizzBuzz = require('./FizzBuzz');

// good
import * as React from 'react';
import Z, { A } from 'foobar';

import { B, C } from '../utility/utilities';
import FizzBuzz from './FizzBuzz';
```

### Exports
- Each file should export a single default class.
- A file can have auxiliary exports, but they should be closely related to the default export. E.g.
  interfaces, enums, etc.

```javascript
export interface IMyClass {
  // ...
}

export default class MyClass implements IMyClass {
  // ...
}
```

### Naming
- Use PascalCase for class and interface names.
- Use camelCase for variables, objects and method names.
- Use `_` to prefix private members.
- Use `I` to prefix interfaces

```javascript
interface IFoo {
  // ...
}

// PascalCase class name
class FooBar implements IFoo {
  private _bar: string;

  constructor() {
  this._bar = 'fizz buzz';
  }

  // camelCase method name
  public bazBar(): string {
  // ...
  }
}

```

### Naming
- Use PascalCase for class and interface names.
- Use camelCase for variables, objects and method names.
- Use `_` to prefix private members.
- Use `I` to prefix interfaces

```javascript
interface IFoo {
  // ...
}

// PascalCase class name
class FooBar implements IFoo {
  private _bar: string;

  constructor() {
  this._bar = 'fizz buzz';
  }

  // camelCase method name
  public bazBar(): string {
  // ...
  }
}

```

### Comments (_Tentative_)
- Use JSDoc style comments for classes and class members (`/**...*/`).
  - Do not annotate the type.
- Comments should provide value beyond what the code implies.
> Do not write comments trivially explainable by the context of the code.
- Use `@class` for class definitions.
- Use `@param` for function and method parameters.
- Use `@return` for non-void return values.
  - You do not have to annotate the `constructor`.
- Use `//` for inline comments. Inline comments should be preceded by an empty
  line and be above the code its annotating.
- Use `@remarks` for special remarks.
- Use `@internalremarks` for remarks that are meant for internal Microsoft developers only. These remarks are not
  expected to be available in external third-party documenation. e.g. Please make sure this interface definition matches
  with the server side class XYZManifest.cs.


```javascript
//bad
/**
 * Gets the width
 * @return The width
 */
public getWidth(): number {
  // ...
}

// good
/**
 * Calculates and returns the current width dependent on visibility.
 * @return The amount in px of width visible on the screen. If not visible, will return 0.
 */
public getWidth(): number {
  // Suppose a super complex calculation occurs here
}

// good (Note: no comments)
public getWidth(): number {
  // Suppose it just returns the width, and does not have any special considerations
}
```

- Use `@todo` for incomplete work items or known bugs. A description and bug ID
  must follow the tag. VSO can automatically find bugs prefixed with `#`.
  - Do not use this gratuitously. It is occasionally unavoidable, but we want
  to minimize technical debt as early as possible.

```javascript
@todo Need to add an example here. #
```

### End of file
- There must be a single empty newline at the end of each file.

## Syntax

### Types
- __Primitives__: When you access a primitive type you work directly on its
  value.
  - `string`
  - `number`
  - `boolean`
  - `null`
  - `undefined`

```javascript
const foo: number = 1;
let bar: number = foo;

bar = 9;

// foo is unchanged despite changing bar because bar's reference changed
console.log(foo, bar); // => 1, 9
```

- __Complex__: When you access a complex type you work on a reference to its
  value.
  - object
  - array
  - function

```javascript
const foo: number[] = [1, 2];
const bar: number[] = foo;
bar[0] = 9;

// foo has changed because foo and bar are both referencing the same object
console.log(foo[0], bar[0]); // => 9, 9
```

- Always specify types.
  - No space before the colon and a single space after the colon.
> There is no syntax to define explicit types when using destructuring. Add
  comments explaining the type inference.
- Use `T[]` instead of `Array<T>`.
> The former syntax is more succinct.
- Do not use `any`.
> Most cases an interface could replace the usage of `any` and is still type safe.
- Use `as` for type assertion. Do not use `<...>` for type assertion.
> The current type is checked at compile time if it matches the shape of the asserted type. `any`
  will always be allowed.

```javascript
// bad
const foo: Foo = <Foo>xyz;

// good
const foo: Foo = xyz as Foo;
```

- __Rarely__, you might need to force an assertion. You can use the double assertion syntax.

```javascript
const foo: Foo = xyz as any as Foo;
```

### Strings

- Use backticks for string interpolation.

```javascript
// bad
const john: string = 'John';
console.log('Hello' + john + '.'); // => Hello John.

// good
const john: string = 'John';
console.log(`Hello ${john}.`); // => Hello John.
```

- Never `eval` a string.


### Arrays
- Use literal syntax for array creation.

```javascript
// bad
const items: number[] = new Array();

// good
const items: number[] = [];
```

- Use `push` instead of direct assignment.

```javascript
// bad
items[items.length] = 1;

// good
items.push(1);
```

- Use spreads `...` to copy arrays.

```javascript
// bad
const len: number = items.length;
const itemsCopy: number[] = [];
for (let i = 0; i < len; ++i) {
  itemsCopy[i] = items[i];
}

// good
const itemsCopy: number[] = [...items];
```

- Use `for ... of` when iterating over an array.

```javascript
const myArray: number[] = [1, 2, 3];

// bad
const len: number = myArray.length;
for (let i: number = 0; i < len; ++i) {
  // ...
}

// good (Note: a type cannot be specified when using this syntax but it can always be inferred)
for (let item of myArray) {
  // ...
}
```

- Do not use sparse arrays.
- Do not use a trailing comma at the end of a sequence.

### Destructuring
- Use object destructuring when accessing and using multiple properties of an
  object.
> Destructuring saves you from creating temporary references for those
  properties.

```javascript
// bad
public getFullName(user: User): string {
  const firstName = user.firstName;
  const lastName = user.lastName;

  return `${firstName} ${lastName}`;
}

// good
public getFullName({ firstName: string, lastName: string }): string {
  return `${firstName} ${lastName}`;
}
```

- Use array destructuring.

```javascript
const arr: number[] = [1, 2, 3, 4];

// bad
const first: number = arr[0];
const second: number = arr[1];

// good
let [first, second]: number[] = arr; // first = 1, second = 2

// good, one line swap
[first, second] = [second, first]; // first = 2, second = 1

// good, you could also us the spread operator to get the rest of the array
let [first, second, ...remainder]: number[] = arr; // first = 1, second = 2, remainder = [3, 4]

// good, you can ignore indices
let [first, , third, ...remainder]: number[] = arr; // first = 1, third = 3, remainder = [4]
```

- Do not use array destructuring for multiple return values.
- Use object destructuring for multiple return values.
> Take advantage of property shorthand notation.

```javascript
// bad, need to consider ordering
const processInput = (input: string): number[] => {
  // left, right, top, bottom are initialized to some values
  return [left, right, top, bottom];
}

const [left, __, top] = processInput(someInput);

// good, independent of ordering and only get data you want
const processInput = (input: string): PositionValues => {
  // left, right, top, bottom are initialized to some values
  return { left, right, top, bottom };
}

const { left, right } = processInput(someInput);
```

### Objects
- Use the literal syntax for object creation

```javascript
// bad
const item = new Object();

// good
const item = {};
```

- Use property value shorthand
> Shorter to write and more descriptive than using the full syntax
- Group shorthand properties at the beginning of the object declaration,
  and alphabetize the rest of the properties.

```javascript
const sharePoint = 'SharePoint';

// bad
const obj = {
  microsoft: 'Microsoft',
  sharePoint: sharePoint,
  oneDrive: 'OneDrive'
};

// good
const obj = {
  sharePoint,
  microsoft: 'Microsoft',
  oneDrive: 'OneDrive'
};
```

### Functions/Methods
- Do not name a parameter `arguments`. This will take precedence over the
  `arguments` object given to every function scope.

```javascript
// bad
public nope(arguments): void {
  // ...
}

// good
public yup(args): void {
  // ...
}
```

- Put default parameters last.

```javascript
// bad
public handleThings(opts = {}, x): void {
  // ...
}

// good
public handleThings(x, opts = {}): void {
  // ...
}
```

## References
Some of the guidelines have been adopted and modified from the following resources:
- [Airbnb JavaScript Style Guide](https://github.com/airbnb/javascript)
- [Khan Academy JavaScript Style Guide](https://github.com/Khan/style-guides/blob/master/style/javascript.md)
