# [sp-client](../../../README.md) › How To › Add A Package Dependency

## How do I add a new library dependency to my project?

Since we're using the Rush tool to install our packages,

> **DO NOT DO THIS:**  `npm install --save example-lib`

Instead, follow the ["Updating a package.json file"](https://github.com/Microsoft/web-build-tools/wiki/Rush-~-Everyday-commands#updating-a-packagejson-file)
guidance from the Rush wiki.

If the package you're using is already approved for our repo, great!  You're done!

But if it is something new, then you need an "**External Package Review**."  You will
know because "rush install" will modify the **browser-approved-packages.json** file, and
when you commit that change, it will trigger a VSO policy requiring approval from
the "**External Package Approvers**" group.  Keep reading...  :-)


## What is an "External Package"?

An **External Package** is any package created/managed by a group of people who are external to the
**sp-client** repo.  This includes libraries used by the build tools (NodeJS environment)
as well as in the web browser (bundled by WebPack).  Generally these libraries are NPM packages,
but other scripts are also considered.  Libraries created by other teams at Microsoft (e.g. **odsp-utilities**)
do not require special legal approval for distribution, but are included in this definition as well,
since they have the same cost/quality/architecture considerations.

The SharePoint Framework has taken a strong stance in embracing code reuse and open source.
This is a great step ahead, but it brings serious responsibility.  Instead of asking:

  - "How do I code the best solution to this problem?"

...we now must look at other people's libraries and consider many questions to determine
the best choice.

## Evaluating External Packages

When evaluating an external packages into our ecosystem, we ask these questions:

1. Does the **sp-client** repo already have an approved package that solves this problem? If yes, use it.

2. For small utility functions, we have standardized on the [lodash](https://lodash.com/) library.
   If lodash can solve this problem, import **@microsoft/sp-lodash-subset**.  If the **lodash** function
   you need is not part of the subset, propose to add it.

3. We generally prefer open source to homebrew solutions.  Is there an open source
   [NPM package](https://www.npmjs.com) that solves the problem?  The https://npms.io
   and http://node-modules.com services both provide a ranked report.

4. Is there an internal library that already solves this problem?

5. If there are several options to choose from, we focus on these criteria:
   - **License type:** is it [easy to get
     LCA approval](https://microsoft.sharepoint.com/sites/lcaweb/Home/Product-Development/Using-Open-Source-Software/OSS-FAQ)?
   - **Size** (kilobytes, minified, non-gzipped):  For libraries that will run in the web
     browser, this directly impacts perf and is a primary concern
   - **Popularity:**  e.g. number of NPM installs, number of GitHub contributors, user comments
   - **Maturity:** e.g. how many GitHub commits are there, how many versions were released?
   - **Support:** e.g. when was the last GitHub commit?  Also check for risk of "[I got bored](https://github.com/WebReflection/es6-collections)
     and [am not bothering to find another maintainer](https://github.com/harthur/nomnom),
     so [good luck with that](http://www.theregister.co.uk/2016/03/23/npm_left_pad_chaos/)"
   - **TypeScript friendliness**: there should be @types already available, and the API should
     be usable without disabling [tslint](http://palantir.github.io/tslint/) rules
   - **Modernity:** e.g. no global window variables, no redefining of system types,
     doesn't inject ad hoc script tags into the DOM, etc
   - **Quality:** examine the source code:  Would you hire someone who codes like that?
     Is it a 5 line wrapper for the real library?  [Is there any actual code in there at all?](http://www.haneycodes.net/npm-left-pad-have-we-forgotten-how-to-program/))
   - **Performance**
   - **Security**

If the **sp-client** repo already contains an approved library, but you found something that
seems better, do not simply "use both."  You should start a proper deprecation process
to replace the old library with the new one.

Once you have decided to include an external package, please the steps outlined below.
This process has several benefits:

- Avoids legal trouble, which has happened in the past when developers incorporated
  OSS libraries without following Microsoft's [legal policies](https://microsoft.sharepoint.com/sites/lcaweb/Home/Product-Development/Using-Open-Source-Software)
- Avoids low quality libraries, which have caused trouble in the past
- Avoids redundant libraries, which is confusing for developers and degrades performance
- Raises awareness about the new library, so other devs can make use of it
- Helps avoid "npm install" sprawl, which is a major factor affecting our CI build times

## Getting Your Library Approved

1. Start by sending an e-mail to
   [spchan](mailto:SP-Client%20Help%20and%20News%20&lt;spchan@service.microsoft.com&gt;)
   with the following template:

   - **E-mail Subject:** *External Package Review for "json-loader"*
   - **Package Name:** *json-loader*
   - **Description:** *json-loader module for webpack*
   - **Project URL (npm/github):** *https://github.com/webpack/json-loader*
   - **Kind (Browser / Non-Browser):** *non-browser*
   - **Minified, Non-gzipped Size (if loaded in a browser):** *5kb*
   - **License:** *MIT*
   - **Alternative options:** *none found*
   - **Does the feature set of this library overlap with an already approved library?** *no*
   - **Why do you need it?**
       *We are using moment.js, which loads timezone files from JSON files, however WebPack
       does not natively understand how to load this data. "json-loader" is a plugin for
       WebPack that provides this functionality.*

    NOTE: The "Browser" category is tracked separately because web pages have additional
    performance considerations that don't apply to NodeJS developer tools.  Only specify
    "Non-Browser" if you are confident that the library will NOT be used in a web page.
    For example, **gulp** requires the NodeJS file system, so it is obviously "Non-Browser".
    The **lodash** library is used by both the browser and the build system, so we classify
    it as "Browser".  The **z-schema** library is technically loadable in a web browser,
    but we have no plans to do that, so it was classified as "Non-Browser".

2. Have a thoughtful discussion and reach a consensus

3. Create a PR that adds your library to the appropriate inventory file:
   - For libraries that will only be used by NodeJS scripts, and NOT loaded in the browser (e.g. *gulp*):
     [nonbrowser-approved-packages.json](../../../common/config/rush/nonbrowser-approved-packages.json)
   - For all other libraries:
     [browser-approved-packages.json](../../../common/config/rush/browser-approved-packages.json)

4. Make sure one of the people who approves your PR is in the "**External Package Approvers**" group.

5. If your library is used by projects that are released to third-party developers
   (e.g. the SPFx platform, tools, or Yeoman generator), contact Pat Miller to go through
   the legal approval process. This requires (1) registering our usage of the dependency
   with CELA, and (2) adding license text to our "ThirdPartyNotice.txt" files as instructed
   by the lawyers.
