# [sp-client](../../../README.md) › How To › Create A Localized String

This document describes the localization pipeline, some limitations of the current implementation,
and future work that will be done to improve the process.

## Contacts
* Questions about tooling:
  * Nick Pape ([nickpape](mailto:nickpape@microsoft.com))
  * Ian Clanton-Thuon ([ianc](mailto:ianc@microsoft.com))
  * Qinwei Zhu ([qinweiz](mailto:qinweiz@microsoft.com))
  * Pete Gonzalez del Solar ([pgonzal](mailto:pgonzal@microsoft.com))
* WEX Localization driver:
  * Anton Labunets ([antonla](mailto:antonla@microsoft.com))

## From a Developer's Perspective
### In TypeScript
A typical **SP-Client** developer is able to use localization simply by using the `@ms/sp-build-internal-web`
build rig. This rig defines several tasks which will make it easy to make use of the existing localization
pipeline. After taking a dependency on this build rig, the developer would then introduce a `resx` file
to describe their string. For example:

```xml
<?xml version="1.0" encoding="utf-8"?>
<root>
  <data name="RecycleBinPageTitle" xml:space="preserve">
    <value>Recycle bin</value>
    <comment>{NeedsReview} The title of recycle bin page.</comment>
  </data>
</root>
```

The `{NeedsReview}` should be used before the string has been voice reviewed, so that we don't pay
to translate unapproved non-final strings. It will also append a `***` to the value of the string
so that it is easy to find. For more information, look at [StringReview.md](./StringReview.md).

During the build, this RESX file will be converted to a TypeScript file that can be imported via:

```TypeScript
import strings from './RecycleBinPage.resx';
```

Nothing else is required. The string resources will be bundled, and by virtue of creating the RESX
file, the strings will also automatically go out to Office Resource Fabric
for translation. This process will take a couple of days.

### UA String Review

**Before enabling your flight in production, be sure to ensure that all strings have been reviewed!**
You don't want embarrassing asterisks appearing in your end-user's UI!

If you want to open a UI Review bug for the User Assistance (UA) team, here’s a task template: [SharePoint UI Request](https://microsoft.sharepoint.com/teams/ODSPCXUIReview/SitePages/Open%20a%20UI%20request.aspx).

And a link to the [MAX notebook](https://microsoft.sharepoint.com/teams/ODSPCXUIReview/_layouts/15/WopiFrame.aspx?sourcedoc=%7B6774ad12-bd2c-41e7-9c04-24aea7b6b2cf%7D&action=edit).

### In a Manifest
TODO

## Known issues

* Will be fixed by [411535](https://onedrive.visualstudio.com/SPPPlat/_workitems/edit/411535)
  "Every bundle from sp-publishing-webparts embeds a copy of every other webpart's strings"
    * Strings for the whole project are bundled together, even if the string resource is being imported in
      a webpack chunk. This may affect perf.
    * You cannot do a file-based import for strings across projects!
    * Moving a `resx` file will cause translations to possibly be lost, sometimes Office Resource Fabric detects
      this "resource shift", but it may not.

## Pipeline & Technical Details
### Export
There are a few steps during the export process, whereby strings are sent to Office Resource Fabric for
translation.

1. During the build, the strings `resx` file is converted to `json`.
1. All `json` string files are gathered into a single project-level `project.strings.json`. A hash
   (based on the relative RESX file path) is appended to the string's key.
1. All project's `project.strings.json` are collected into a single `repo.string.resx` file. The relative path to
   the project is appended to the string's key.
1. This `repo.strings.json` file is converted to a `repo.strings.resx` file.
1. The `repo.strings.resx` file is converted to `repo.strings.resx.lcg` using `lstools.exe`.
1. The `repo.strings.resx.lcg` is exported from the [master build definition](https://onedrive.visualstudio.com/SPPPlat/_build/index?buildId=602057&_a=summary&tab=artifacts)
   as a build artifact. [The actual filename is just `strings.resx.lcg`].

### Import
1. Translators working in Office Resource Fabric convert the strings to other languages.
1. The translated strings are checked into [//sp-client/common/localization/](https://onedrive.visualstudio.com/SPPPlat/_git/sp-client?path=%2Fcommon%2Flocalization&version=GBmaster&_a=contents)
   periodically. This folder contains a folder for every language with a `strings.resx.lcl` file.
1. There is a [build definition](https://onedrive.visualstudio.com/SPPPlat/_build/index?context=allDefinitions&path=%5Csp-client+Bots&definitionId=1092&_a=completed)
   which is run once daily which does several things:
   * Does a build to create the `repo.strings.resx.lcg` file.
   * This file is combined with each translated `lcl` file to produce a `{culture}.repo.strings.resx` file using `lstools.exe`.
   * This is converted to a `{culture}.repo.strings.json` file.
   * This file is split apart based on project name, and each segment is checked into `{project_path}/localization/{culture}.json`.

### During a production build
1. All strings are converted to a `project.strings.json` file, which is placed in `lib/resx-strings/en-us.json`.
1. All localized strings are copied from `localization/{culture}.json` to `lib/resx-strings/{culture}.json`.
1. Each project has an entry in the `localizedResources` section of `config/config.json` mapping the importable
   `resx-string` alias to `lib/resx-strings/{locale}.js`. Note that the `strings.resx.ts` file points at this alias.
1. The build tools collect all localized resources for a single culture into a single bundle (e.g. `{culture}.js`).
1. The build tools build all code in a single bundle (e.g. `code.js`).
1. A bundle is created combining the localized resources bundles with the code bundle (e.g. `code.{culture}.js`).
   This is shipped to production.