# [sp-client](../../../README.md) › How To › Create A New Project

## What is a project?

At a minimum, a "project" is a folder in a [Rush](http://aka.ms/rush) repo that contains these things:

- **package.json** - Since [NPM](https://docs.npmjs.com/getting-started/what-is-npm) is our formalism for
  modularization
- **gulpfile.js** - Since [Gulp](https://gulpjs.com/) is our formalism for building, and projects get built
- **.npmignore** - which describes whether your project gets published or not, and what files to publish

It may also contain these optional things:

- a local **README.md** file, hyperlinked from the root README.md, containing a useful additional information
- **tsconfig.json** with custom TypeScript compiler options
- **bin/** folder - if the project is a command-line tool
- **etc/** folder - containing other miscellaneous files that don't fit in a standard category
- **license/** folder - if the license agreement is many files (e.g. one for each language)
- **config/** folder - with configuration for our
  [gulp-core-build](https://github.com/Microsoft/web-build-tools/wiki/Gulp-Core-Build) based toolchain
- **localization/** folder - allowing your text strings to be processed by SharePoint's localization workflow
- **sharepoint/** folder - for SPFx Feature XML definitions, and *.spkg output
- **src/** folder - where your TypeScript source code goes
- **typings/** folder - only if you need custom \*.d.ts files because
  [DefinitelyTyped](https://github.com/DefinitelyTyped/DefinitelyTyped) does not have an "**@types**" package

It also may create these output folders during the build, which are not tracked by Git:

- **dist/** folder - bundled output, which may be deployed to a CDN
- **lib/** folder - compiler intermediary files, which may be published to NPM
- **node_modules/** folder, for installed NPM dependencies
- **temp/** folder - temporary files created during the build, which may be safely deleted

It should **NOT** contain any of these things:

- Imports using "../.." notation to reference files inside another project's folder -- **_projects must be
  fully isolated and self-contained!_**
- Other top-level folder names that aren't one of the standardized names described above
- Substantial code written in plain JavaScript -- we strongly prefer to use TypeScript for all projects
- Code or assets that are not copyrighted by Microsoft, including code snippets plagiarized from the internet
- A **README.md** file containing useless boilerplate

## When should I create a new project?

It's okay to put multiple things in the same project, even multiple web parts that get deployed as separate
bundles.

Advantages of multiple things together in one project:

- If they are related, they can be built into a common bundle that loads together
- Or they can go in separate bundles, but have their common code automatically extracted into a Webpack
  "[chunk](https://webpack.js.org/guides/code-splitting/#prevent-duplication)")
- Fewer projects means less clutter in our directory organization (although more clutter within a project)

Advantages of individual projects:

- Rush can skip building (and eventually TAB testing) of individual projects, but it will probably never
  skip building individual files/folders within a project
- Small projects are built in parallel, which can be faster
- Because we prohibit circular references between projects, keeping them small will force devs to avoid
  "spaghetti" designs (where every class depends on every other class within a project folder)

## How do I create my new project?

1. **Name your project.**  Since your name will be contributed in a global namespace with everyone else, put
  some careful thought into a concise name that clearly conveys your purpose.  If your project is a
  reusable library, it should have a scoped charter to avoid turning into a dumping ground.
  Examples of good charters: "accessing the Bing Map service", "tracking SharePoint telemetry".
  Examples of poorly scoped charters: "utilities related to the UI", "code shared between my team's projects".

2. **Choose a category folder.**  In this repo, we enforce that projects must always be under a top-level
  "[category folder](
  https://github.com/Microsoft/web-build-tools/wiki/Rush-~-Enabling-policies#projectfoldermindepth-controlling-folder-size
  )". The category folder names are chosen to keep their contents evenly balanced; if your project doesn't seem
  to fit any category, or if one category is collecting too many projects, feel free to propose changes.
  Reorganizing the project hierarchy is inexpensive, useful, and encouraged.<br/><br/>
  The current category folders are:

    - **apps** - applications based on SPFx's **sp-application-base**
    - **integration-tests** - In general, unit tests and TAB tests go inside the project that they
    are testing.  The "integration-tests" folder should *only* be used for exceptional cases, such as:
    end-to-end test projects that aren't associated with any specific package, tests that are unusually
    expensive to build/run, or tests that need their own NPM package due for technical reasons.
    - **libraries** - libraries that can be used in production, but NOT published for third-party developers
    - **prototypes** - experimental projects that do NOT ship to production, and are exempt from the usual
    code review, unit testing, and legal requirements
    - **spfx-core** - [SharePoint Framework](http://aka.ms/spfx) core web libraries, published for third-party
    developers
    - **spfx-libraries** - SharePoint Framework additional web libraries, published for third-party developers
    - **spfx-tools** - SharePoint Framework command-line tooling, published for third-party developers
    - **tools** - internal tooling projects
    - **webparts** - first-party web parts that get deployed to production


3. **Copy a skeleton project.**  Project templates can be found in the
  "[project-templates](../../project-templates)" folder.  Choose an appropriate one, and copy+paste it to your
  new project folder.  (Be careful not to copy the symlinked "node_modules" subfolder!)

4. **DO NOT ADD ANY CODE.**  We ask that your initial PR should just set up the new project; use a separate
  PR to start adding your code. Why?  The PR to create a new project always seems to have interesting code
  review discussions, and the audience will be tools people, not your feature crew.  Your PRs will get
  approved faster if you avoid mixing two different topics/audiences.

5. Add your project to the "[rush.json](../../rush.json)" file.  (Alphabetical order!)

6. **Customize these fields.**  Make sure to choose unique values for these parts of your template:
    - Package name, must be the same as your project folder name
    - Package description
    - The "id" for any *.manifest.json files (must be a v4 GUID in lowercase, e.g. using [this
      tool](https://www.guidgen.com/))
    - The "alias" in your *.manifest.json files

7. If your project has manifests, follow the additional guidance in
   HowTo/[CreateANewManifest.md](./CreateANewManifest.md)

8. **Run "rush install" and "rush build"** to test your new project.  Then create your PR.

9. Add an entry in the repository's [README.md](../../README.md) inventory, telling people what your project does.

10. **IMPORTANT: Configure a VSTS branch policy.**  Microsoft's compliance policy requires that any folder
  containing production code must reviewed and approved by a person with appropriate expertise.  Your feature
  crew should determine your quality bar (code quality level? test strategy? security checklist? documentation
  goals?), and then designate stakeholders who will be required to approve pull requests.  Use the [sp-client
  &gt; master](https://onedrive.visualstudio.com/DefaultCollection/SPPPlat/_admin/_policies?scope=58d9957f4e2344f6a2dbbe00ca15e4b3:refs/heads/master)
  VSTS control panel to create an appropriate branch policy.  (If you don't have permissions, find someone
  from the "**Project Administrators**" VSTS group on [this control panel](
  https://onedrive.visualstudio.com/DefaultCollection/SPPPlat/_admin/_security?_a=members) -- any of
  those people can help you create a branch policy.)  Lastly, create a test PR and confirm that the PR cannot
  be merged without signoff from one of your designated reviewers.


## Adding a new WebPart project

**FIRST** follow all the steps above. The next few steps are specific to setting up a WebPart Project.

1. Add a file to [common/review/webpack-chunk-sizes](https://onedrive.visualstudio.com/SPPPlat/_git/sp-client?path=%2Fcommon%2Freviews%2Fwebpack-chunk-sizes) named [your-folder-name.size.json]. For example, the file for the sp-embed-webparts folder is [sp-embed-webparts.size.json]. This will keep track of your changing project size.

2. Add your WebParts to [/tools/vsts-deploy/sources.json] (https://onedrive.visualstudio.com/SPPPlat/_git/sp-client?path=%2Ftools%2Fvsts-deploy%2Fsources.json). For this, you will need the alias and id you originally set in the WebPart's manifest file.

3. Every time you add a new WebPart to your project, you'll have to update the config.json file and add the new entrypoint and manifest. You will also have to repeat the step above.

4. To add your project to shrinkwrap.json, you can either run "rush generate" or manually add it as such:

```javascript
    "@rush-temp/empty-webpart-project": {
      "version": "0.0.0",
      "from": "projects\\empty-webpart-project",
      "resolved": "file:projects\\empty-webpart-project"
    },
```
5. The tools/webpart-index/src/webpart-configs folder should have .json files added for the new webpart:

```javascript
{
  "$schema": "./../schemas/first-party-webpart-config.schema.json",
  "id": "[PROJECT ID]",
  "data": null,
  "flightName": "EmptyWebParts"
}

```

6. Add a Strings.resx file to src/loc

7. Add a ManifestStrings.resx file describing each of your WebParts in this format:


    <!-- Empty Webpart -->
      <data name="EmptyWebPartTitle" xml:space="preserve">
        <value>Empty WebPart</value>
        <comment>{NeedsReview}[Tag:ODSP-SPPLAT-WebpartIndex]The title of the Empty Webpart.</comment>
      </data>
      <data name="EmptyWebPartDescription" xml:space="preserve">
        <value>Display messages from Empty Webparts.</value>
        <comment>{NeedsReview}[Tag:ODSP-SPPLAT-WebpartIndex]The description of the Empty Webpart.</comment>
      </data>

8. Update the README.md in the root folder with a description of your project.


## Other considerations for a new project

### Test API entry point

If your library exposes test APIs to be consumed by other libraries (e.g. mocks), use the
[test-api.js](../../spfx-core/sp-core-library/test-api.js) pattern.  Consumers can then import these
definitions like this example from
[DigestCache.test.ts](../../spfx-core/sp-http/src/spHttpClient/test/DigestCache.test.ts):

```javascript
import { MockTimeProvider } from '@microsoft/sp-core-library/test-api';
```