# [sp-client](../../../README.md) › Reference › API Extractor

Read this if you made a change that updated a file under **common/reviews/api/**.

## What is "Public API"?

Generally speaking, the SPPPlat team considers client-side "Public API" to be:

> *Any classes, interfaces, or other definitions that are exported by an
> NPM package, if we published the package for use by third-party developers.*

This effectively includes:

- All packages under the "spfx-core" folder

- All definitions directly exported by the **package.json** entry point
  (typically **src/index.ts**)

- Any indirectly reachable definitions.  For example, if **MyClass** is exported
  and **MyClass.getOptions()**  returns **IMyOptions**, then **IMyOptions** will
  have IntelliSense and is effectively usable by developers even if we forgot
  to explicitly export it.

- All class members with the TypeScript **public** keyword (not to be confused
  with the **@public** JSDoc tag)

- All **protected** members of exported classes unless there is some mechanism
  to prevent subclassing

- All **base classes** or **base interfaces** of an exported class/interface
  (but obviously not their private members)

It does NOT include:

- Classes such as **_SPKillSwitch** that have been designated as "internal-only"
  because of a compelling reason why third-party developers won't want it

- Entire NPM packages that have been designated as "internal-only" and will not
  be published to npmjs.com

*NOTE: For server APIs such as REST and CSOM, we have a separate process
that is described here:* http://aka.ms/spapirev


## Why do we care about Public API?

The vast majority of the JavaScript community works in a cycle like this:

1. Upgrade your framework and libraries
2. Fix some bugs, implement features
3. Compile everything
4. Test everything
5. Deploy to production
6. Wait for users to complain or suggest features, then GOTO step #1

If the framework and libraries follow [SemVer](http://semver.org/),
then a minor version bump in step #1 shouldn't cause any compiler errors in step #3,
but it could introduce bugs that are found in step #4.  For a major version bump,
the developer may have to do significant work to get their project to compile again,
but it's their choice when to upgrade.  Yup, sounds cool...

**[LOUD RECORD SCRATCH!]**  This is not how web parts work!  Our cycle is more
like this:

1. A third-party developer builds a web part
2. They compile it into a bundle and manifest
3. They test it
4. They deploy it to SharePoint, and then we may never hear from them again
5. Microsoft swaps out our Framework with an updated release
6. Microsoft swaps out the React library with a new version
7. Microsoft ships a new version of the **sp-pages** application
8. GOTO step #5 once per month, for twenty months
9. Janice in accounting tries to add the web part to her page, possibly in a completely
   different application that didn't even exist in step #3...  AND IF THE WEB PART CRASHES,
   SHE WILL BE VERY UPSET.  Even if Janice knows how to contact the web part developer,
   they might be a consultant whose support contract has expired.  Uh oh!

For steps #5 through #9, note that there is no test pass, and we don't even get
to recompile the source code.  The versioning has to be mostly minor bumps
(to avoid expensive side-by-side loading or shims), and all libraries must perfectly
honor the SemVer semantics.

Try to explain these requirements to a typical JavaScript developer, and their
head might explode when they realize the implications.  "Impossible -- it can't be done!"
But if you think about it, this is exactly how SharePoint web parts have *always* worked
in C#.  SPPPlat is making the bold claim that we can recreate that same scenario
using TypeScript in a web browser.  (Obviously this is a complex topic, and a reason for you
to immediately join our team if you're not afraid of a challenge!)  But that's a story
for another day...

The main takeaway is that formalizing our Public API surface is crucial to ensuring
that Janice's web part never breaks, so she will perceive SharePoint to be a reliable
product.  There are many other preconditions for this success, but the Public API is
the biggest one and the one that requires the most brainpower from you, the well-meaning
hero who wants to add a new API to the framework.


## What is the bar for Public API?

As with the [SPApiRev](http://aka.ms/spapirev) process for C#, our goal isn't to burn
up huge amounts of time trying to make every class perfect so we'll never have to
change it again.  But we do believe that the API contract will be much more stable if it
receives a higher level of scrutiny than everyday code.  The general quality bar:

- Clearly designate it so that people know it is a Public API (see next section)

- Complete a serious code review and signoff

- Provide AEDoc summaries for each type, member, function parameter, and return value,
  using professional English writing style

- Describe the valid inputs for each parameter and perform appropriate validation

- Provide unit test code coverage for each API function unless there's a good reason why not

- Follow our [coding guidelines](./TSStyleGuide.md) and TSLint rules

- Do not rely on archaic coding practices such as global variables,
  injected JavaScript, reflection, etc that produce brittle API contracts (c'mon!)

None of these is particularly complicated or onerous.  They are the same
best practices that great devs always aspire to; we're just being more careful for
Public APIs, where the design decisions we make will be very costly to change
after third parties have started writing code.


## What is the signoff process?

The process works like this:

- Add your class, interface, or other definition to the **package.json** entry
  point (typically **src/index.ts**).  You need to do this explicitly even if the
  definition is already indirectly reachable via some other type.

- Add a [release tag](https://github.com/Microsoft/web-build-tools/wiki/API-Extractor-~-AEDoc-tags#release-tags)
   (**@alpha** or **@internal**) to your AEDoc, as described below.

- Review the quality bar checklist above, then create a Pull Request,
  and get it approved by someone who:
  - Is a member of the
    [SPPPlat API Approvers](https://onedrive.visualstudio.com/DefaultCollection/SPPPlat/_admin/_security?_a=members)
    VSO group
  - Is aware that they are approving a Public API
  - Is not the same person who created the Pull Request :-)


## Can I export an "internal" API?

Yes, however:

1. This should be very rare.  If you encounter a need for a new framework API,
   third-party developers are likely to encounter the exact same need.  We urge
   you to put in the extra effort to design a Public API that everyone can
   benefit from.  **SharePoint is a developer platform, not just an application.**
   Failing to provide a great developer experience isn't just lazy -- it's like
   stealing from the company!  :-D

2. Even for an **@internal** type, the quality bar is still higher than regular code.
   Once an API is exported to other packages, it becomes much more expensive to change.


## AEDoc for exported APIs

All exported definitions should have informative
[AEDoc](https://github.com/Microsoft/web-build-tools/wiki/API-Extractor-~-AEDoc-tags)
summaries for every type, member, function parameter, and return value.
Write documentation that communicates **non-obvious** information about what your API does
and how to use it.  Your audience is people who are proficient with TypeScript,
have only basic familiarity with SharePoint, and may be completely unfamiliar
with your feature.

**IMPORTANT:** Your code comments for any **@beta** and **@public** types in the SPFx packages
will be published on the [API Reference](https://docs.microsoft.com/en-us/javascript/api/sp-application-base)
web site.  This means that each sentence you write will eventually "go live" on a public stage
for millions of readers to see.  (No pressure! :-) )  To help you out, the UA team maintains an excellent
[Microsoft Writing Style Guide](https://worldready.cloudapp.net/Styleguide/Read?id=2700&topicid=26472)
with concise voicing guidance, plus a ton of practical phrasing examples.  Print out a copy.
Read it.  Wear it on a golden chain like a hip hop medallion and rock those API docs, champ!


## What if I have questions?

E-mail the **sppplatdev@microsoft.com** DL.
This is your Bat Phone(tm) for SPFx related questions.
