﻿<#
.SYNOPSIS
    Parse test case result files and code churn coverage files on VSTS build agent and save the data points to SPO-RT dashboard database.

.DESCRIPTION
    This script shall be configured as a VSTS build step in the pullrequest build definition of sp-client project.

    It shall be added after the build step: 'rush rebuild --vso --production'.

    The code churn coverage files that searched by the script are generated by VerifyCoverageTask gulp task during the build.

    The test result files that searched by the script are generated by karma test result reporter during the build.

    Parsed data points will be saved to SPO-RT dashboard database at: 'na4v1as4h4.database.windows.net'.

    The dashboard of the collected data points can be found at:
    - http://spo-rt/CodeHealth/VstsCodeChurnCoverageTrend?UserAlias=<alias>
    - http://spo-rt/CodeHealth/VstsCodeChurnCoverageRank

.PARAMETER ConnectionString
    Connection string of SPO-RT dashboard database, the RW permission is required.

.PARAMETER ProjectId
    ID (GUID) of VSTS project: SPPPlat.

.PARAMETER RepositoryId
    ID (GUID) of VSTS repository: sp-client.

.PARAMETER CodeChurnCoverageFileName
    Code churn coverage file name.

.PARAMETER TestResultsFileName
    Test results file name.

.PARAMETER TimeOutInSeconds
    Timeout to stop the script in seconds.

.EXAMPLE
    D:\Agent>CollectTestDataOnVSTSBuildAgent.ps1 -ConnectionString "RW_CONN_STR" -ProjectId 'PRJ_GUID' -RepositoryId 'REPO_GUID' -CodeChurnCoverageFileName 'code-churn-coverage.json' -TestResultsFileName 'result.json' -TimeOutInSeconds 10
#>

param(
    [Parameter(Mandatory=$true)]
    [string]$ConnectionString,
    [Parameter(Mandatory=$true)]
    [string]$ProjectId,
    [Parameter(Mandatory=$true)]
    [string]$RepositoryId,
    [Parameter(Mandatory=$true)]
    [string]$CodeChurnCoverageFileName,
    [Parameter(Mandatory=$true)]
    [string]$TestResultsFileName,
    [Parameter(Mandatory=$true)]
    [int]$TimeOutInSeconds
)

function Get-PullRequestId()
{
    if ($SourceBranchName -eq "merge")
    {
        if ($SourceBranch.Split('/').Count -eq 4)
        {
            return $SourceBranch.Split('/')[2]
        }
    }

    -1
}

# ------------------------------------------------------------
# Prerequisites to check
# ------------------------------------------------------------

$WorkingDir = $env:SYSTEM_DEFAULTWORKINGDIRECTORY
$SourceBranchName = $env:BUILD_SOURCEBRANCHNAME
$SourceBranch = $env:BUILD_SOURCEBRANCH
$BuildRunId = $env:BUILD_BUILDID
$PullRequestId = Get-PullRequestId

# ------------------------------------------------------------
# Declare collect test results job
# ------------------------------------------------------------

$CollectTestResultsJobScriptBlock = {
    param(
        [Parameter(Mandatory=$true)]
        [string]$ConnectionString,
        [Parameter(Mandatory=$true)]
        [string]$ProjectId,
        [Parameter(Mandatory=$true)]
        [string]$RepositoryId,
        [Parameter(Mandatory=$true)]
        [string]$PullRequestId,
        [Parameter(Mandatory=$true)]
        [string]$BuildRunId,
        [Parameter(Mandatory=$true)]
        [string]$WorkingDir,
        [Parameter(Mandatory=$true)]
        [string]$TestResultsFileName
    )

    Write-Host "`nCollect test results job start."

    #region Function Definitions

    function Add-TestCaseResults
    {
        [CmdletBinding()]
        param (
            [Parameter(Mandatory=$true)]
            [System.Data.SqlClient.SqlConnection]$Connection,
            [Parameter(Mandatory=$true)]
            [string]$ProjectId,
            [Parameter(Mandatory=$true)]
            [string]$RepositoryId,
            [Parameter(Mandatory=$true)]
            [int]$PullRequestId,
            [Parameter(Mandatory=$true)]
            [int]$BuildRunId,
            [Parameter(Mandatory=$true)]
            [System.Collections.Hashtable]$TestCaseResults
        )

        $DataTable = New-Object -TypeName System.Data.DataTable

        $Col = New-Object -TypeName System.Data.DataColumn
        $Col.ColumnName = 'SubProject'
        $Col.DataType = [String]
        $DataTable.Columns.Add($Col)

        $Col = New-Object -TypeName System.Data.DataColumn
        $Col.ColumnName = 'TestSuite'
        $Col.DataType = [String]
        $DataTable.Columns.Add($Col)

        $Col = New-Object -TypeName System.Data.DataColumn
        $Col.ColumnName = 'TestDescription'
        $Col.DataType = [String]
        $DataTable.Columns.Add($Col)

        $Col = New-Object -TypeName System.Data.DataColumn
        $Col.ColumnName = 'RunStatus'
        $Col.DataType = [String]
        $DataTable.Columns.Add($Col)

        $Col = New-Object -TypeName System.Data.DataColumn
        $Col.ColumnName = 'RunDateUTC'
        $Col.DataType = [DateTime]
        $DataTable.Columns.Add($Col)

        $Col = New-Object -TypeName System.Data.DataColumn
        $Col.ColumnName = 'DurationMs'
        $Col.DataType = [Int32]
        $DataTable.Columns.Add($Col)

        foreach ($Result in $TestCaseResults.Values)
        {
            $Row = $DataTable.NewRow()
            $Row.Item('SubProject') = $Result['SubProject']
            $Row.Item('TestSuite') = $Result['TestSuite']
            $Row.Item('TestDescription') = $Result['TestDescription']
            $Row.Item('RunStatus') = $Result['RunStatus']
            $Row.Item('RunDateUTC') = $Result['RunDateUTC']
            $Row.Item('DurationMs') = $Result['DurationMs']
            $DataTable.Rows.Add($Row)
        }

        $Command = $Connection.CreateCommand()
        $Command.CommandText = 'AddTestCaseResultsFromTable'
        $Command.CommandType = [System.Data.CommandType]'StoredProcedure'

        $Command.Parameters.Add('@TestCaseResultsTable', [System.Data.SqlDbType]'Structured') | Out-Null
        $Command.Parameters['@TestCaseResultsTable'].Value = [System.Data.DataTable]$DataTable

        $Command.Parameters.Add('@ProjectId', [System.Data.SqlDbType]'NVarChar') | Out-Null
        $Command.Parameters['@ProjectId'].Value = [string]$ProjectId

        $Command.Parameters.Add('@RepositoryId', [System.Data.SqlDbType]'NVarChar') | Out-Null
        $Command.Parameters['@RepositoryId'].Value = [string]$RepositoryId

        $Command.Parameters.Add('@PullRequestId', [System.Data.SqlDbType]'Int') | Out-Null
        $Command.Parameters['@PullRequestId'].Value = [int]$PullRequestId

        $Command.Parameters.Add('@BuildRunId', [System.Data.SqlDbType]'Int') | Out-Null
        $Command.Parameters['@BuildRunId'].Value = [int]$BuildRunId

        $Command.ExecuteNonQuery() | Out-Null
    }

    #endregion

    #region Main Script of CollectTestResultsJobScriptBlock

    Try
    {
        $Connection = New-Object -TypeName System.Data.SqlClient.SqlConnection
        $Connection.ConnectionString = $ConnectionString

        $TestCaseResults = @{}

        # ------------------------------------------------------------
        # Get all test results
        # ------------------------------------------------------------

        $RushJsonFile = "{0}\{1}" -f ($WorkingDir, 'rush.json')
        $RushJsonFileContent = Get-Content $RushJsonFile -Raw | Out-String
        # Since rush.json contains comments, have to remove comments in first place, otherwise ConvertFrom-Json will fail.
        $RushJsonFileContent = $RushJsonFileContent -Replace "//.*|/\*[\d\D]*\*/" , ""
        $RushJsonObject = ConvertFrom-Json -InputObject $RushJsonFileContent

        foreach ($Project in $RushJsonObject.projects)
        {
            $TestCaseResultFilePath = "{0}\{1}\temp\testResult\{2}" -f ($WorkingDir, $Project.projectFolder, $TestResultsFileName)
            $TestCaseResultFilePath = $TestCaseResultFilePath -Replace "/", "\"

            if (Test-Path $TestCaseResultFilePath -PathType Leaf)
            {
                $TestCaseResultFileContent = Get-Content $TestCaseResultFilePath -Raw | Out-String
                $TestCaseResultFileJson = ConvertFrom-Json $TestCaseResultFileContent

                foreach ($Result in $TestCaseResultFileJson)
                {
                    $RunStatus = "Failed"
                    if ($Result.success)
                    {
                        $RunStatus = "Passed"
                    }

                    if ($Result.skipped)
                    {
                        $RunStatus = "Skipped"
                    }

                    $Item = @{
                        'SubProject' = $Project.projectFolder;
                        'TestSuite' = $Result.suite -Join "#";
                        'TestDescription' = $Result.description;
                        'RunStatus' = $RunStatus;
                        'RunDateUTC' = Get-Date;
                        'DurationMs' = $Result.time
                    }

                    $TestCaseResults[$Item.SubProject + "#" + $Item.TestSuite + "#" + $Item.TestDescription] = $Item
                }
                # Delete the test result file after the data is read
                Remove-Item $TestCaseResultFilePath
            }
        }

        if ($TestCaseResults.Count -eq 0)
        {
            Write-Host "Nothing to process, stop parsing data and exit."
            exit
        }

        # ------------------------------------------------------------
        # Store test result data points to database of SPO-RT dashboard
        # ------------------------------------------------------------

        Write-Host "Total test cases need to process: $($TestCaseResults.Count)"

        $Connection.Open()

        if ($TestCaseResults.Count -gt 0)
        {
            $Param = @{
                'Connection' = $Connection;
                'ProjectId' = $ProjectId;
                'RepositoryId' = $RepositoryId;
                'PullRequestId' = $PullRequestId;
                'BuildRunId' = $BuildRunId;
                'TestCaseResults' = $TestCaseResults;
            }

            Add-TestCaseResults @Param
        }

    }
    Catch
    {
        Write-Host "!!! Exceptions !!!"
        Write-Host $_.Exception.Message
    }
    Finally
    {
        if ($Connection -ne $null)
        {
            $Connection.Close()
        }
    }

    Write-Host "Collect test results job complete."
    exit

    #endregion
}

# ------------------------------------------------------------
# Declare collect code churn coverage job
# ------------------------------------------------------------

$CollectCodeChurnCoverageJobScriptBlock = {
    param(
        [Parameter(Mandatory=$true)]
        [string]$ConnectionString,
        [Parameter(Mandatory=$true)]
        [string]$ProjectId,
        [Parameter(Mandatory=$true)]
        [string]$RepositoryId,
        [Parameter(Mandatory=$true)]
        [string]$PullRequestId,
        [Parameter(Mandatory=$true)]
        [string]$WorkingDir,
        [Parameter(Mandatory=$true)]
        [string]$CodeChurnCoverageFileName
    )

    Write-Host "`nCollect code churn coverage job start."

    if ($PullRequestId -eq -1)
    {
        Write-Host "This is not a merge pull request, stop collecting data."
        exit
    }

    #region Function Definitions

    function Convert-Count
    {
        [CmdletBinding()]
        param (
            $CountValue
        )

        if ($CountValue -ne $null)
        {
            $CountValue
        }
        else
        {
            -1
        }
    }

    function Get-FilePathId
    {
        [CmdletBinding()]
        param (
            [Parameter(Mandatory=$true)]
            [System.Data.SqlClient.SqlConnection]$Connection,
            [Parameter(Mandatory=$true)]
            [string]$ProjectId,
            [Parameter(Mandatory=$true)]
            [string]$RepositoryId,
            [Parameter(Mandatory=$true)]
            [string]$FilePath
        )

        $Command = $Connection.CreateCommand()
        $Command.CommandText = 'EnsureFilePath'
        $Command.CommandType = [System.Data.CommandType]'StoredProcedure'

        $Command.Parameters.Add('@ProjectId', [System.Data.SqlDbType]'NVarChar') | Out-Null
        $Command.Parameters['@ProjectId'].Value = [string]$ProjectId

        $Command.Parameters.Add('@RepositoryId', [System.Data.SqlDbType]'NVarChar') | Out-Null
        $Command.Parameters['@RepositoryId'].Value = [string]$RepositoryId

        $Command.Parameters.Add('@FilePath', [System.Data.SqlDbType]'NVarChar') | Out-Null
        $Command.Parameters['@FilePath'].Value = [string]$FilePath

        $Command.Parameters.Add('@FilePathId', [System.Data.SqlDbType]'Int') | Out-Null
        $Command.Parameters['@FilePathId'].Direction = [System.Data.ParameterDirection]::Output

        $Command.ExecuteNonQuery() | Out-Null

        # return FilePathId
        $Command.Parameters['@FilePathId'].Value
    }

    function Add-CodeCoverage
    {
        [CmdletBinding()]
        param (
            [Parameter(Mandatory=$true)]
            [System.Data.SqlClient.SqlConnection]$Connection,
            [Parameter(Mandatory=$false)]
            [System.Array]$CovInfoList
        )

        $DataTable = New-Object -TypeName System.Data.DataTable

        $Col = New-Object -TypeName System.Data.DataColumn
        $Col.ColumnName = 'FilePathId'
        $Col.DataType = [Int32]
        $DataTable.Columns.Add($Col)

        $Col = New-Object -TypeName System.Data.DataColumn
        $Col.ColumnName = 'PullRequestId'
        $Col.DataType = [Int32]
        $DataTable.Columns.Add($Col)

        $Col = New-Object -TypeName System.Data.DataColumn
        $Col.ColumnName = 'StatementCount'
        $Col.DataType = [Int32]
        $DataTable.Columns.Add($Col)

        $Col = New-Object -TypeName System.Data.DataColumn
        $Col.ColumnName = 'HitStatementCount'
        $Col.DataType = [Int32]
        $DataTable.Columns.Add($Col)

        $Col = New-Object -TypeName System.Data.DataColumn
        $Col.ColumnName = 'BranchCount'
        $Col.DataType = [Int32]
        $DataTable.Columns.Add($Col)

        $Col = New-Object -TypeName System.Data.DataColumn
        $Col.ColumnName = 'HitBranchCount'
        $Col.DataType = [Int32]
        $DataTable.Columns.Add($Col)

        $Col = New-Object -TypeName System.Data.DataColumn
        $Col.ColumnName = 'FunctionCount'
        $Col.DataType = [Int32]
        $DataTable.Columns.Add($Col)

        $Col = New-Object -TypeName System.Data.DataColumn
        $Col.ColumnName = 'HitFunctionCount'
        $Col.DataType = [Int32]
        $DataTable.Columns.Add($Col)

        $Col = New-Object -TypeName System.Data.DataColumn
        $Col.ColumnName = 'LineCount'
        $Col.DataType = [Int32]
        $DataTable.Columns.Add($Col)

        $Col = New-Object -TypeName System.Data.DataColumn
        $Col.ColumnName = 'HitLineCount'
        $Col.DataType = [Int32]
        $DataTable.Columns.Add($Col)

        $Col = New-Object -TypeName System.Data.DataColumn
        $Col.ColumnName = 'CodeChurnCount'
        $Col.DataType = [Int32]
        $DataTable.Columns.Add($Col)

        $Col = New-Object -TypeName System.Data.DataColumn
        $Col.ColumnName = 'HitCodeChurnCount'
        $Col.DataType = [Int32]
        $DataTable.Columns.Add($Col)

        foreach ($CovInfo in $CovInfoList)
        {
            $Row = $DataTable.NewRow()
            $Row.Item('FilePathId') = $CovInfo['filePathId']
            $Row.Item('PullRequestId') = $CovInfo['pullRequestId']
            $Row.Item('StatementCount') = $CovInfo['statementCount']
            $Row.Item('HitStatementCount') = $CovInfo['hitStatementCount']
            $Row.Item('BranchCount') = $CovInfo['branchCount']
            $Row.Item('HitBranchCount') = $CovInfo['hitBranchCount']
            $Row.Item('FunctionCount') = $CovInfo['functionCount']
            $Row.Item('HitFunctionCount') = $CovInfo['hitFunctionCount']
            $Row.Item('LineCount') = $CovInfo['lineCount']
            $Row.Item('HitLineCount') = $CovInfo['hitLineCount']
            $Row.Item('CodeChurnCount') = $CovInfo['codeChurnCount']
            $Row.Item('HitCodeChurnCount') = $CovInfo['hitCodeChurnCount']
            $DataTable.Rows.Add($Row)
        }

        $Command = $Connection.CreateCommand()
        $Command.CommandText = 'AddCodeCoverageFromTable'
        $Command.CommandType = [System.Data.CommandType]'StoredProcedure'

        $Command.Parameters.Add('@ProjectId', [System.Data.SqlDbType]'NVarChar') | Out-Null
        $Command.Parameters['@ProjectId'].Value = [string]$ProjectId

        $Command.Parameters.Add('@RepositoryId', [System.Data.SqlDbType]'NVarChar') | Out-Null
        $Command.Parameters['@RepositoryId'].Value = [string]$RepositoryId

        $Command.Parameters.Add('@PullRequestId', [System.Data.SqlDbType]'Int') | Out-Null
        $Command.Parameters['@PullRequestId'].Value = [int]$PullRequestId

        $Command.Parameters.Add('@CodeCoverageTable', [System.Data.SqlDbType]'Structured') | Out-Null
        $Command.Parameters['@CodeCoverageTable'].Value = [System.Data.DataTable]$DataTable

        $Command.ExecuteNonQuery() | Out-Null
    }

    function Add-CoverageJsonFile
    {
        [CmdletBinding()]
        param (
            [Parameter(Mandatory=$true)]
            [System.Data.SqlClient.SqlConnection]$Connection,
            [Parameter(Mandatory=$true)]
            [string]$ProjectId,
            [Parameter(Mandatory=$true)]
            [string]$RepositoryId,
            [Parameter(Mandatory=$true)]
            [int]$PullRequestId,
            [Parameter(Mandatory=$false)]
            [string]$Content
        )

        $Command = $Connection.CreateCommand()
        $Command.CommandText = '[AddCoverageJsonFiles]'
        $Command.CommandType = [System.Data.CommandType]'StoredProcedure'

        $Command.Parameters.Add('@ProjectId', [System.Data.SqlDbType]'NVarChar') | Out-Null
        $Command.Parameters['@ProjectId'].Value = [string]$ProjectId

        $Command.Parameters.Add('@RepositoryId', [System.Data.SqlDbType]'NVarChar') | Out-Null
        $Command.Parameters['@RepositoryId'].Value = [string]$RepositoryId

        $Command.Parameters.Add('@PullRequestId', [System.Data.SqlDbType]'Int') | Out-Null
        $Command.Parameters['@PullRequestId'].Value = [int]$PullRequestId

        $Command.Parameters.Add('@Content', [System.Data.SqlDbType]'NVarChar') | Out-Null
        $Command.Parameters['@Content'].Value = [string]$Content

        $Command.ExecuteNonQuery() | Out-Null
    }

    #endregion

    #region Main Script of CollectCodeChurnCoverageJobScriptBlock

    Try
    {
        $Connection = New-Object -TypeName System.Data.SqlClient.SqlConnection
        $Connection.ConnectionString = $ConnectionString

        $CodeChurnCoverageFiles = @()
        $CovInfoList = @()
        $CodeChurnCoverageFileContents = @()

        # ------------------------------------------------------------
        # Get all path of code churn coverage files
        # ------------------------------------------------------------

        $RushJsonFile = "{0}\{1}" -f ($WorkingDir, 'rush.json')
        $RushJsonFileContent = Get-Content $RushJsonFile -Raw | Out-String
        # Since rush.json contains comments, have to remove comments in first place, otherwise ConvertFrom-Json will fail.
        $RushJsonFileContent = $RushJsonFileContent -Replace "//.*|/\*[\d\D]*\*/" , ""
        $RushJsonObject = ConvertFrom-Json -InputObject $RushJsonFileContent

        foreach ($Project in $RushJsonObject.projects)
        {
            $CodeChurnCoverageFilePath = "{0}\{1}\temp\coverage\{2}" -f ($WorkingDir, $Project.projectFolder, $CodeChurnCoverageFileName)
            $CodeChurnCoverageFilePath = $CodeChurnCoverageFilePath -Replace "/", "\"

            if (Test-Path $CodeChurnCoverageFilePath -PathType Leaf)
            {
                $CodeChurnCoverageFiles += $CodeChurnCoverageFilePath
            }
        }

        if ($CodeChurnCoverageFiles.Count -eq 0)
        {
            Write-Host "Nothing to process, stop parsing data and exit."
            exit
        }

        # ------------------------------------------------------------
        # Collect code churn coverage data
        # ------------------------------------------------------------

        Write-Host "Total code churn coverage files need to process: $($CodeChurnCoverageFiles.Count)"

        $Connection.Open()

        foreach ($FileName in $CodeChurnCoverageFiles)
        {
            $FileContent = Get-Content $FileName -Raw | Out-String
            $FileJsonObject = Get-Content $FileName -Raw | ConvertFrom-Json
            $IsFileContentAppended = $false

            foreach ($SrcFile in $FileJsonObject) {
                if ($SrcFile.fullName.Length -gt 0)
                {
                    Write-Host "Processing coverage info for file: $($SrcFile.fullName)"

                    if (!$IsFileContentAppended)
                    {
                        $CodeChurnCoverageFileContents += $FileContent
                        $IsFileContentAppended = $true
                    }

                    $FilePath = $SrcFile.fullName.Substring($WorkingDir.Length) -Replace '\\', '/'

                    $Param = @{
                        'connection' = $Connection;
                        'projectId' = $ProjectId;
                        'repositoryId' = $RepositoryId;
                        'filePath' = $FilePath;
                    }

                    $FilePathId = Get-FilePathId @Param

                    $CovInfo = @{
                        'filePathId' = $FilePathId;
                        'pullRequestId' = $PullRequestId;
                        'statementCount' = Convert-Count $SrcFile.fileCovInfo.stmtsCount;
                        'hitStatementCount' = Convert-Count $SrcFile.fileCovInfo.covStmtsCount;
                        'branchCount' = Convert-Count $SrcFile.fileCovInfo.branchCount;
                        'hitBranchCount' = Convert-Count $SrcFile.fileCovInfo.covBranchCount;
                        'functionCount' = Convert-Count $SrcFile.fileCovInfo.funcsCount;
                        'hitFunctionCount' = Convert-Count $SrcFile.fileCovInfo.covFuncsCount;
                        'lineCount' = Convert-Count $SrcFile.fileCovInfo.linesCount;
                        'hitLineCount' = Convert-Count $SrcFile.fileCovInfo.covLinesCount;
                        'codeChurnCount' = Convert-Count $SrcFile.codeChurnCovInfo.linesCount;
                        'hitCodeChurnCount' = Convert-Count $SrcFile.codeChurnCovInfo.covLinesCount;
                    }

                    $CovInfoList += $CovInfo
                }

            }
            # Delete the code churn coverage file after the data is read
            Remove-Item $FileName
        }

        # ------------------------------------------------------------
        # Store coverage data points to database of SPO-RT dashboard
        # ------------------------------------------------------------

        Add-CodeCoverage $Connection $CovInfoList

        $Param = @{
            'connection' = $Connection;
            'projectId' = $ProjectId;
            'repositoryId' = $RepositoryId;
            'pullRequestId' = $PullRequestId;
            'content' = $CodeChurnCoverageFileContents -Join ',';
        }

        Add-CoverageJsonFile @Param
    }
    Catch
    {
        Write-Host "!!! Exceptions !!!"
        Write-Host $_.Exception.Message
    }
    Finally
    {
        if ($Connection -ne $null)
        {
            $Connection.Close()
        }
    }

    Write-Host "Collect code churn coverage job complete."
    exit

    #endregion
}

# ------------------------------------------------------------
# Script Execution
# ------------------------------------------------------------

# If the job execution time of CollectTestResultsJobScriptBlock or CollectCodeChurnCoverageJobScriptBlock is longer than the TimeOutInSeconds, it will be terminated.

Try
{
    $CollectTestResultsJobGUID = [system.Guid]::NewGuid()
    $CollectTestResultsJob = Start-Job -Name $CollectTestResultsJobGUID -ScriptBlock $CollectTestResultsJobScriptBlock -ArgumentList $ConnectionString, $ProjectId, $RepositoryId, $PullRequestId, $BuildRunId, $WorkingDir, $TestResultsFileName
    $CollectTestResultsJob | Wait-Job -Timeout $TimeOutInSeconds
    $CollectTestResultsJob | Where-Object {$_.State -ne "Completed"} | Stop-Job

    $CollectCodeChurnCoverageJobGUID = [system.Guid]::NewGuid()
    $CollectCodeChurnCoverageJob = Start-Job -Name $CollectCodeChurnCoverageJobGUID -ScriptBlock $CollectCodeChurnCoverageJobScriptBlock -ArgumentList $ConnectionString, $ProjectId, $RepositoryId, $PullRequestId, $WorkingDir, $CodeChurnCoverageFileName
    $CollectCodeChurnCoverageJob | Wait-Job -Timeout $TimeOutInSeconds
    $CollectCodeChurnCoverageJob | Where-Object {$_.State -ne "Completed"} | Stop-Job

    Receive-Job $CollectTestResultsJob, $CollectCodeChurnCoverageJob
}
Catch
{
    Write-Host "!!! Job execution exceptions !!!"
    Write-Host $_.Exception.Message
}
