/**
 * @file BaseCollection.tsx
 * @copyright Microsoft Corporation. All rights reserved.
 */

import * as React from 'react';
import * as ReactDOM from 'react-dom';
import { DisplayMode } from '@microsoft/sp-core-library';
import { _EngagementLogger, _QosMonitor } from '@ms/sp-telemetry';
import {
  autobind,
  BaseComponent,
  MessageBar
} from 'office-ui-fabric-react';

import { IBaseCollectionWebPartViewState } from './models/IBaseCollectionWebPartViewState';
import { IBaseCollectionWebPartViewProperties } from './models/IBaseCollectionWebPartViewProperties';

import { ReadingMode, ScreenReaderAlert } from '@ms/sp-component-utilities';
import WebPartChrome from '../webPartChrome/WebPartChrome';

import strings from './loc/Strings.resx';
import styles from './style/BaseCollection.module.scss';

export default class BaseCollection
  extends BaseComponent<IBaseCollectionWebPartViewProperties, IBaseCollectionWebPartViewState> {

  private _webPartChrome: WebPartChrome;
  private _layoutContainer: HTMLElement;

  constructor(props: IBaseCollectionWebPartViewProperties) {
    super(props);
  }

  public get title(): string {
    return this._webPartChrome ? this._webPartChrome.title : '';
  }

  public render(): JSX.Element {
    const qosMonitorRender: _QosMonitor = new _QosMonitor(`${this.props.qosPrefix}ComponentRender`);
    let emptyResult: JSX.Element;

    // if we have a response and the response contains 0 rows
    const hasEmptyResponse: boolean = !!this.props.items && this.props.items.length < 1;
    let emptyResponseText: string;
    if (hasEmptyResponse && !this.props.hasInitialLoadingState) {
      // if no results and hideWebPartWhenEmpty is set true
      if (this.props.hideWebPartWhenEmpty && this.props.displayMode === DisplayMode.Read) {
        emptyResponseText = undefined;
      } else {
        emptyResponseText = strings.NoResults;
      }

      try {
        emptyResult = <span className={ styles.emptyResult }>{ emptyResponseText }</span>;
      } catch (er) {
        this.props.errorCallback(er);
        qosMonitorRender.writeUnexpectedFailure('FailedToRenderEmptyResult');
      }
    }

    // If not displaying the "no results" message, continue to build the view of the data.
    if (!emptyResult) {
      const qosMonitorRenderLayout: _QosMonitor = new _QosMonitor(`${this.props.qosPrefix}ComponentRenderLayout`);
      try {
        // tslint:disable-next-line:no-any
        this.props.layoutInfo.loadLayout().then((componentClass: React.ComponentClass<any>) => {
          const layout: JSX.Element = React.createElement(componentClass, this.props);
          ReactDOM.render(layout, this._layoutContainer);
          qosMonitorRenderLayout.writeSuccess();
        });
      } catch (ex) {
        this.props.errorCallback(ex);
        qosMonitorRenderLayout.writeUnexpectedFailure('FailedToRenderLayout', ex);
      }
    }

    let result: JSX.Element;
    try {
      let title: string = this.props.title || '';

      if (this.props.displayMode === DisplayMode.Read) {
        // if webpart opts in, render the placeholder title in read mode
        if (this.props.displayPlaceholderTitleInReadMode && !title) {
          title = this.props.defaultTitle;
        }
      }

      // Display the title if showing the initial loading state, if there is data to display,
      // or the webpart is rendering with an empty response message. Avoid displaying the title
      // if there is no content below it.
      const isRenderingContent: boolean =
        this.props.hasInitialLoadingState || hasEmptyResponse === false || !!emptyResponseText;
      const showTitle: boolean =
        (this.props.showTitle && this.props.displayMode === DisplayMode.Edit) ||
        (isRenderingContent && (this.props.showTitle && !!title));

      result =
        <WebPartChrome
          ref={ this._resolveRef('_webPartChrome') }
          showTitle={ this.props.showTitle && showTitle }
          title={ title }
          onTitleChange={ this._onTitleChange }
          placeholder={ this.props.defaultTitle }
          isEditMode={ this.props.displayMode === DisplayMode.Edit }
          seeAllLink={ this.props.seeAllLink }
          addButton={ this.props.addButton }
          contentElement={ this._renderWebPartDisplay(emptyResult) }
          pagination={ this.props.paginationProps }
        />;
    } catch (ex) {
      this.props.errorCallback(ex);
      qosMonitorRender.writeUnexpectedFailure('FailedToRender', ex);
    }

    qosMonitorRender.writeSuccess();
    return result;
  }

  public focusOnTitle(): void {
    if (this._webPartChrome) {
      this._webPartChrome.focusOnTitle();
    }
  }

  private _renderWebPartDisplay(emptyResult: JSX.Element): JSX.Element {
    return (
      <div>
        { this.props.displayMode === DisplayMode.Edit && !this.props.renderedOnce &&
          this.props.addToPageScreenReaderLabel &&
          <ScreenReaderAlert
            readingMode={ ReadingMode.ReadImmediately }
            message={ this.props.addToPageScreenReaderLabel }
          />
        }
        { this.props.messageView &&
          <MessageBar
            className={ styles.messageView }
            ariaLabel={ this.props.messageView.ariaLabel }
            messageBarType={ this.props.messageView.messageBarType }>
            { this.props.messageView.message }
          </MessageBar>
        }
        { emptyResult || <div ref={(div) => this._layoutContainer = div}></div> }
      </div>);
  }

  @autobind
  private _onTitleChange(ev: React.FormEvent<HTMLTextAreaElement>): void {
    _EngagementLogger.logEvent('WebPartTitle.Click', ev.type);
  }
}
