/**
 * @file BaseCollectionWebPart.ts
 * @copyright Microsoft Corporation. All rights reserved.
 *
 * @remark A base class for WebParts that share common features for rendering SharePoint content.  Includes support
 *         for multiple layouts and data providers via display field mapping.
 */

import * as React from 'react';
import * as ReactDOM from 'react-dom';
import * as lodash from '@microsoft/sp-lodash-subset';

import { _QosMonitor } from '@ms/sp-telemetry';
import { _SPFlight, Environment, EnvironmentType, ServiceScope } from '@microsoft/sp-core-library';
import {
  BaseClientSideWebPart,
  IPropertyPaneChoiceGroupOption,
  _IPropertyPaneChoiceGroupOptionInternal,
  IWebPartPropertiesMetadata
} from '@microsoft/sp-webpart-base';
import { Async } from 'office-ui-fabric-react';
import {
  PageContext,
  SPPermission
} from '@microsoft/sp-page-context';
import { IDataProvider, IDataRow } from '@ms/sp-component-utilities';

import DisplayFieldMap from '../displayFieldMap/DisplayFieldMap';
import BaseCollection from './BaseCollection';
import {
  IBaseCollectionWebPartOptions,
  IBaseCollectionWebPartLayoutState
} from './models/IBaseCollectionWebPartOptions';
import { IBaseCollectionWebPartProperties } from './models/IBaseCollectionWebPartProperties';
import { IBaseCollectionWebPartViewProperties } from './models/IBaseCollectionWebPartViewProperties';
import { IBaseCollectionWebPartViewState } from './models/IBaseCollectionWebPartViewState';

import { IDataProviderInfo } from './models/IDataProviderInfo';
import { ILayoutInfo } from './models/ILayoutInfo';
import { ILayoutOption } from './models/ILayoutOption';
import { MySiteWebPartCacheManager } from '../mySiteWebPartCache/MySiteWebPartCacheManager';
import strings from './loc/Strings.resx';

export enum RefreshType {
  Append = 0,
  Replace = 1
}

export abstract class BaseCollectionWebPart<T> extends BaseClientSideWebPart<IBaseCollectionWebPartProperties & T> {
  /**
   * A generated method called from onPropertyPaneFieldChanged to retrieve data and re-render.  It will wait for a
   * break in input on the Property Pane to avoid too many data requests.
   */
  public delayedGetDataAndRender: () => void;

  protected pageContext: PageContext;
  private _async: Async;
  private _options: IBaseCollectionWebPartOptions;
  private _layoutInfos: ILayoutInfo[];
  private _activeLayoutInfo: ILayoutInfo;
  private _activeDataProviderInfo: IDataProviderInfo;
  private _renderError: (error: Error) => void;
  private _response: IDataRow[];
  private _component: React.Component<IBaseCollectionWebPartViewProperties, IBaseCollectionWebPartViewState>;
  private _mySiteCacheWebPartManager: MySiteWebPartCacheManager;
  private _usingEmptyStateRows: boolean;

  /**
   * @returns The react component of the actively rendered BaseCollectionWebPart
   * @see IBaseCollectionWebPartViewProperties
   * @see IBaseCollectionWebPartViewState
   */
  protected get component(): React.Component<IBaseCollectionWebPartViewProperties, IBaseCollectionWebPartViewState> {
    return this._component;
  }

  /**
   * @returns Configuration property bag for the actively used layout by the BaseCollectionWebPart
   * @see ILayoutInfo
   */
  protected get activeLayoutInfo(): ILayoutInfo {
    return this._activeLayoutInfo;
  }

  /**
   * @returns Configuration property bag for the actively used data provider by the BaseCollectionWebPart
   * @see IDataProviderInfo
   */
  protected get activeDataProviderInfo(): IDataProviderInfo {
    return this._activeDataProviderInfo;
  }

  /**
   * @returns True if current user has permissions to add list items. False otherwise.
   */
  protected get isContributor(): boolean {
    const permissions: SPPermission = this.context.pageContext.web.permissions;
    return !!permissions && permissions.hasPermission(SPPermission.addListItems);
  }

/**
 * Sets or Updates a prop value in the active Layout. The prop change will take effect on the next rendering.
 */
  protected updateActiveLayoutProp(prop: string, value: any): void { // tslint:disable-line:no-any
    this._activeLayoutInfo.properties[prop] = value;
  }

  /**
   * Initialize the WebPart.  If overridden, be sure to call super.onInit() at some point and return
   * it's value from your onInit.
   */
  protected onInit(): Promise<void> {
    this._response = [];
    this._options = this.options;
    const qosMonitorMount: _QosMonitor = new _QosMonitor(`${this._options.qosPrefix}Mount`);

    this._renderError = this.renderError.bind(this);
    this.setDefaultProperties(this.properties);

    this._layoutInfos = this.layouts;
    // @todo: support dynamic loading of layouts
    // this._loadLayoutsFromUrls(this._options.layoutUrls);
    this._setSelectedLayout();

    // @todo: support dynamic loading of dataProviders
    // this._loadDataProvidersFromUrls(this._options.dataProviderUrls);
    this._setSelectedDataProvider();

    this._getDisplayFieldMaps();

    // perform initial empty render if the layout supports it
    if (this._activeLayoutInfo.hasInitialLoadingState) {
      this.render();
    }

    return new Promise<void>((complete: () => void, reject: (error: Error) => void) => {
      const serviceScope: ServiceScope = this.context.serviceScope;
      serviceScope.whenFinished(() => {
        // Get resources from the service scope.
        this.pageContext = serviceScope.consume(PageContext.serviceKey);
        if (_SPFlight.isEnabled(959 /*WexModernWebPartCache*/) && this._options.canCacheResults) {
          this._mySiteCacheWebPartManager = serviceScope.consume(MySiteWebPartCacheManager.serviceKey);
        }
        this._setPageContextVariables();

        this._async = new Async(this);
        // 500 is totally arbitrary from just playing around / suggestions from Stack Overflow
        // [!WEX VSO: 207916] - Investigate reasonable, user-friendly time delay for text field rendering
        const delayInMilliseconds: number =
          this._options.delayBeforeNoResultsMessage === undefined ? 500 : this._options.delayBeforeNoResultsMessage;
        // @todo [!WEX VSO: 323801] - Consider initializing this in the constructor.
        this.delayedGetDataAndRender = this._async.debounce(this.refreshData, delayInMilliseconds);

        let request: Promise<IDataRow[]>;
        const prefetchedData: IDataRow[] = this.getPrefetchedData(this._activeDataProviderInfo.id, this._activeLayoutInfo.id);  // tslint:disable-line

        if (prefetchedData !== undefined) {
          request = new Promise<IDataRow[]>((resolve) => {
            return resolve(prefetchedData);
          });
        } else if (_SPFlight.isEnabled(959 /*WexModernWebPartCache*/) && this._options.canCacheResults) {
          request = this._mySiteCacheWebPartManager.requestWebPartCacheData(
            this._requestData.bind(this),
            this.setResponseAndRender.bind(this),
            this.getRequest(this._activeDataProviderInfo.id, this._activeLayoutInfo.id),
            this.context.instanceId,
            this.context.manifest.alias);
        } else {
          request = this._requestData();
        }

        request
          .then((response: IDataRow[]) => {
            this._response = response;
            if (!this.renderedOnce) {
              qosMonitorMount.writeSuccess();
            }

            complete();
          })
          .catch((error: Error) => {
            this.renderError(error);
            qosMonitorMount.writeUnexpectedFailure('FailedInitialDataRequest', error);
            reject(error);
          });
      });
    });
  }

  /**
   * Override to initialize which features of the BaseCollectionWebPart are utilized.
   *
   * @returns A hash object with the values to initialize the various options of the BaseCollectionWebPart.
   * See the documentation for IBaseCollectionWebPartOptions.
   */
  protected abstract get options(): IBaseCollectionWebPartOptions;

  /**
   * Override to return that layout(s) available for rendering.
   */
  protected abstract get layouts(): ILayoutInfo[];

  /**
   * Call this to set the active layout by id.
   *
   * @param layoutId - Indicates which layout is selected for rendering.
   */
  protected setLayoutById(layoutId: string): void {
    this.properties.layoutId = layoutId;
    this._setSelectedLayout();
  }

  /**
   * Call this to set the active layout by key.
   *
   * @param key - Indicates which layout is selected for rendering.
   */
  protected setLayoutByKey(key: number): void {
    for (const layout of this._layoutInfos) {
      if (layout.option && layout.option.key === key) {
        this.setLayoutById(layout.id);
        break;
      }
    }
  }

  /**
   * Override to return that available data providers.
   */
  protected abstract get dataProviders(): IDataProviderInfo[];

  /**
   * Override to get Prefetched Data if available
  */
  protected getPrefetchedData(dataProviderId: string, layoutId: string): IDataRow[] {
    return undefined;
  }

  /**
   * Returns the active data provider.  The intent is to allow your web part to invoke
   * additional data provider methods in order to support CRUD operations.
   */
  // tslint:disable:no-any
  protected get dataProvider(): Promise<IDataProvider<any, any>> {
    const promise: Promise<IDataProvider<any, any>> = new Promise<IDataProvider<any, any>> (
      (complete: (response: IDataProvider<any, any>) => void, reject: (error: Error) => void) => {
        if (this._activeDataProviderInfo.dataProvider) {
          // Avoid re-allocating the data provider.
          complete(this._activeDataProviderInfo.dataProvider);
        } else {
          // Load the data provider if it has not been loaded or the selected data provider has changed.
          this._activeDataProviderInfo.loadDataProvider(this._activeDataProviderInfo.options ||
                                                        { serviceScope: this.context.serviceScope })
          .then((dataProvider: IDataProvider<any, any>) => {
            // store the instance so it does not have to be re-allocated
            this._activeDataProviderInfo.dataProvider = dataProvider;
            complete(dataProvider);
          }, (error: Error) => {
            reject(error);
          });
        }
      });
    return promise;
  }
  // tslint:enable:no-any

  /**
   * Returns true if the web part is being populated with empty state rows.
   */
  protected get usingEmptyStateRows(): boolean {
    return this._usingEmptyStateRows;
  }

  /**
   * Call this to set the active data provider by id.
   *
   * @param dataProviderId - Indicates the selected data provider.
   */
  protected setDataProviderById(dataProviderId: string): void {
    this.properties.dataProviderId = dataProviderId;
    this._setSelectedDataProvider();
  }

  /**
   * Call this to set the active data provider by key.
   *
   * @param key - Indicates the selected data provider.
   */
  protected setDataProviderByKey(key: number): void {
    for (const dataProvider of this.dataProviders) {
      if (dataProvider.option && dataProvider.option.key === key) {
        this.setDataProviderById(dataProvider.id);
        break;
      }
    }
  }

  /**
   * Override to return the display field map appropriate for the active layout and data provider.
   *
   * @param layoutId - Indicates the active layout.
   * @param dataProviderId - Indicates the active data provider.
   *
   * @returns A display field map.
   */
  protected getDisplayFieldMap(layoutId: string, dataProviderId: string): DisplayFieldMap | undefined {
    // With no mapping the data will be sent to the layout as received from the data provider.
    return undefined;
  }

  /*
   * The derived WebPart should implement getPropertyPaneConfiguration(): IPropertyPaneConfiguration
   * per the requirements of the BaseClientSideWebPart.
   */

  /**
   * Returns configuration panel for the layout selections.  This should be leveraged in the WebPart's
   * getPropertyPaneConfiguration implementation.
   *
   * @returns An array of layout options derived from the WebPart's layouts.
   */
  protected get propertyPaneLayoutOptions(): IPropertyPaneChoiceGroupOption[] {
    const iconSidePixels: number = 32;

    return this._layoutInfos.map((layoutInfo: ILayoutInfo): _IPropertyPaneChoiceGroupOptionInternal => {
      const option: ILayoutOption = layoutInfo.option;
      return {
        imageSize: { width: iconSidePixels, height: iconSidePixels },
        imageSrc: option.imageSrc,
        key: option.key,
        text: option.title,
        themableSvg: option.themableSvg,
        iconProps: option.iconProps
      };
    });
  }

  /**
   * Override this function to convert the WebPart properties to the appropriate options for the data provider's
   * requestData function call.
   *
   * @param dataProviderId - Indicates the active data provider.
   *
   * @param layoutId - Indicates the active layout.
   *
   * @returns Any kind of input required by the data provider's requestData method.
   */
  protected abstract getRequest(dataProviderId: string, layoutId: string): any; // tslint:disable-line:no-any

  /**
   * Call refreshData() from your WebPart to invoke a new request for data.
   *
   * @param request  - Optional input for the data provider. If not provided, the getRequest method will be invoked to
   * obtain the input for the data provider.
   * @param refreshType - Defaults to REFRESH_TYPE.REPLACE which replaces the current data with the results from the
   * new request.  Set to REFRESH_TYPE.APPEND in order to append the new data to the end of existing results.
   * @param shouldRender - Defaults to true, in which case the WebPart redraws.  Set to false to prevent a redraw so
   * that the WebPart can call the render() method on its own timing.
   *
   * @returns A promise with an IDataRow[] which is the resulting data available to the WebPart.
   */
  protected refreshData(
    request?: any, // tslint:disable-line:no-any
    refreshType: RefreshType = RefreshType.Replace,
    shouldRender: boolean = true
    ): Promise<IDataRow[]> {

    return this._requestData(request).then((response: IDataRow[]): Promise<IDataRow[]> => {
        this._response = refreshType === RefreshType.Append ? this._response.concat(response) : response;
        if (shouldRender) {
          this.render();
        }
        return Promise.resolve(this._response);
      }).catch((error: Error) => {
        this.renderError(error);
        return Promise.resolve(undefined);
      });
  }

  /**
   * Override this function to perform any adjustments to a data row before it is rendered.
   * The default implementation does not alter the data.
   *
   * @param layoutId - Indicates the active layout.
   * @param dataProviderId - Indicates the active data provider.
   * @param row - A row of data as received from the data source.
   *
   * @returns A row of data after it has been transposed from the original form to the desired format.
   */
  protected transposeRow(layoutId: string, dataProviderId: string, row: any): any { // tslint:disable-line:no-any
    return row;
  }

  /**
   * Performs the display field map transpose of your data provider results and renders the WebPart with the
   * provided layout.
   */
  protected render(): void {
    const qosMonitor: _QosMonitor = new _QosMonitor(`${this._options.qosPrefix}Render`);

    if (!this._activeLayoutInfo) {
      const error: Error = new Error(strings.ErrorNoLayout);
      this.renderError(error);
      qosMonitor.writeUnexpectedFailure('FailedNoLayout', error);
      return;
    }

    // Allow the WebPart to run its own code prior to the render.
    this.onPreRender(this._activeLayoutInfo.id, this._activeDataProviderInfo.id, this._response)
      .then((renderOptions: IBaseCollectionWebPartLayoutState) => {
        // update configuration per any changes returned in the onPreRender method.
        if (renderOptions) {
          for (const name in renderOptions) {
            if (renderOptions.hasOwnProperty(name)) {
              this._options[name] = renderOptions[name];
            }
          }
        }

        // forward properties to the layout
        const viewProps: IBaseCollectionWebPartViewProperties = lodash.extend(
            this._options, this.properties, this._activeLayoutInfo && this._activeLayoutInfo.properties);

        viewProps.hasInitialLoadingState = this._activeLayoutInfo.hasInitialLoadingState;
        viewProps.displayMode = this.displayMode;
        viewProps.errorCallback = this._renderError;
        viewProps.renderedOnce = this.renderedOnce;
        viewProps.defaultTitle = this.defaultTitle;

        viewProps.layoutInfo = this._activeLayoutInfo;
        viewProps.width = this.width;

        if (this._activeLayoutInfo.displayFieldMaps) {
          viewProps.displayFieldMap = this._activeLayoutInfo.displayFieldMaps[this._activeDataProviderInfo.id];
        }

        viewProps.items = undefined;

        // Check if our layout has display field maps and we have a data provider.
        if (this._activeLayoutInfo.displayFieldMaps && this._activeDataProviderInfo) {
          const map: DisplayFieldMap = this._activeLayoutInfo.displayFieldMaps[this._activeDataProviderInfo.id];

          // If we have a map then apply it to the data provider response to generate our items.
          if (map) {
            // Some layouts (such as list) need access to the display field map to determine which elements
            // (e.g., columns) to display.
            viewProps.displayFieldMap = map;

            viewProps.items = map.transpose(this._response, (row: IDataRow): IDataRow => {
              return this.transposeRow(this._activeLayoutInfo.id, this._activeDataProviderInfo.id, row);
            });
          }
        }

        // By default with no map and no provided transpose function, the items become the response.
        if (!viewProps.items) {
          // Provide an opportunity for a transpose the data even if a map was not used.
          viewProps.items = this._response.map((row: any): IDataRow => { // tslint:disable-line:no-any
            return this.transposeRow(this._activeLayoutInfo.id, this._activeDataProviderInfo.id, row);
          });
        }

        try {
          // any is required to avoid having to expose IBaseCollectionWebPartOptions to IBaseCollectionWebPartProperties
          const element: React.ReactElement<IBaseCollectionWebPartViewProperties> =
            React.createElement(BaseCollection as any, viewProps); // tslint:disable-line:no-any
          this._component =
            <React.Component<IBaseCollectionWebPartViewProperties, IBaseCollectionWebPartViewState>>
            ReactDOM.render(element, this.domElement);

          if (!this.renderedOnce && this._component) {
            (this._component as BaseCollection).focusOnTitle();
          }

          // Allow the WebPart to run its own code after the render.
          this.onPostRender(this._activeLayoutInfo.id, this._activeDataProviderInfo.id, this._response)
            .then(() => {
              qosMonitor.writeSuccess();
            })
            .catch((error: Error) => {
              qosMonitor.writeUnexpectedFailure('FailedPostRender', error);
            });
        } catch (error) {
          qosMonitor.writeUnexpectedFailure('FailedRender', error);
        }
      })
      .catch((error: Error) => {
        this.renderError(error);
        qosMonitor.writeUnexpectedFailure('FailedPreRender', error);
        return;
      });
  }

  /**
   * Override this function to receive notification when render is about to start. Use this method to interrogate
   * the response and adjust the WebPart state if necessary prior to rendering. An example use would be to change
   * the current layout based upon data found (or not found) in the data response.
   *
   * Use the return value to update BaseCollectionWebPart configuration prior to rendering. Any configuration values
   * provided are appended to the current BaseCollectionWebPart configuration. Use this to set configuration dynamically
   * for the feature options in IBaseCollectionWebPartLayoutState which include whether to render a title, +Add
   * button, See all link, pagination control or message view.
   *
   * @param layoutId - Indicates the active layout.
   * @param dataProviderId - Indicates the active data provider.
   * @param response - The data rows as received from the data source.
   *
   * @returns A promise is returned that may result in success or error. A success result can include a hash object of
   * type IBaseCollectionWebPartLayoutState, with values to be appended to the current set of options.
   */
  protected onPreRender(
    layoutId: string, dataProviderId: string, response: IDataRow[]): Promise<IBaseCollectionWebPartLayoutState> {

    return Promise.resolve(); // return success with no configuration changes
  }

  /*
   * Override this function to receive notification when render has completed.  The intent is to provide a facility
   * for implementing followup data requests and updates to the rendered display.  The current layoutId and
   * dataProviderId are passed in case the WebPart supports alternative layouts/data providers and follow up steps
   * after using one might be different than after using another.
   *
   * @param layoutId - Indicates the active layout.
   * @param dataProviderId - Indicates the active data provider.
   * @param response - The data rows as received from the data source.
   *
   * @returns A promise is returned that may result in success or error.
   */
  protected onPostRender(layoutId: string, dataProviderId: string, response: IDataRow[]): Promise<{}> {
    return Promise.resolve(); // return success
    // use Promise.reject(error) if an error is to be returned.
  }

  /**
   * Serialize edits to the WebPart title per the BaseClientSideWebPart's onBeforeSerialize function.  If you override
   * then call super.onBeforeSerialize() in order to allow BaseCollectionWebPart to serialize the WebPart title.
   */
  protected onBeforeSerialize(): void {
    // Serialize changes entered in the title input box.
    if (this._component) {
      let title: string = (this._component as BaseCollection).title;
      if (title) {
        // remove surrounding white space
        title = title.trim();
      }
      this.properties.title = title;
    }
  }

  /**
   * Return the WebPart properties metadata per the BaseClientSideWebPart's propertiesMetadata function.  If you
   * override then call super.propertiesMetadata() to get the initial set of properties before adding your own.
   *
   * @returns The WebPart properties metadata.
   */
  protected get propertiesMetadata(): IWebPartPropertiesMetadata {
    return {
      'title': {isSearchablePlainText: true},
      'baseUrl': {isLink: true}
    };
  }

  /**
   * Called when the property pane is closed. Invokes a data request and then renders the result.
   */
  protected onDiscard(): void {
    this.refreshData();
  }

  /**
   * Override to set property default values where no values are present.
   * The default implementation does nothing.
   *
   * @param properties - The WebPart properties object for modification.
   */
  protected setDefaultProperties(properties: IBaseCollectionWebPartProperties): void {
    // Override to initialize the WebPart properties object with default values where none have been provided.
    return;
  }

  /**
   * Override to provide your WebPart with a default generated title. The user may type over the generated title
   * to specifically title the instance of the WebPart.  The default implementation returns placeholder
   * text for the title.
   *
   * @returns The title as a string.
   */
  protected get defaultTitle(): string {
    return strings.TitlePlaceholder;
  }

  /**
   * The set of data items to return in place of empty results from data provider.
   * @param layoutId - Indicates the active layout.
   * @param dataProviderId - Indicates the active data provider.
   * @param response - The data rows as received from the data source.
   *
   * @returns Set of data to be displayed instead of empty set.
   *
   * @example Override to specify data fo user to understand the layout of webpart instead of plain "no results"
   * experience.
   */
  protected getEmptyStateRows(layoutId: string, dataProviderId: string): IDataRow[] {
    return undefined;
  }

  /**
   *
   * @param response - The data rows to be rendered. This wil overrwrite the current response.
   */
  protected setResponseAndRender(response: IDataRow[]): void {
    this._response = response;
    this.render();
  }

  private _setPageContextVariables(): void {
    if (this._options.requirePageContextVariables) {
      // validate expected page context values are present
      if (!this.pageContext) {
        throw new Error(strings.ErrorMissingPageContext);
      }

      if (!this.pageContext.web) {
        throw new Error(strings.ErrorMissingWeb);
      }

      // for test pages that aren't localhost and all other hostTypes
      if (this.pageContext.web.absoluteUrl) {
        // if test page
        if (UNIT_TEST || DEBUG && Environment.type === EnvironmentType.Local) {
          if (this.pageContext.web.absoluteUrl.indexOf('https://microsoft.sharepoint.com') === 0) {
            // MSIT
            this.properties.webId = 'c7ecb072-9d0b-4b2c-9f6e-75337cb6c06a';
            // @todo: set test siteId for MSIT
          } else {
            // eDog
            this.properties.webId = 'a50ff2b3-f789-4251-aad1-5909f23aed4f';
            // @todo: set test siteId for eDog
          }
        // not test pages and not localhost
        } else {
          if (!this.pageContext.web.id) {
            throw new Error(strings.ErrorMissingWebId);
          }
          this.properties.webId = this.pageContext.web.id.toString();

          if (!this.pageContext.site) {
            throw new Error(strings.ErrorMissingSite);
          }
          if (!this.pageContext.site.id) {
            throw new Error(strings.ErrorMissingSiteId);
          }
          this.properties.siteId = this.pageContext.site.id.toString();
        }

        this.properties.baseUrl = this.pageContext.web.absoluteUrl;
      } else {
        // localhost, mimic eDog
        this.properties.webId = 'a50ff2b3-f789-4251-aad1-5909f23aed4f';
        // @todo: set test siteId for eDog
        this.properties.baseUrl = 'https://msft.spoppe.com/sites/wex';
      }
    }
  }

  private _setSelectedLayout(): void {
    // Default to the first layout if none are selected.
    if (!this.properties.layoutId && this._layoutInfos.length > 0) {
      this.properties.layoutId = this._layoutInfos[0].id;
    }

    if (this.properties.layoutId) {
      for (const layoutInfo of this._layoutInfos) {
        if (layoutInfo.id === this.properties.layoutId) {
          this._activeLayoutInfo = layoutInfo;
          break;
        }
      }
    }
  }

  private _setSelectedDataProvider(): void {
    // Default to the first data provider if none are selected.
    if (!this.properties.dataProviderId && this.dataProviders.length > 0) {
      this.properties.dataProviderId = this.dataProviders[0].id;
    }

    if (this.properties.dataProviderId) {
      for (const dataProviderInfo of this.dataProviders) {
        if (dataProviderInfo.id === this.properties.dataProviderId) {
          this._activeDataProviderInfo = dataProviderInfo;
          break;
        }
      }
    }
  }

  private _getDisplayFieldMaps(): void {
    this._layoutInfos.map((layout: ILayoutInfo) => {
      if (!layout.displayFieldMaps) {
        layout.displayFieldMaps = {};
      }
      this.dataProviders.map((dataProvider: IDataProviderInfo) => {
        layout.displayFieldMaps[dataProvider.id] = this.getDisplayFieldMap(layout.id, dataProvider.id);
      });
    });
  }

  // @todo: We can make a type alias on "any" and use it in this file. That would allow readers to understand the
  // typing and avoid multiple use of tslint-disable.
  private _requestData(request?: any): Promise<IDataRow[]> { // tslint:disable-line:no-any
    return (
      new Promise<IDataRow[]> ((complete: (response: IDataRow[]) => void, reject: (error: Error) => void) => {
        if (this._activeDataProviderInfo) {
          const qosMonitor: _QosMonitor = new _QosMonitor(`${this._options.qosPrefix}RequestData`);
          this.dataProvider
          .then((dataProvider: IDataProvider<undefined, undefined>) => {
            // Default to the optionally provided request via the refreshData or requestMoreData methods,
            // otherwise invoke the WebPart's getRequest method.

            request = request || this.getRequest(this._activeDataProviderInfo.id, this._activeLayoutInfo.id);
            dataProvider.requestData(request).then(
              (response: IDataRow[]): void => {
                this._usingEmptyStateRows = false;
                if (_SPFlight.isEnabled(932/*WexSitePagePublishing*/) && // empty states paired with comm sites for now
                    (!response || response.length === 0)) {
                  this._usingEmptyStateRows = true;
                  const empty: IDataRow[] = this.getEmptyStateRows(
                    this._activeLayoutInfo.id, dataProvider.id);
                  if (empty) {
                    response = empty;
                  }
                }

                complete(response);
                qosMonitor.writeSuccess();
              },
              (error: Error): void => {
                this.context.statusRenderer.clearLoadingIndicator(this.domElement);
                qosMonitor.writeUnexpectedFailure('FailedDataRequest', error);
                reject(error);
              });
          },
          (error: Error): void => {
            qosMonitor.writeUnexpectedFailure('FailedToLoadDataProvider', error);
            reject(error);
          });
        } else {
          // The derived WebPart class has not provided a data provider.
          reject(new Error(strings.ErrorNoDataProvider));
        }
    }));
  }

  // @todo: load shared data providers from a script in a central location.
  // private _loadDataProvidersFromUrls(urls: string[]): Promise<IDataProviderInfo[]> {
  //   ...
  // }

  // @todo: load shared layouts from a script in a central location.
  // private _loadLayoutsFromUrls(urls: string[]): Promise<ILayoutInfo[]> {
  //   ...
  // }
}