import * as React from 'react';
import { css, BaseComponent, IBaseProps } from 'office-ui-fabric-react';
import TextareaAutosize from 'react-textarea-autosize';

import styles from './CaptionElement.module.scss';
import strings from './loc/Strings.resx';

export enum CaptionElementType {
  Caption = 1,
  Title = 2
}

export enum CaptionElementAlignPosition {
  PageDirection = 1,
  Center = 2
}

export interface ICaptionElementProps extends IBaseProps {
  /**
   * Whether we are in edit mode or read mode.
   */
  isEditMode: boolean;

  /**
   * The text that this component contains.
   * @todo: Better to rename to 'text' or similar.
   */
  captionText: string;

  /**
   * Text to be shown when a text value is not set. This provides instructions to the user
   * on what text is expected.
   */
  placeholder: string;

  /**
   * Specifies if this editable text is being used to render a
   * Caption (e.g. for an image) or for a web part's title text.
   *
   * @default CaptionElementType.Caption
   */
  type?: CaptionElementType;

  /**
   * Custom class name to apply to the component.
   */
  className?: string;

  /**
   * Specify the text should be one of page direction aligned or center aligned with RTL considered.
   * Page direction aligned is left aligned adjusted to right aligned for RTL.
   *
   * There is a bug on Edge browser when input or center aligned or right aligned with a non-empty placeholder,
   * the cursor will be incorrectly left aligned.
   * {@link https://developer.microsoft.com/en-us/microsoft-edge/platform/issues/4468563/}
   *
   * @default CaptionElementAlignPosition.PageDirection
   */
  alignPosition?: CaptionElementAlignPosition;

  /**
   * Callback function that is called whenever the text value changes.
   */
  onCaptionTextChange?: (ev: React.FormEvent<HTMLTextAreaElement>) => void;
}

/**
 * Caption for various content webparts.
 * NOTE: Only display if the content has been selected!
 *
 * @todo: Consider making this component generic, something like 'EditableTextSpan'. Rather than
 *        building in text styles specifically for captions and web part titles, those styles could be
 *        set by the parent using the className prop. Then this could be used in more places.
*/
export default class CaptionElement extends BaseComponent<ICaptionElementProps, {}> {
  public static defaultProps: ICaptionElementProps = {
    isEditMode: false,
    captionText: undefined,
    placeholder: undefined,
    type: CaptionElementType.Caption,
    alignPosition: CaptionElementAlignPosition.Center
  };

  private _textarea: HTMLTextAreaElement;
  private _span: HTMLSpanElement;

  constructor(props: ICaptionElementProps) {
    super(props);
  }

  public render(): JSX.Element {
    return (
      <div className={
        css(
          this.props.className,
          { [styles.captionElement__caption]: this.props.type === CaptionElementType.Caption },
          { [styles.captionElement__title]: this.props.type === CaptionElementType.Title },
          { [styles.captionElement__leftAlign]:
            this.props.alignPosition === CaptionElementAlignPosition.PageDirection },
          { [styles.captionElement__centerAlign]: this.props.alignPosition === CaptionElementAlignPosition.Center }
        )
      }>
        { this.props.isEditMode ? this._renderEditMode : this._renderViewMode }
      </div>
    );
  }

  public value(): string {
    return this.props.isEditMode ?
      (this._textarea ? this._textarea.value : '') :
      (this._span ? this._span.textContent : '');
  }

  public focusOnCaptionElement(): void {
    if (this._textarea) {
      this._textarea.focus();
    }
  }

  /**
   * Render an editable text area.
   */
  private get _renderEditMode(): JSX.Element {
    return (
      <TextareaAutosize
        className={ styles.textarea }
        onChange={ this.props.onCaptionTextChange }
        ref={ this._resolveRef('_textarea') }
        placeholder={ this.props.placeholder }
        defaultValue={ this.props.captionText }
        aria-label={
          (this.props.type && this.props.type === CaptionElementType.Title) ?
          strings.TitlePlaceholder :
          strings.CaptionPlaceholder
        }
      />
    );
  }

  /**
   * Render a read-only text span.
   */
  private get _renderViewMode(): JSX.Element {
    return (
      <span ref={ this._resolveRef('_span') }>{ this.props.captionText }</span>
    );
  }
}