/// <reference types="mocha" />

declare var sinon: sinon.SinonStatic;
import * as React from 'react';
import { expect } from 'chai';
import { mount, ReactWrapper, shallow, ShallowWrapper } from 'enzyme';

import {
  default as CaptionElement,
  CaptionElementAlignPosition,
  CaptionElementType,
  ICaptionElementProps
 } from '../CaptionElement';

import styles from '../CaptionElement.module.scss';

/* tslint:disable:no-unused-expression */

describe('CaptionElement tests', () => {
  let props: ICaptionElementProps;

  beforeEach(() => {
    props = {
      isEditMode: false,
      captionText: '',
      placeholder: '',
      type: CaptionElementType.Caption,
      className: undefined,
      alignPosition: CaptionElementAlignPosition.Center,
      onCaptionTextChange: undefined
    };
  });

  describe('Render a caption or title in read mode', () => {
    let caption: ShallowWrapper<ICaptionElementProps, void>;
    beforeEach(() => {
      props.isEditMode = false;
    });

    afterEach(() => {
      caption.unmount();
    });

    describe('given caption text', () => {
      beforeEach(() => {
        props.captionText = 'Caption Text';
        caption = shallow<ICaptionElementProps, void>(<CaptionElement { ...props } />);
      });

      it('should render exactly one span element', () => {
        expect(caption.find('span').length).to.be.equal(1);
      });

      it('should render a span element with the correct text', () => {
        expect(caption.find('span').text()).to.be.equal(props.captionText);
      });
    });

    it('should render an empty caption', () => {
      props.captionText = '';
      props.placeholder = 'Enter your caption';
      caption = shallow<ICaptionElementProps, void>(<CaptionElement { ...props } />);
      expect(caption.find('span').text()).to.be.equal('');
    });

    describe('given a centered align position', () => {
      beforeEach(() => {
        props.captionText = 'Caption Text';
        caption = shallow<ICaptionElementProps, void>(<CaptionElement { ...props } />);
      });

      it('should render the correct text', () => {
        expect(caption.find('span').text()).to.be.equal(props.captionText);
      });

      it('should include the centerAlign class', () => {
        expect(caption.hasClass(styles.captionElement__centerAlign)).to.be.true;
      });

      it('should not include the leftAlign class', () => {
        expect(caption.hasClass(styles.captionElement__leftAlign)).to.be.false;
      });
    });

    describe('given a directional flow caption', () => {
      beforeEach(() => {
        props.captionText = 'Page Direction Caption';
        props.alignPosition = CaptionElementAlignPosition.PageDirection;
        caption = shallow<ICaptionElementProps, void>(<CaptionElement { ...props } />);
      });

      it('should render the correct text', () => {
        expect(caption.find('span').text()).to.be.equal(props.captionText);
      });

      it('should include the leftAlign class', () => {
        expect(caption.hasClass(styles.captionElement__leftAlign)).to.be.true;
      });

      it('should not include the centerAlign class', () => {
        expect(caption.hasClass(styles.captionElement__centerAlign)).to.be.false;
      });
    });

    describe('given an undefined align position', () => {
      beforeEach(() => {
        props.captionText = 'Default Caption';
        props.alignPosition = undefined;
        caption = shallow<ICaptionElementProps, void>(<CaptionElement { ...props } />);
      });

      it('should render the correct text', () => {
        expect(caption.find('span').text()).to.be.equal(props.captionText);
      });

      it('should default to centered caption', () => {
        expect(caption.hasClass(styles.captionElement__centerAlign)).to.be.true;
      });
    });

    describe('given a caption element type', () => {
      beforeEach(() => {
        props.captionText = 'Caption Text';
        props.type = CaptionElementType.Caption;
        caption = shallow<ICaptionElementProps, void>(<CaptionElement { ...props } />);
      });

      it('should render the correct text', () => {
        expect(caption.find('span').text()).to.be.equal(props.captionText);
      });

      it('should include a caption class', () => {
        expect(caption.hasClass(styles.captionElement__caption)).to.be.true;
      });

      it('should not include a title class', () => {
        expect(caption.hasClass(styles.captionElement__title)).to.be.false;
      });
    });

    describe('given a title element type', () => {
      beforeEach(() => {
        props.captionText = 'Title Text';
        props.type = CaptionElementType.Title;
        caption = shallow<ICaptionElementProps, void>(<CaptionElement { ...props } />);
      });

      it('should render the correct text', () => {
        expect(caption.find('span').text()).to.be.equal(props.captionText);
      });

      it('should include a title class', () => {
        expect(caption.hasClass(styles.captionElement__title)).to.be.true;
      });

      it('should not include a caption class', () => {
        expect(caption.hasClass(styles.captionElement__caption)).to.be.false;
      });
    });

    describe('given a caption with custom formatting', () => {
      beforeEach(() => {
        props.captionText = 'Custom Formatted Text';
        props.className = 'Custom';
        caption = shallow<ICaptionElementProps, void>(<CaptionElement { ...props } />);
      });

      it('should render the correct text', () => {
        expect(caption.find('span').text()).to.be.equal(props.captionText);
      });

      it('should include the custom format class', () => {
        expect(caption.hasClass(props.className)).to.be.true;
      });
    });
  });

  describe('Render a caption or title in edit mode', () => {
    let caption: ReactWrapper<ICaptionElementProps, void>;

    beforeEach(() => {
      props.isEditMode = true;
    });

    afterEach(() => {
      caption.unmount();
    });

    describe('given caption text', () => {
      beforeEach(() => {
        props.captionText = 'Caption Text';
        caption = mount<ICaptionElementProps, void>(<CaptionElement { ...props } />);
      });

      it('should render exactly one textarea element', () => {
        expect(caption.find('textarea').length).to.be.equal(1);
      });

      it('should render a textarea element with the correct text', () => {
        expect(caption.find('textarea').text()).to.be.equal(props.captionText);
      });
    });

    it('should render empty when no placeholder or entry text is present', () => {
      props.captionText = undefined;
      props.placeholder = undefined;
      caption = mount<ICaptionElementProps, void>(<CaptionElement { ...props } />);
      expect(caption.find('textarea').text()).to.be.equal('');
    });

    it('should render placeholder text when no entry is present', () => {
      props.captionText = undefined;
      props.placeholder = 'Enter your caption';
      caption = mount<ICaptionElementProps, void>(<CaptionElement { ...props } />);
      expect(caption.find(`textarea[placeholder="${props.placeholder}"]`).length).to.be.equal(1);
    });

    it('should render placeholder text when the entry is an empty string', () => {
      props.captionText = '';
      props.placeholder = 'Enter your caption';
      caption = mount<ICaptionElementProps, void>(<CaptionElement { ...props } />);
      expect(caption.find(`textarea[placeholder="${props.placeholder}"]`).length).to.be.equal(1);
    });

    describe('given a centered align position', () => {
      beforeEach(() => {
        props.captionText = 'Centered Caption';
        props.alignPosition = CaptionElementAlignPosition.Center;
        caption = mount<ICaptionElementProps, void>(<CaptionElement { ...props } />);
      });

      it('should render the correct text', () => {
        expect(caption.hasClass(styles.captionElement__caption)).to.be.true;
      });

      it('should include the centerAlign class', () => {
        expect(caption.hasClass(styles.captionElement__centerAlign)).to.be.true;
      });

      it('should not include the leftAlign class', () => {
        expect(caption.hasClass(styles.captionElement__leftAlign)).to.be.false;
      });
    });

    describe('given a page directional flow align position', () => {
      beforeEach(() => {
        props.captionText = 'Page Direction Caption';
        props.alignPosition = CaptionElementAlignPosition.PageDirection;
        caption = mount<ICaptionElementProps, void>(<CaptionElement { ...props } />);
      });

      it('should render the correct text', () => {
        expect(caption.hasClass(styles.captionElement__caption)).to.be.true;
      });

      it('should include the leftAlign class', () => {
        expect(caption.hasClass(styles.captionElement__leftAlign)).to.be.true;
      });

      it('should not include the centerAlign class', () => {
        expect(caption.hasClass(styles.captionElement__centerAlign)).to.be.false;
      });
    });

    describe('given an undefined align position', () => {
      beforeEach(() => {
        props.captionText = 'Default Caption';
        props.alignPosition = undefined;
        caption = mount<ICaptionElementProps, void>(<CaptionElement { ...props } />);
      });

      it('should render the correct text', () => {
        expect(caption.find('textarea').text()).to.be.equal(props.captionText);
      });

      it('should default to centered caption', () => {
        expect(caption.hasClass(styles.captionElement__centerAlign)).to.be.true;
      });
    });

    describe('given a caption element type', () => {
      beforeEach(() => {
        props.captionText = 'Caption Text';
        props.type = CaptionElementType.Caption;
        caption = mount<ICaptionElementProps, void>(<CaptionElement { ...props } />);
      });

      it('should render the correct text', () => {
        expect(caption.find('textarea').text()).to.be.equal(props.captionText);
      });

      it('should include a caption class', () => {
        expect(caption.hasClass(styles.captionElement__caption)).to.be.true;
      });

      it('should not include a title class', () => {
        expect(caption.hasClass(styles.captionElement__title)).to.be.false;
      });
    });

    describe('given a title element type', () => {
      beforeEach(() => {
        props.captionText = 'Title Text';
        props.type = CaptionElementType.Title;
        caption = mount<ICaptionElementProps, void>(<CaptionElement { ...props } />);
      });

      it('should render the correct text', () => {
        expect(caption.find('textarea').text()).to.be.equal(props.captionText);
      });

      it('should include a title class', () => {
        expect(caption.hasClass(styles.captionElement__title)).to.be.true;
      });

      it('should not include a caption class', () => {
        expect(caption.hasClass(styles.captionElement__caption)).to.be.false;
      });
    });

    describe('given a caption with custom formatting', () => {
      beforeEach(() => {
        props.captionText = 'Custom Formatted Text';
        props.className = 'Custom';
        caption = mount<ICaptionElementProps, void>(<CaptionElement { ...props } />);
      });

      it('should render the correct text', () => {
        expect(caption.find('textarea').text()).to.be.equal(props.captionText);
      });

      it('should include the custom format class', () => {
        expect(caption.hasClass(props.className)).to.be.true;
      });
    });
  });

  describe('CaptionElement event model', () => {
    let caption: ReactWrapper<ICaptionElementProps, void>;
    beforeEach(() => {
      props.isEditMode = true;
    });

    afterEach(() => {
      caption.unmount();
    });

    it('should invoke the onCaptionTextChange method when receiving data entry', () => {
      const spy: sinon.SinonSpy = sinon.spy();
      props.onCaptionTextChange = spy;
      caption = mount<ICaptionElementProps, void>(<CaptionElement { ...props } />);
      caption.find('textarea').simulate('change', { target: { value: 'New Caption' } });
      expect(spy.calledOnce).to.be.true;
    });

    // You can use spy.args[0][0] to inspect the input to onCaptionTextChange.
    // @todo: Would like to add a test here that validates the onCaptionTextChange event
    // receives the correct entry. Not sure that using { target: { value: 'New Caption' } }
    // is actually the correct input format to .simulate(). The object received by
    // onCaptionTextChange ends up looking like { 1: 'N', 2: 'e', 3: 'w'... and so on.
  });
});