/**
 * @Copyright (c) Microsoft Corporation.  All rights reserved.
 */

import { autobind } from 'office-ui-fabric-react';
import { BaseClientSideWebPart } from '@microsoft/sp-webpart-base';
import WebPartDataProvider from './WebPartDataProvider';
import { IDataProviderConfiguration } from './IDataProvider';
import { IRequest, IResponse } from './IDataProvider';

/**
 * This is a prototype at the current point. The goal is to build some base code that will
 * help our customers follow the performance best practices to build web parts.
 *
 * @alpha
 */
export abstract class BaseDataWebPart<TProperties, TData> extends BaseClientSideWebPart<TProperties> {

  private _dataProvider: WebPartDataProvider;

  /*
   * @virtual
   */
  protected get dataProviderConfiguration(): IDataProviderConfiguration {
    return {
      /* tslint:disable-next-line:no-any */
      serviceScope: this.context.serviceScope as any
    };
  }

  /**
   * @sealed
   */
  protected get dataProvider(): WebPartDataProvider {
    if (!this._dataProvider) {
      this._dataProvider = new WebPartDataProvider(this.dataProviderConfiguration);
    }

    return this._dataProvider;
  }

  /**
   * Fetch the web part data.
   */
  protected abstract fetchData(): Promise<TData>;

  /**
   * Callback when date is fetched and updated. In case of a cache hit, this callback
   * is called twice. Once for data from the cache and next for data from the actual
   * source. In case of a cache miss, the callback happens only once with data from
   * the actual source only.
   */
  protected abstract dataReceived(response: IResponse<TData>): void;

  /**
   * The default value of the cacheKey is the web part instanceId.
   *
   * @virtual
   */
  protected get cacheKey(): string {
    return this.context.instanceId;
  }

  /**
   * The default value of the cacheKey is the web part instanceId.
   *
   * @virtual
   */
  protected get alias(): string {
    return this.context.manifest.alias;
  }

  /**
   * The cache parameters. By default its the properties of the web part.
   *
   * @virtual
   */
  protected get cacheParameters(): { [key: string]: Object } {
    /* tslint:disable-next-line:no-any */
    return this.properties as any as { [key: string]: Object };
  }

  /**
   * @private
   */
  protected _internalGetData(): Promise<void> {

    const request: IRequest<TData> = {
      fetchData: this.fetchData,
      dataReceived: this._internalDataReceived,
      cacheKey: this.cacheKey,
      alias: this.alias,
      cacheParameters: this.cacheParameters
    };

    return this.dataProvider.requestData(request).then((data: TData) => {
      request.dataReceived({
        data: data,
        metadata: { }
      });
    });
  }

  /**
   * Data updated callback
   * @todo: data should be typed as IResponse<TData>. (VSO#399477) tracks fixing that in the MySiteWebPartCacheManager.
   */
  @autobind
  /* tslint:disable-next-line:no-any */
  private _internalDataReceived(data: any): void {
    this.dataReceived({
        data: data,
        metadata: { }
      });
  }
}