/**
 * @Copyright (c) Microsoft Corporation.  All rights reserved.
 */

import { ServiceScope } from '@microsoft/sp-core-library';

/**
 * Data provider interface
 *
 * @alpha
 */
export interface IDataProvider {

  /**
   * Data provider configuration.
   */
  readonly configuration: IDataProviderConfiguration;

  /**
   * API to request data. If the data is found in the cache, the promise is
   * resolved immediately with the data. In the case of a cache miss, promise is resolved
   * after the data is fetched from the remote service.
   **/
  requestData<TData>(request: IRequest<TData>): Promise<TData>;
}

/**
 * Data provider configuration. This configuration is required during the initialization
 * of the DataProvider.
 *
 * @alpha
 */
export interface IDataProviderConfiguration {
  serviceScope: ServiceScope;
}

/**
 * Request parameters.
 *
 * @alpha
 */
export interface IRequest<TData> {

  /**
   * Delegate that helps in fetching data for the consuming component. Since the consuming component
   * could fetch data in multiple ways from multiple sources.
   */
  fetchData: () => Promise<TData>;

  /**
   * Callback when date is fetched and updated. In case of a cache hit, this callback
   * is called twice. Once for data from the cache and next for data from the actual
   * source. In case of a cache miss, the callback happens only once with data from
   * the actual source only.
   */
  dataReceived: (response: IResponse<TData>) => void;

  /**
   * The key to be used for the cache. e.g. a consuming components instance id. It is
   * upto the discretion of the consuming component to use a different key.
   */
  cacheKey: string;

  /**
   * The string identifying the component using the cache for logging purposes.
   */
  alias: string;

  /**
   * A hash of this JSON object is used to extract data from the cache. The purpose
   * of this parameter is to make sure invalid results are not displayed to the
   * user. For example, if the query parameters for the consuming component are changed as
   * compared to when the cache was generated, displaying the cached results would
   * be incorrect.
   */
  cacheParameters: { [key: string]: Object };
}

/**
 * Response data from the DataProvider.
 */
export interface IResponse<TData> {
  data: TData;
  metadata: IResponseMetadata;
}

/**
 * Metadata associated with the response. This property shall
 * be used in the future.
 *
 * @alpha
 */
export interface IResponseMetadata {
}