/**
 * @file DateBox.test.tsx
 * @copyright (c) Microsoft Corporation.  All rights reserved.
 */
import * as React from 'react';
import { expect } from 'chai';
import { ShallowWrapper, shallow } from 'enzyme';
import DateBox, { DateBoxSize, IDateBoxProps } from './DateBox';

/* tslint:disable:no-unused-expression */

describe('Date Box', () => {
  const props: IDateBoxProps = {
    startDate: new Date('2017-07-11T07:27:15.982Z'),
    endDate: new Date('2017-07-12T07:27:15.982Z'), // 24 hours later than start date
    size: DateBoxSize.Medium
  };

  it('should renders multiple-day box when start and end date are not the same day', () => {
    const dateBox: ShallowWrapper<IDateBoxProps, {}> = shallow(<DateBox { ...props } />);

    expect(dateBox.find('[data-automation-id="singleDayBox"]')).to.have.lengthOf(0);
    expect(dateBox.find('[data-automation-id="multipleDayBox"]')).to.have.lengthOf(1);
  });

  it('should renders single-day box when start and end date are the same day', () => {
    const dateBox: ShallowWrapper<IDateBoxProps, {}> = shallow(
      <DateBox
        { ...props }
        endDate={ props.startDate }
      />
    );

    expect(dateBox.find('[data-automation-id="singleDayBox"]')).to.have.lengthOf(1);
    expect(dateBox.find('[data-automation-id="multipleDayBox"]')).to.have.lengthOf(0);
  });

  it('should renders single-day box when end date is undefined', () => {
    const dateBox: ShallowWrapper<IDateBoxProps, {}> = shallow(
      <DateBox
        { ...props }
        endDate={ undefined }
      />
    );

    expect(dateBox.find('[data-automation-id="singleDayBox"]')).to.have.lengthOf(1);
    expect(dateBox.find('[data-automation-id="multipleDayBox"]')).to.have.lengthOf(0);
  });

  it('should render empty state if both start and end date are undefined', () => {
    // Empty state is a single-day box, inside which the first row displays 'Month' in local language,
    // and the second row displays '01'.
    const dateBox: ShallowWrapper<IDateBoxProps, {}> = shallow(
      <DateBox
        startDate={ undefined }
        endDate={ undefined }
        size={ DateBoxSize.Medium }
      />
    );

    expect(dateBox.find('[data-automation-id="singleDayBox"]')).to.have.lengthOf(1);
    expect(dateBox.find('[data-automation-id="multipleDayBox"]')).to.have.lengthOf(0);
    expect(dateBox.find('[data-automation-id="singleDayMonthContainer"]').text()).to.not.be.empty;
    expect(dateBox.find('[data-automation-id="singleDayDayContainer"]').text()).to.equal('01');
  });

  it('should render single-day date string with formating function defined in props', () => {
    const dateBox: ShallowWrapper<IDateBoxProps, {}> = shallow(
      <DateBox
        { ...props }
        endDate={ props.startDate }
        formatMonth={ (date) => 'exampleMonthFormat' }
        formatDay={ (date) => 'exampleDayFormat' }
      />
    );

    expect(dateBox.find('[data-automation-id="singleDayBox"]')).to.have.lengthOf(1);
    expect(dateBox.find('[data-automation-id="multipleDayBox"]')).to.have.lengthOf(0);
    expect(dateBox.find('[data-automation-id="singleDayMonthContainer"]').text()).to.equal('exampleMonthFormat');
    expect(dateBox.find('[data-automation-id="singleDayDayContainer"]').text()).to.equal('exampleDayFormat');
  });

  it('should render multiple-day date string with formating function defined in props', () => {
    const dateBox: ShallowWrapper<IDateBoxProps, {}> = shallow(
      <DateBox
        { ...props }
        formatDate={ (date) => 'exampleFormat' }
      />
    );

    expect(dateBox.find('[data-automation-id="singleDayBox"]')).to.have.lengthOf(0);
    expect(dateBox.find('[data-automation-id="multipleDayBox"]')).to.have.lengthOf(1);
    expect(dateBox.find('[data-automation-id="multipleDayStartDateContainer"]').text()).to.equal('exampleFormat');
    expect(dateBox.find('[data-automation-id="multipleDayEndDateContainer"]').text()).to.equal('exampleFormat');
  });
});