import * as React from 'react';
import { css } from 'office-ui-fabric-react';
import * as lodash from '@microsoft/sp-lodash-subset';
import { LocaleFormat } from '@ms/i18n-utilities';

import { isSameDay } from '@ms/sp-component-utilities';
import styles from './DateBox.module.scss';
import strings from './DateBox.resx';

/**
 * This enums defines the size options of the date box.
 */
export enum DateBoxSize {
  Small = 0,
  Medium = 1,
  Large = 2
}

/**
 * This type is used for the mapping from box size to class names.
 */
type SizeClassMapType = { [key: number]: string };

/**
 * Mapping from the date box size to class names that indicates size.
 * The keys of the mapping should align with the DateBoxSize enum.
 * @todo #281147: Use the enum as the key after upgrade to TypeScript 2.1.3.
 */
const sizeClassNameMapping: SizeClassMapType = {
  0: styles.boxIsSmall,
  1: styles.boxIsMedium,
  2: styles.boxIsLarge
};

export interface IDateBoxProps {
  /**
   * The start date object.
   */
  startDate: Date;

  /**
   * The end date object.
   *
   * If the endDate is the same day as the startDate, DateBox will show single day view of the startDate.
   * If the endDate or startDate is undefined, DateBox will show single day view of the defined date.
   * If the endDate and startDate are both undefined, DateBox will show the empty state view.
   */
  endDate: Date;

  /**
   * Indicates the DateBox size, which depends on the user scenario (in page title, in 'see all page', etc.).
   */
  size: DateBoxSize;

  /**
   * Customized style for DateBox.
   */
  className?: string;

  /**
   * Optional method to format the date to a date string, e.g. 'Nov 1', 'Jan 15'.
   */
  formatDate?: (date: Date) => string;

  /**
   * Optional method to format the month of the date to a month string, e.g. 'Nov', 'Jan'.
   */
  formatMonth?: (date: Date) => string;

  /**
   * Optional method to format the day of the date to a day string, e.g. '1', '15'.
   */
  formatDay?: (date: Date) => string;
}

/**
 * The date box component.
 */
export default class DateBox extends React.PureComponent<IDateBoxProps, {}> {
  public static defaultProps: Partial<IDateBoxProps> = {
    formatDate: (date: Date) => {
      return LocaleFormat.formatDate(date, { skeleton: 'MMMd' }).toUpperCase();
    },
    formatMonth: (date: Date) => {
      return LocaleFormat.formatDate(date, { skeleton: 'MMM' }).toUpperCase();
    },
    formatDay: (date: Date) => {
      return LocaleFormat.formatDate(date, { skeleton: 'd' });
    }
  };

  private _sizeClassName: string;

  public render(): JSX.Element {
    const { startDate, endDate } = this.props;
    this._sizeClassName = lodash.get<SizeClassMapType, string>(sizeClassNameMapping, this.props.size);

    if (!startDate || !endDate || isSameDay(startDate, endDate)) {
      return this._singleDayDateBoxElement;
    } else {
      return this._multipleDaysDateBoxElement;
    }
  }

  private get _singleDayDateBoxElement(): JSX.Element {
    const { startDate, endDate, formatMonth, formatDay } = this.props;
    const date: Date = startDate || endDate;

    const month: string = date ? formatMonth(date) : strings.DefaultMonthText;

    // @todo, May replace this '01' with a string generated from i18n-utilities
    const day: string = date ? formatDay(date) : '01';

    return (
      <div
        className={ css(styles.box, styles.boxIsSingleDay, this._sizeClassName, this.props.className) }
        data-automation-id='singleDayBox'
      >
        <div
          className={ styles.month }
          data-automation-id='singleDayMonthContainer'
        >
          { month }
        </div>
        <div
          className={ styles.day }
          data-automation-id='singleDayDayContainer'
        >
          { day }
        </div>
      </div>
    );
  }

  private get _multipleDaysDateBoxElement(): JSX.Element {
    const { startDate, endDate, formatDate } = this.props;

    return (
      <div
        className={ css(styles.box, styles.boxIsMultipleDays, this._sizeClassName, this.props.className) }
        data-automation-id='multipleDayBox'
      >
        <div
          className={ styles.date }
          data-automation-id='multipleDayStartDateContainer'
        >
          { formatDate(startDate) }
        </div>
        <hr className={ styles.separator } />
        <div
          className={ styles.date }
          data-automation-id='multipleDayEndDateContainer'
        >
          { formatDate(endDate) }
        </div>
      </div>
    );
  }
}
