/**
 * @file Functions to transpose data from one named structure to another that is compatible with a display template.
 * @copyright Microsoft Corporation. All rights reserved.
 */

import CalculatedField from './CalculatedField';
import { IDataRow } from '@ms/sp-component-utilities';
import MappedField from './MappedField';
export const LINK_FIELDNAME_SUFFIX: string = '_linkUrl';

export default class DisplayFieldMap {
  private _map: (MappedField | CalculatedField)[];

  public constructor(map: (MappedField | CalculatedField)[]) {
    this._map = map;
  }

  // Return an individual value from an object.
  public fieldValue(field: MappedField | CalculatedField, row: IDataRow): Object {
    if (field instanceof CalculatedField) {
      return field.formatter(row);
    } else {
      return this._firstIdentifiedValue(row, field.sources);
    }
  }

  // Return an individual linkUrl from an object.
  public fieldLinkUrl(field: MappedField | CalculatedField, row: IDataRow): Object {
    return this._firstIdentifiedValue(row, field.linkUrls);
  }

  // Return an object with its attributes re-mapped to different property names as identified in the
  // display field map.  Optionally pass in a preProcess function to manipulate the data prior to
  // applying the display map.
  public transposeRow(row: IDataRow, preProcess?: (row: IDataRow) => IDataRow): IDataRow {
    if (!row) {
      // marker for triggering onRenderMissingRow for infinite scroll
      return row;
    }
    const sourceRow: IDataRow = preProcess ? preProcess(row) : row;
    const result: IDataRow = {};

    if (this._map) {
      for (let i: number = 0; i < this._map.length; i++) {
        const field: MappedField | CalculatedField = this._map[i];
        result[field.target] = this.fieldValue(field, sourceRow);
        if (field.linkUrls) {
          result[field.target + LINK_FIELDNAME_SUFFIX] = this.fieldLinkUrl(field, sourceRow);
        }
      }
    }
    return result;
  }

  // Returns all objects re-mapped via the transposeRow function.
  public transpose(rows: IDataRow[], preProcess?: (row: IDataRow) => IDataRow): IDataRow[] {
    const result: IDataRow[] = [];
    for (let i: number = 0; i < rows.length; i++) {
      result[i] = this.transposeRow(rows[i], preProcess);
    }
    return result;
  }

  get map(): (MappedField | CalculatedField)[] {
    return this._map;
  }

  private _firstIdentifiedValue(row: IDataRow, sources: string[]): Object {
    for (let i: number = 0; i < sources.length; i++) {
      const fieldName: string = sources[i];
      const value: Object = row[fieldName];
      // Skip if the value is empty (undefined, null or an empty string)
      if (value !== undefined && value !== null && value !== '') { // tslint:disable-line:no-null-keyword
        return row[fieldName];
      }
    }

    return undefined; // defaults to undefined
  }
}