/// <reference types="mocha" />

import { expect } from 'chai';
import { IDataRow } from '@ms/sp-component-utilities';
import CalculatedField from '../CalculatedField';
import DisplayFieldMap from '../DisplayFieldMap';

import MappedField from '../MappedField';

interface IData {
  Name?: string;
  JobTitle?: string;
  Distributor?: string;
  Birthdate?: string;
  Released?: string;
  Budget?: number;
  Fictional?: boolean;
  Documentary?: boolean;
}

interface IDisplay {
  heading?: string;
  detail?: string;
  subnote?: string;
  footer?: string;
}

/* tslint:disable:no-unused-expression */

describe('DisplayFieldMap tests', () => {
  const rows: IData[] = [
    {
      Name: 'Buster Keaton',
      JobTitle: 'Director',
      Birthdate: '10/04/1895'
    },
    {
      Name: 'The General',
      Distributor: 'United Artists',
      Released: '12/31/1926',
      Budget: 750000,
      Fictional: true,
      Documentary: false
    },
    {
      Name: 'Charles Chaplin',
      JobTitle: '',
      Distributor: 'United Artists',
      Birthdate: '04/16/1889'
    }
  ];
  const displayFieldMap: DisplayFieldMap = new DisplayFieldMap([
    new MappedField('heading', { sources: ['Name'] }),
    new MappedField('detail', { sources: ['JobTitle', 'Distributor'] }),
    new MappedField('subnote', { sources: ['Birthdate', 'Released'] }),
    new MappedField('footer', { sources: ['Budget'] })
  ]);
  const displayFieldMap2: DisplayFieldMap = new DisplayFieldMap([
    new CalculatedField('heading', {}, (row: IData): Object => {
      return row.Name + ': ' + row.Distributor;
    })
  ]);
  const displayFieldMap3: DisplayFieldMap = new DisplayFieldMap([
    new MappedField('heading', { sources: ['Fictional', 'Documentary'] }),
    new MappedField('detail', { sources: ['Documentary', 'Fictional'] })
  ]);

  it('can lookup an individual field value', () => {
    let result: Object;
    result = displayFieldMap.fieldValue(displayFieldMap.map[1], rows[0] as IDataRow);
    expect(result).to.equal('Director');
    result = displayFieldMap.fieldValue(displayFieldMap.map[1], rows[1] as IDataRow);
    expect(result).to.equal('United Artists');
  });

  it('can transpose a row', () => {
    const result: IDisplay = displayFieldMap.transposeRow(rows[0] as IDataRow);
    expect(result.heading).to.equal('Buster Keaton');
    expect(result.detail).to.equal('Director');
    expect(result.subnote).to.equal('10/04/1895');
    expect(result.footer).to.be.undefined;
  });

  it('can transpose rows', () => {
    const result: IDisplay[] = displayFieldMap.transpose(rows as IDataRow[]);
    expect(result[0].heading).to.equal('Buster Keaton');
    expect(result[0].detail).to.equal('Director');
    expect(result[0].subnote).to.equal('10/04/1895');
    expect(result[0].footer).to.be.undefined;
    expect(result[1].heading).to.equal('The General');
    expect(result[1].detail).to.equal('United Artists');
    expect(result[1].subnote).to.equal('12/31/1926');
    expect(result[1].footer).to.equal(750000);
  });

  it('will treat empty string as an undefined value', () => {
    const result: Object = displayFieldMap.fieldValue(displayFieldMap.map[1], rows[2] as IDataRow);
    expect(result).to.equal('United Artists');
  });

  it('can support calculated fields', () => {
    const result: IDisplay = displayFieldMap2.transposeRow(rows[1] as IDataRow);
    expect(result.heading).to.equal('The General: United Artists');
  });

  it('can support boolean values', () => {
    const result: IDisplay = displayFieldMap3.transposeRow(rows[1] as IDataRow);
    expect(result.heading).to.be.true;
    expect(result.detail).to.be.false;
  });

  it('will treat undefined boolean as an undefined value', () => {
    const result: IDisplay = displayFieldMap3.transposeRow(rows[0] as IDataRow);
    expect(result.heading).to.be.undefined;
  });
});