/**
 * @file DocLibraryContext.ts
 * @Copyright (c) Microsoft Corporation.  All rights reserved.
 */
import {
  Environment,
  EnvironmentType,
  ServiceScope
} from '@microsoft/sp-core-library';

import { SPHttpClient } from '@microsoft/sp-http';
import { PageContext } from '@microsoft/sp-page-context';
import { Uri } from '@ms/odsp-utilities-bundle';

export interface ISiteContext {
  webAbsoluteUrl?: string;
  libraryName: string;
  defaultLibraryId: string;
  pickerDomain: string;
  siteId: string;
  webId: string;
}

export interface IGetListWithRootFolderResult {
  Id: string; // Guid
  Title: string;
  Description: string;
  BaseTemplate: number;
  ParentWebUrl: string;
  RootFolder: IListRootFolder;
}

export interface IListRootFolder {
  ServerRelativeUrl: string;
  Name: string;
}

export default class DocLibraryContext {
  private _errorCallback: () => void;
  private _spHttpClient: SPHttpClient;
  private _webAbsoluteUrl: string;
  private _siteContext: ISiteContext;
  private _siteId: string;
  private _webId: string;

  constructor(serviceScope: ServiceScope, errorCallback?: () => void) {
    let pageContext: PageContext;
    serviceScope.whenFinished(() => {
      this._spHttpClient = serviceScope.consume(SPHttpClient.serviceKey);
      pageContext = serviceScope.consume(PageContext.serviceKey);
    });

    if (pageContext && pageContext.web) {
      this._webAbsoluteUrl = pageContext.web.absoluteUrl;
      this._webId = pageContext.web.id && pageContext.web.id.toString() || '';
    } else {
      this._webAbsoluteUrl = this._webId = '';
    }
    this._siteId = pageContext && pageContext.site && pageContext.site.id && pageContext.site.id.toString() || '';
    this._errorCallback = errorCallback;
  }

  public get defaultDocumentLibrary(): Promise<IGetListWithRootFolderResult> {
    const url: string = Uri.concatenate(this._webAbsoluteUrl, '_api/web/DefaultDocumentLibrary?$expand=RootFolder');
    return this._spHttpClient.get(
      url,
      SPHttpClient.configurations.v1
    ).then((response: Response) => response.json());
  }

  public get siteContext(): Promise<ISiteContext> {
    if (this._siteContext) {
      return Promise.resolve(this._siteContext);
    } else {
      return this._getContainingFolderSiteContext().then((response: ISiteContext) => {
        this._siteContext = response;
        return response;
      });
    }
  }

  // [WEX! VSO: 240908] Instrument DocLibraryContext
  private _getContainingFolderSiteContext(): Promise<ISiteContext> {
    const domain: string = window.location.protocol + '//' + window.location.hostname;

    if (UNIT_TEST || DEBUG && Environment.type === EnvironmentType.Local) {
      this._webAbsoluteUrl = 'https://msft.spoppe.com/sites/wex';
      return Promise.resolve({
        webAbsoluteUrl: this._webAbsoluteUrl,
        libraryName: 'Shared Documents',
        defaultLibraryId: '{9BFFA1A1-0C18-4F0C-A7F2-7DB2564273D2}',
        pickerDomain: domain,
        siteId: '301e2bbb-a597-42d2-9495-9863d685b011',
        webId: 'a50ff2b3-f789-4251-aad1-5909f23aed4f'
      });
    } else {
      return new Promise<ISiteContext>(
        (complete: (contextInfo: ISiteContext) => void,
        reject: (error: Error) => void) => {
          this.defaultDocumentLibrary.then((list: IGetListWithRootFolderResult) => {
            complete({
              webAbsoluteUrl: this._webAbsoluteUrl,
              libraryName: list && list.RootFolder && list.RootFolder.Name || '',
              defaultLibraryId: list && list.Id || '',
              pickerDomain: domain,
              siteId: this._siteId,
              webId: this._webId
            });
          }).catch((error: Error) => {
            if (this._errorCallback) {
              this._errorCallback();
            }

            reject(error);
          });
        }
      );
    }
  }
}