/**
 * Lazy Loader internal event implementation.
 * @Copyright (c) Microsoft Corporation.  All rights reserved.
 *
 * @file LazyEventHandler.tsx
 */

import * as React from 'react';
import * as ReactDom from 'react-dom';
import { Async } from 'office-ui-fabric-react';
import { _EngagementLogger, _TraceLogger, _LogSource } from '@ms/sp-telemetry';

import { ILazyLoaderProps, ILazyLoaderState } from './LazyLoader';

type LazyComponent = React.Component<ILazyLoaderProps, ILazyLoaderState>;

/**
 * Event handler to support lazy loading. This class takes heavy load of managing
 * registered components to be rendered lazily. It listens to the scroll/resize events
 * and mark components rendered as they become closer to the viewport
 */
export default class LazyEventHandler {
  // Registered lazy components which needs to be rendered at right time
  private static _components: Set<LazyComponent> = new Set<LazyComponent>();

  // Attribute name to be used to track how many lazy components child to same
  // scrollable parent, to smartly unregister from scroll/resize events
  private static readonly _lazyLoadedTracker: string = 'sp-client-lazyload-data';

  // Utility object to debounce the event handler
  private static _async: Async;

  // This is debounced version of the Lazy handler.
  // We use this to ensure only one copy of the event handler
  private static _lazyHandler: () => void;

  private static _logSource: _LogSource = _LogSource.create('LazyLoader');
  private static _initTime: Date;
  private static _isListeningWindow: boolean = false;

  /**
   * Register a component which should be lazy loaded
   */
  public static registerComponent(component: React.Component<ILazyLoaderProps, ILazyLoaderState>): void {
    if (!component) {
      return;
    }

    if (!LazyEventHandler._lazyHandler) {
      LazyEventHandler._initialize();
    }

    LazyEventHandler._handleComponentVisibility(component);
    if (component.state.isRendered) {
      // Component is already in the view and loaded
      _EngagementLogger.logEvent('Webpart.LazyLoad.Scroll', 'StartLoaded');
      return;
    }

    const componentNode: Element = ReactDom.findDOMNode(component);
    if (!componentNode) {
      _TraceLogger.logError(
        LazyEventHandler._logSource,
        new Error(`LazyLoader - ReactDom.findDOMNode returned ${componentNode}`));
      LazyEventHandler._renderComponent(component);

      _EngagementLogger.logEvent('Webpart.LazyLoad.Scroll', 'StartLoaded');
      return;
    }

    const parent: Element = LazyEventHandler._findScrollableParent(componentNode);

    if (parent) {
      const count: number = LazyEventHandler._updateRegisteredChildren(parent, 1);
      if (count === 1) {
        LazyEventHandler._listenEvents(parent);
      }
    }

    LazyEventHandler._components.add(component);
    LazyEventHandler._listenEvents(window);
  }

  /**
   * Unregister a loaded component which should not be tracked any more
   */
  public static unregisterComponent(component: React.Component<ILazyLoaderProps, ILazyLoaderState>): void {
    if (!LazyEventHandler._components.has(component)) {
      return;
    }

    _EngagementLogger.logEvent('Webpart.LazyLoad.Scroll', 'LazyLoaded');

    LazyEventHandler._components.delete(component);
    const componentNode: Element = ReactDom.findDOMNode(component);
    if (componentNode) {
      const parent: Element = LazyEventHandler._findScrollableParent(componentNode);
      if (parent) {
        const count: number = LazyEventHandler._updateRegisteredChildren(parent, -1);
        if (count <= 0) {
          LazyEventHandler._unlistenEvents(parent);
        }
      }
    }

    if (LazyEventHandler._components.size === 0) {
      // Done loading all lazy components
      LazyEventHandler._unlistenEvents(window);
      LazyEventHandler._dispose();
    }
  }

  /**
   * Event handler which is called upon scroll and resize events
   */
  private static _lazyLoaderEventHandler(): void {
    if (!LazyEventHandler._components || LazyEventHandler._components.size === 0) {
      _TraceLogger.logError(
        LazyEventHandler._logSource,
        new Error('LazyLoader - _lazyLoaderEventHandler is called when no lazy components to load any more'));
      return;
    }

    LazyEventHandler._components.forEach((
      value: LazyComponent,
      index: LazyComponent,
      theSet: Set<LazyComponent>) => {
        LazyEventHandler._handleComponentVisibility(value);
      }
    );
  }

  /**
  * Handler to make component visible if it is close enough to the viewport
  */
  private static _handleComponentVisibility(component: LazyComponent): void {
    let renderComponent: boolean = false;
    try {
      const componentNode: Element = ReactDom.findDOMNode(component);
      if (!componentNode) {
        renderComponent = true;
        _TraceLogger.logVerbose(LazyEventHandler._logSource, 'ReactDom.findDOMNode returned null');
      } else {
        const parent: Element = LazyEventHandler._findScrollableParent(componentNode);
        const canRender: boolean =
          LazyEventHandler._isCloseToViewport(parent, componentNode, component.props.offset);

        if (canRender) {
          renderComponent = true;
        }
      }
    } catch (error) {
       _TraceLogger.logError(LazyEventHandler._logSource, error);
      renderComponent = true;
    }

    if (renderComponent) {
      LazyEventHandler._renderComponent(component);
      LazyEventHandler.unregisterComponent(component);
    }

    return;
  }

  /**
  * Check if component is close enough to the viewport to be shown
  */
  private static _isCloseToViewport(
    parent: Element, // tslint:disable-line:no-any
    componentNode: Element,
    offset: number
  ): boolean {
    try {
      const { top, height }: { top: number, height: number } =
        componentNode.getBoundingClientRect();
      const documentHeight: number =
        window.innerHeight || document.documentElement.clientHeight;

      if (!parent) {
        // There is no overflow parent, check distance from window
        const topInOrAboveViewport: boolean = (top - offset) <= documentHeight;
        const bottomNotAboveViewport: boolean = (top + height + offset) >= 0;

        return  topInOrAboveViewport && bottomNotAboveViewport;
      }

      // tslint:disable-next-line:no-use-before-declare
      const { top: parentTop, height: parentHeight }: { top: number, height: number }
        = parent.getBoundingClientRect();

      // Parent and viewport's intersection. 0 is top of the viewport
      const parentIntersectionTop: number = Math.max(0, parentTop);
      const parentIntersectionHeight: number =
        Math.min(parentTop + parentHeight, documentHeight) - parentIntersectionTop;
      // element's top relative to intersection
      const offsetTop: number = top - parentIntersectionTop;

      const parentTopInOrAboveViewport: boolean = (offsetTop - offset) <= parentIntersectionHeight;
      const parentBottomNotAboveViewport: boolean = (offsetTop + height + offset) >= 0;

      return parentTopInOrAboveViewport && parentBottomNotAboveViewport;
    } catch (error) {
      _TraceLogger.logError(LazyEventHandler._logSource, error);
      // We want to avoid lazy loading as a fallback
      return true;
    }
  }

  /**
   * Traverses up the DOM for the element with the data-is-scrollable=true attribute, or returns
   * window
   */
  private static _findScrollableParent(startingElement: Element): Element {
    const DATA_IS_SCROLLABLE_ATTRIBUTE: string = 'data-is-scrollable';
    let el: Element = startingElement;
    const body: HTMLElement = document.body;
    // First do a quick scan for the scrollable attribute.
    while (el && el !== body) {
      if (el.getAttribute(DATA_IS_SCROLLABLE_ATTRIBUTE) === 'true') {
        return el;
      }
      el = el.parentElement;
    }

    // If we haven't found it, the use the slower method: compute styles to evaluate if overflow is set.
    el = startingElement;

    while (el && el !== body) {
      if (el.getAttribute(DATA_IS_SCROLLABLE_ATTRIBUTE) !== 'false') {
        const styles: CSSStyleDeclaration = getComputedStyle(el);
        const overflowY: string = styles ? styles.getPropertyValue('overflow-y') : '';

        if (overflowY && (overflowY === 'scroll' || overflowY === 'auto')) {
          return el;
        }
      }

      el = el.parentElement;
    }

    // Fall back to window scroll.
    if (!el || el === body) {
      el = undefined;
    }

    return el;
  }

  private static _initialize(): void {
    LazyEventHandler._async = new Async();
    LazyEventHandler._initTime = new Date();
    LazyEventHandler._lazyHandler =
      LazyEventHandler._async.throttle(LazyEventHandler._lazyLoaderEventHandler, 100);
    _TraceLogger.logVerbose(LazyEventHandler._logSource, 'Webpart.LazyLoad.Initialize');
  }

  private static _dispose(): void {
    const timeSpent: number =  (new Date()).getTime() - LazyEventHandler._initTime.getTime();
    _EngagementLogger.logEvent('Webpart.LazyLoad.LazyLoadingCompleted', timeSpent.toString());

    // Undefine all members
    LazyEventHandler._lazyHandler = undefined;
    LazyEventHandler._initTime = undefined;
    LazyEventHandler._async = undefined;
  }

  /**
  * Start listening scroll and resize events
  */
  private static _listenEvents(parent: Element | Window): void {
    if (!parent) {
      return;
    }

    switch (parent) {
      case window:
        if (!LazyEventHandler._isListeningWindow) {
          parent.addEventListener('scroll', LazyEventHandler._lazyHandler, false);
          parent.addEventListener('resize', LazyEventHandler._lazyHandler, false);
          LazyEventHandler._isListeningWindow = true;
        }

        break;

      default:
        parent.addEventListener('scroll', LazyEventHandler._lazyHandler, false);
        parent.addEventListener('resize', LazyEventHandler._lazyHandler, false);
        break;
    }

    return;
  }

  /**
  * Stop listening scroll and resize events
  */
  private static _unlistenEvents(parent: Element | Window): void {
    if (!parent) {
      return;
    }

    switch (parent) {
      case window:
        if (LazyEventHandler._isListeningWindow) {
          parent.removeEventListener('resize', LazyEventHandler._lazyHandler);
          parent.removeEventListener('scroll', LazyEventHandler._lazyHandler);
          LazyEventHandler._isListeningWindow = false;
        }

        break;

      default:
        parent.removeEventListener('resize', LazyEventHandler._lazyHandler);
        parent.removeEventListener('scroll', LazyEventHandler._lazyHandler);
        break;
    }

    return;
  }

  /**
   * It keeps track of how many Lazy childs are registered to a scrollable parent
   * This record keeping is required to deregister event handler from the parent
   */
  private static _updateRegisteredChildren(element: Element, count: number): number {
    const lazyChilds: number =
      count + Number(element.getAttribute(LazyEventHandler._lazyLoadedTracker));
    if (count !== 0) {
      element.setAttribute(LazyEventHandler._lazyLoadedTracker, lazyChilds.toString());
    }
    return lazyChilds;
  }

  private static _renderComponent(component: LazyComponent): void {
    if (component && !component.state.isRendered) {
      component.setState({ isRendered: true });
      component.forceUpdate();
      if (component.props.onRendered) {
        component.props.onRendered();
      }
    }
  }
}
