/**
 * Lazy Loader internal implementation. Use SPLazyLoader for lazy components
 * @Copyright (c) Microsoft Corporation.  All rights reserved.
 *
 * @file LazyLoader.tsx
 */

import * as React from 'react';
import LazyEventHandler from './LazyEventHandler';

export interface ILazyLoaderProps extends React.Props<any> { // tslint:disable-line:no-any
  height?: number;
  offset?: number;
  placeholder?: React.ReactElement<void>;
  onRendered?: () => void;
}

export interface ILazyLoaderState {
  isRendered: boolean;
}

/**
 * LazyLoader component ensures that the only child component is rendered only  when
 * is it going to be visible. It rather renders provided placeholder or empty division with provided height.
 */
export default class LazyLoader extends React.Component<ILazyLoaderProps, ILazyLoaderState> {
  public static defaultProps: ILazyLoaderProps = {
    height: 250,
    offset:  (window.innerHeight || document.documentElement.clientHeight) / 3
  };

  constructor(props: ILazyLoaderProps) {
    super(props);
    this.state = {
      isRendered: false
    };
  }

  public componentDidMount(): void {
    LazyEventHandler.registerComponent(this);
  }

  public shouldComponentUpdate(): boolean {
    return this.state.isRendered;
  }

  public componentWillUnmount(): void {
    LazyEventHandler.unregisterComponent(this);
  }

  public render(): React.ReactElement<any> { // tslint:disable-line:no-any
    // Close to viewport? let's render actual child component
    if (this.state.isRendered) {
      return React.Children.only(this.props.children);
    }

    // Render placeholder if provided
    if (this.props.placeholder) {
      return this.props.placeholder;
    }

    // Render default empty placeholder with the height provided
    return (
      <div
        { ...{ height: this.props.height } /* VSO:391730 Remove this workaround */ }
        className={ 'office-ui-lazyloader-placeholder' } >
      </div>);
  }
}
