/**
 * @copyright Microsoft Corporation. All rights reserved.
 *
 * @file LessText.test.ts
 */

import { expect } from 'chai';
import { mount, ReactWrapper } from 'enzyme';
import * as React from 'react';

import ILessTextProps from './ILessTextProps';
import LessText from './LessText';

describe('<LessText />', () => {
  const textToBeTruncated: string = 'It is a very long string that needs to be truncated by LessText component.';
  let lessText: ReactWrapper<ILessTextProps, void>;
  let lessTextContainer: HTMLDivElement;
  let mockProps: ILessTextProps;

  beforeEach(() => {
    lessTextContainer = document.createElement('div');
    lessTextContainer.style.width = '100px';
    lessTextContainer.style.height = '40px';
    lessTextContainer.style.lineHeight = '20px';

    document.body.appendChild(lessTextContainer);

    mockProps = {
      text: textToBeTruncated,
      className: 'testClass',
      showTitle: true,
      lines: 2,
      omission: '+++' // Rename it for testing.
    };
  });

  afterEach(() => {
    if (lessText) {
      lessText.unmount();
    }
  });

  it('should display the passed className.', () => {
    lessText = mount<ILessTextProps, void>(<LessText {...mockProps} />, {
      attachTo: lessTextContainer
    });
    expect(lessText.find('.testClass')).to.have.lengthOf(1);
  });

  it('should have the passed text in title attribute.', () => {
    lessText = mount<ILessTextProps, void>(<LessText {...mockProps} />, {
      attachTo: lessTextContainer
    });
    expect(lessText.find(`[data-automation-id='less-text']`)).to.have.lengthOf(1);
    expect(lessText.find(`[data-automation-id='less-text']`).first().prop('title')).to.equal(textToBeTruncated);
  });

  it('should have text truncated when text is too long', (done) => {
    lessText = mount<ILessTextProps, void>(<LessText {...mockProps} />, {
      attachTo: lessTextContainer
    });

    // Text is updated after a new animation frame. Use timeout to check after that.
    setTimeout(() => {
      const actualText: string = document.querySelector(`[data-automation-id='less-text']`)!.textContent!;
      const actualTextWithoutOmission: string = actualText.substring(0, actualText.indexOf('+++'));

      expect(actualText.indexOf('+++')).to.not.equal(-1);
      expect(textToBeTruncated.indexOf(actualTextWithoutOmission)).to.equal(0);
      expect(actualTextWithoutOmission.length).to.be.below(textToBeTruncated.length);
      done();
    }, 100);
  });

  it('should not truncate text when text is short', (done) => {
    lessText = mount<ILessTextProps, void>(<LessText {...mockProps} text='short text' />, {
      attachTo: lessTextContainer
    });

    // Text is updated after a new animation frame. Use timeout to check after that.
    setTimeout(() => {
      const actualText: string = document.querySelector(`[data-automation-id='less-text']`)!.textContent!;

      expect(actualText).to.equal('short text');
      done();
    }, 100);
  });

  it('should truncate text when window resize is triggered', (done) => {
    lessText = mount<ILessTextProps, void>(<LessText {...mockProps} />, {
      attachTo: lessTextContainer
    });

    // Text is updated after a new animation frame. Use timeout to check after that.
    setTimeout(() => {
      const firstTruncatedText: string = document.querySelector(`[data-automation-id='less-text']`)!.textContent!;

      lessTextContainer.style.width = '200px';
      window.dispatchEvent(new Event('resize'));

      // Text is updated after a new animation frame. Use timeout to check after that.
      setTimeout(() => {
        const updatedText: string = document.querySelector(`[data-automation-id='less-text']`)!.textContent!;
        expect(firstTruncatedText).to.not.equal(updatedText);

        done();
      }, 100);
    }, 100);
  });

  it('should truncate text when window resize is triggered', (done) => {
    lessText = mount<ILessTextProps, void>(<LessText {...mockProps} />, {
      attachTo: lessTextContainer
    });

    // Text is updated after a new animation frame. Use timeout to check after that.
    setTimeout(() => {
      const lessTextElement: HTMLDivElement =
        document.querySelector(`[data-automation-id='less-text']`)! as HTMLDivElement;

      expect(lessTextElement.clientHeight).to.equal(40);
      done();
    }, 100);
  });

  it('should truncate with a single long word', (done) => {
    const singleLongWord:  string = 'Lopadotemachoselachogaleokranioleipsanodrimhypotrimmatosilphiopar' +
      'aomelitokatakechymenokichlepikossyphophattoperisteralektryonoptekephalliokigklope' +
      'leiolagoiosiraiobaphetraganopterygon';

    lessText = mount<ILessTextProps, void>(<LessText {...mockProps} text={singleLongWord} />, {
      attachTo: lessTextContainer
    });

    // Text is updated after a new animation frame. Use timeout to check after that.
    setTimeout(() => {
      const lessTextElement: HTMLDivElement =
        document.querySelector(`[data-automation-id='less-text']`)! as HTMLDivElement;
      const actualText: string = lessTextElement.textContent!;

      expect(lessTextElement.clientHeight).to.equal(40);
      expect(actualText.indexOf('+++')).to.not.equal(-1);

      done();
    }, 100);
  });
});
