/**
 * @copyright Microsoft Corporation. All rights reserved.
 *
 * @file LessText.tsx
 */

import { autobind, css, Async, EventGroup } from 'office-ui-fabric-react';
import * as React from 'react';

import ILessTextProps from './ILessTextProps';

import styles from './LessText.module.scss';

// Note: Any changes in this file should be in sync with LessText in sp-canvas-toolbox project.
// Ticket to add support from OUFR level: https://github.com/OfficeDev/office-ui-fabric-react/issues/2882
export default class LessText extends React.PureComponent<ILessTextProps, {}> {

  private _async: Async;
  private _eventGroup: EventGroup;
  private _omission: string;
  private _separator: string;
  private _textArray: string[];
  private _textContainerElement: null | HTMLElement;

  constructor(props: ILessTextProps) {
    super(props);
    this._async = new Async(this);
    this._eventGroup = new EventGroup(this);
    this._separator = props.separator || ''; // Break between each characters by default.
    this._omission = props.omission || '…';
    const delay: number =
      this.props.resizeDelayInMilliseconds ? this.props.resizeDelayInMilliseconds : 50;

    this._handleWindowResize = this._async.debounce(this._handleWindowResize, delay);
  }

  public componentDidMount(): void {
    // Leverage `setTimeout` here to wait for the batch async styles are loaded.
    this._async.setTimeout(this.update, 0);
    this._eventGroup.on(window, 'resize', this._handleWindowResize);
  }

  public componentDidUpdate(): void {
    this.update();
  }

  public componentWillUnmount(): void {
    this._eventGroup.off(window, 'resize', this._handleWindowResize);
    this._eventGroup.dispose();
    this._async.dispose();
  }

  public render(): JSX.Element {
    return (
      <div
        data-automation-id='less-text'
        ref={(ref) => this._textContainerElement = ref}
        title={this.props.showTitle ? this.props.text : undefined}
        className={css(styles.lessText, this.props.className)}
      >
        {this.props.text}
      </div>
    );
  }

  /**
   * Call update at runtime to truncate the text whenever on needs.
   */
  @autobind
  public update(): void {
    if (this._textContainerElement) {
      this._async.requestAnimationFrame(this._truncateLines);
    }
  }

  private _handleWindowResize(): void {
    this.update();
  }

  private get _contentHeight(): number {
    if (this._textContainerElement) {
      const contentStyles: CSSStyleDeclaration = window.getComputedStyle(this._textContainerElement);
      let padding: number = 0;
      if (contentStyles && contentStyles.paddingTop && contentStyles.paddingBottom) {
        padding = parseFloat(contentStyles.paddingTop) + parseFloat(contentStyles.paddingBottom);
      }

      return this._textContainerElement.offsetHeight - padding;
    }

    return 0;
  }

  @autobind
  private _truncateLines(): void {
    if (!this._textContainerElement) {
      return;
    }

    const text: string = this.props.text || '';
    this._textContainerElement.textContent = text;

    const lines: number = Math.round(this.props.lines);
    const lineHeight: string = window
      .getComputedStyle(this._textContainerElement, undefined)
      .getPropertyValue('line-height');
    const maxHeight: number = lines * parseInt(lineHeight, 10);
    const currHeight: number = this._contentHeight;
    this._textArray = text.split(this._separator);

    if (currHeight > maxHeight && lines > 0) {
      const trimmedLength: number = this._trimAndCalculateLength(currHeight, maxHeight);
      this._applyTextNumber(trimmedLength);
    }
  }

  private _trimAndCalculateLength(
    currentValue: number,
    maxValue: number
  ): number {
    const allowedTextNumber: number = this._textArray.length;
    if (!this._textContainerElement) {
      return allowedTextNumber;
    }

    let minTextNumber: number = 1, maxTextNumber: number = allowedTextNumber;

    do {
      const currentTextNumber: number = Math.floor((minTextNumber + maxTextNumber + 1) / 2);
      this._applyTextNumber(currentTextNumber);
      currentValue = this._contentHeight;

      if (currentValue > maxValue) {
        maxTextNumber = currentTextNumber - 1;
      } else {
        minTextNumber = currentTextNumber;
      }
    } while (minTextNumber < maxTextNumber);

    return minTextNumber;
  }

  private _applyTextNumber(newTextNumber: number): void {
    const newText: string[] = this._textArray.slice(0, newTextNumber);
    const newTextJoined: string = newText.join(this._separator);

    if (this._textContainerElement) {
      this._textContainerElement.textContent = this._textContainerElement.textContent === newTextJoined
        ? newTextJoined
        : newTextJoined + this._omission;
    }
  }
}