/**
 * @copyright Microsoft Corporation. All rights reserved.
 *
 * @file GridList.
 */

import * as React from 'react';
import TouchArea, { SwipeDirection } from '../touchArea/TouchArea';
import {
  autobind,
  Panel,
  PanelType,
  KeyCodes,
  FocusTrapZone
} from 'office-ui-fabric-react';
import SlideChangeButton, { ButtonDirection } from './SlideChangeButton';
import styles from './css/LightBox.module.scss';
import strings from './loc/Strings.resx';

export interface ILightBoxProps<T> {
  items: T[];
  startIndex: number;
  isShown: boolean;
  onRenderItem: (item: T, index: number) => JSX.Element;
  onClose: () => void;
  elementToFocusOnDismiss?: HTMLElement;
}

export interface ILightBoxState {
  itemCount?: number;
  currentIndex?: number;
}

export default class LightBox<T> extends React.Component<ILightBoxProps<T>, ILightBoxState> {

  private _isInternalIndexChange: boolean = false;

  constructor(props: ILightBoxProps<T>) {
    super(props);
    this._closePanel = this._closePanel.bind(this);
    this._indexForward = this._indexForward.bind(this);
    this._indexBack = this._indexBack.bind(this);
    this._onKeyPress = this._onKeyPress.bind(this);
    this._onSwipe = this._onSwipe.bind(this);
    this._getItemsToRender = this._getItemsToRender.bind(this);
    this.state = {
      itemCount: props.items.length,
      currentIndex: props.startIndex
    };
  }

  public componentWillReceiveProps(nextProps: ILightBoxProps<T>): void {
    const currentIndex: number = this._isInternalIndexChange ?
      this.state.currentIndex :
      nextProps.startIndex >= 0 && nextProps.startIndex < nextProps.items.length ?
        nextProps.startIndex : 0;
    this.setState({ currentIndex: currentIndex });
  }

  public render(): JSX.Element {
    if (!this.props.items || this.props.items.length === 0) {
      return null; // tslint:disable-line:no-null-keyword
    }

    return (
      <Panel className={ styles.lightBoxPanel }
        isOpen={ this.props.isShown }
        type={ PanelType.smallFluid }
        onDismiss={ this._closePanel }
        closeButtonAriaLabel={ strings.LightBoxButtonCloseLabel }
        headerText={ strings.LightBoxLabel }
        headerClassName={ styles.screenReaderOnly }>
        { this.props.isShown &&
          <FocusTrapZone
            isClickableOutsideFocusTrap={ true }
            forceFocusInsideTrap={ true }
            elementToFocusOnDismiss={ this.props.elementToFocusOnDismiss }
            ignoreExternalFocusing={ false }>
            <div role='application'
              className={ styles.subPanel }
              onKeyDown={ this._onKeyPress }>
              { this._getLeftSlideChangeButton() }
              { this._getRightSlideChangeButton() }
              <TouchArea onSwipe={ this._onSwipe }
                preventDefault= { true }>
                { this._getItemsToRender() }
              </TouchArea>
            </div>
          </FocusTrapZone>
        }
      </Panel>);
  }

  private _closePanel(): boolean {
    this._isInternalIndexChange = false;
    this.props.onClose();
    return true;
  }

  private _indexBack(event?: React.MouseEvent<any>): void { // tslint:disable-line:no-any
    this._isInternalIndexChange = true;
    let index: number = this.state.currentIndex;
    if (index === 0) {
      index = this.props.items.length - 1;
    } else {
      index--;
    }

    this.setState({
      currentIndex: index
    });

    if (event) {
      event.nativeEvent.returnValue = false; // Avoid classic page postback
    }
  }

  private _indexForward(event?: React.MouseEvent<any>): void { // tslint:disable-line:no-any
    this._isInternalIndexChange = true;
    let index: number = this.state.currentIndex;
    if (index === this.props.items.length - 1) {
      index = 0;
    } else {
      index++;
    }
    this.setState({
      currentIndex: index
    });

    if (event) {
      event.nativeEvent.returnValue = false; // Avoid classic page postback
    }
  }

  /* tslint:disable-next-line:no-any */
  private _onKeyPress(ev: React.KeyboardEvent<any>): void {
    switch (ev.keyCode) {
      case KeyCodes.right:
        this._indexForward();
        break;
      case KeyCodes.left:
        this._indexBack();
        break;
    }
  }

  private _onSwipe(direction: SwipeDirection): void {
    switch (direction) {
      case SwipeDirection.Left:
        this._indexBack();
        break;
      case SwipeDirection.Right:
        this._indexForward();
        break;
    }
  }

  private _getItemsToRender(): JSX.Element {
    const itemNumber: number = this.props.items.length;
    const prevIndex: number = this.state.currentIndex > 0 ? this.state.currentIndex - 1 : itemNumber - 1;

    const prevItem: JSX.Element = itemNumber > 2 &&
      <div className={ styles.hiddenSlide } data-automation-id='prev-item' aria-hidden={ true }>
        { this.props.onRenderItem(this.props.items[prevIndex], prevIndex) }
      </div>;

    const nextIndex: number = this.state.currentIndex === itemNumber - 1 ? 0 : this.state.currentIndex + 1;

    const nextItem: JSX.Element = itemNumber > 1 &&
      <div className={ styles.hiddenSlide } data-automation-id='next-item' aria-hidden={ true }>
        { this.props.onRenderItem(this.props.items[nextIndex], nextIndex) }
      </div>;

    return (<div className={ styles.contentHolder }>
      { prevItem }
      <div aria-live='polite'>
        { this.props.onRenderItem(this.props.items[this.state.currentIndex], this.state.currentIndex) }
      </div>
      { nextItem }
    </div>
    );
  }

  @autobind
  private _getLeftSlideChangeButton(): JSX.Element {
    return this.props.items.length > 1 &&
      <SlideChangeButton
        onClick={ this._indexBack }
        buttonDirection={ ButtonDirection.Prev } />;
  }

  @autobind
  private _getRightSlideChangeButton(): JSX.Element {
    return this.props.items.length > 1 &&
      <SlideChangeButton
        onClick={ this._indexForward }
        buttonDirection={ ButtonDirection.Next }
        shouldFocus={ !this._isInternalIndexChange && this.props.isShown === true } />;
  }
}
