/**
 * @file MySiteWebPartCacheDataProvider.ts
 * @copyright Microsoft Corporation. All rights reserved.
 *
 * @remark DataProvider for the MySiteWebPartCacheManager.
 */

import { Uri } from '@ms/odsp-utilities-bundle';
import { Environment, Guid, ServiceScope, Validate } from '@microsoft/sp-core-library';
import { SPHttpClient } from '@microsoft/sp-http';
import { PageContext } from '@microsoft/sp-page-context';
import { Async, autobind } from 'office-ui-fabric-react';
import { _EngagementLogger, _LogSource, _PerformanceLogger, _QosMonitor, _TraceLogger } from '@ms/sp-telemetry';
import { IDataRow } from '@ms/sp-component-utilities';
import { IMySiteWebPartCacheDataItem } from './IMySiteWebPartCacheDataItem';
import { MySiteWebPartCacheHelper } from './MySiteWebPartCacheHelper';

// Window Variables
const PREFETCHED_WEBPARTDATA: string = `_spWebPartData`;
// Property and Field titles
const FIELDNAME: string = `FieldName`;
const FIELDVALUE: string = `FieldValue`;
const SPBUILTINFIELDID_ID: string = `Id`;
const SPBUILTINFIELDID_TITLE: string = `Title`;
const SPHOMECACHELIST_HASHCODE_FIELDNAME: string = `HashCode`;
const SPHOMECACHELIST_CACHEVERSION_FIELDNAME: string = `CacheVersion`;
const SPHOMECACHELIST_VALUE_FIELDNAME: string = `Value`;
// QOS tags
const QOS_PREFIX: string = `MySiteWebPartCacheDataProvider.`;
// setTimeout delay
const DELAY: number = 3000;
// Events
// WORKAROUND: sboyle: remove once proper in-place navigation is implemented (VSO:305861)
const SITEPAGE_LOADED: string = `sitePageLoaded: `;

export class MySiteWebPartCacheDataProvider {
  private static _logSource: _LogSource = _LogSource.create('MySiteWebPartCacheDataProvider');
  private _cacheItemsMap: Map<string, IMySiteWebPartCacheDataItem> = new Map<string, IMySiteWebPartCacheDataItem>();
  private _shouldCreatePageCacheFolder: boolean;
  private _shouldCreateWebPartCacheFolder: boolean;
  private _currentPageCacheFolderKey: string;
  private _currentWebAbsoluteUrl: string;
  private _mySiteAbsoluteUrl: string;
  private _spHttpClient: SPHttpClient;
  private _utcTimeStamp: Date;
  private _webPartCachePath: string;
  private _pageContext: PageContext;
  private _async: Async;
  // WORKAROUND: sboyle: remove eventlistener once proper in-place navigation is implemented (VSO:305861)
  private _messageEventListener: (ev: MessageEvent) => void;

  public constructor(serviceScope: ServiceScope) {
    serviceScope.whenFinished(() => {
      this._pageContext = serviceScope.consume(PageContext.serviceKey);
      this._spHttpClient = serviceScope.consume(SPHttpClient.serviceKey);
      this._async = new Async(this);
      // WORKAROUND: sboyle: remove eventlistener once proper in-place navigation is implemented (VSO:305861)
      this._messageEventListener = this._onReceivePageNavMessage;
      window.addEventListener('message', this._messageEventListener);
      this._initialize();
      _EngagementLogger.logEvent('MySiteWebPartCacheDataProvider.Constructor');
    });
  }

  /**
   * Returns cached data for the calling WebPart while handling caching and re-rendering.
   *
   * @param requestData - A function that returns an IDataRow[] and makes the service data request.
   * @param dataUpdated - A function that returns the updated data when there is a cache hit on the original
   *                      request.
   * @param config - The current WebPart config.
   * @param webPartInstanceId - The WebPart instanceId.
   * @param webPartAlias - The WebPart alias (manifest alias).
   * @param rerender - Controls whether or not to re-render on new data after rendering cached data.
   *                   Default: Always re-render.
   * @param secondsToInvalidate - A number of seconds representing the amount of time to expire cached results.
   *                              Default: Don't expire cached results automatically.
   * @returns - A promise that represents WebPart data returned from the request or cache.
   */
  public getWebPartData(
    requestData: () => Promise<IDataRow[]>,
    dataUpdated: (response: IDataRow[]) => void,
    config: any, // tslint:disable-line:no-any
    webPartInstanceId: string,
    webPartAlias: string,
    rerender: boolean = true,
    secondsToInvalidate: number = -1): Promise<IDataRow[]> {
      Validate.isNotNullOrUndefined(requestData, 'requestData');
      Validate.isNotNullOrUndefined(dataUpdated, 'dataUpdated');
      Validate.isNotNullOrUndefined(config, 'config');
      Validate.isNonemptyString(webPartInstanceId, 'webPartInstanceId');
      Validate.isNonemptyString(webPartAlias, 'webPartAlias');

      const qosSuffix: string = `getWebPartData`;
      const qosMonitor: _QosMonitor = new _QosMonitor(QOS_PREFIX + qosSuffix);
      let rawResponseString: string = ``;
      let response: IDataRow[] = new Array<IDataRow>();
      const cachedWebPart: IMySiteWebPartCacheDataItem = this._cacheItemsMap.get(webPartInstanceId);
      const webPartHash: string = MySiteWebPartCacheHelper.generateHash(JSON.stringify(config));
      _TraceLogger.logVerbose(
        MySiteWebPartCacheDataProvider._logSource,
        `WebPart: ${webPartInstanceId}, Page: ${this._currentPageCacheFolderKey}, Environment: ${Environment.type}.`);
      _EngagementLogger.logEvent('MySiteWebPartCacheDataProvider.GetWebPartData');
      _PerformanceLogger.startMarkForComponent(
        `WebPart.${webPartAlias}.${webPartInstanceId}`);

      if (cachedWebPart) {
        if (MySiteWebPartCacheHelper.isExpired(this._utcTimeStamp, cachedWebPart.modified, secondsToInvalidate)) {
              // This WebPart is cached but the cached item has passed its invalidation time.
              _TraceLogger.logVerbose(
                MySiteWebPartCacheDataProvider._logSource,
                `Requested cached WebPart item ${webPartInstanceId} exists but is expired.`);

              return requestData().then((rawResponse: IDataRow[]) => {
                response = rawResponse;
                rawResponseString = JSON.stringify(response);

                this._async.setTimeout(() => {
                  this._tryUpdateCache(
                    webPartInstanceId,
                    webPartAlias,
                    rawResponseString,
                    webPartHash,
                    cachedWebPart.id);
                },
                DELAY);

                qosMonitor.writeSuccess(this._getWebPartIdLogObject(webPartAlias, webPartInstanceId));
              })

              .catch((error: Error) => { // tslint:disable-line:typedef
                // This will tell us if requestData() or dataUpdated fails, specifically.
                _TraceLogger.logError(
                  MySiteWebPartCacheDataProvider._logSource,
                  error,
                  `getWebPartData: error while processing cached WebPart item ${webPartInstanceId}.`);
                  qosMonitor.writeUnexpectedFailure(
                    `ExpiredCacheUpdateError`,
                    error,
                    this._getWebPartIdLogObject(webPartAlias, webPartInstanceId)
                  );
                if (DEBUG) {
                  throw error;
                }
              })

              .then(() => {
                return response;
              });

        } else {
          try {
            if (webPartHash === cachedWebPart.hash) {
              // This WebPart is cached; the config has not changed.
              _TraceLogger.logVerbose(
                MySiteWebPartCacheDataProvider._logSource,
                `Found cached item ${webPartInstanceId} with matching config.`);

              requestData().then((rawResponse: IDataRow[]) => {
                response = rawResponse;
                rawResponseString = JSON.stringify(response);
                if (cachedWebPart.result !== rawResponseString) {
                  // This WebPart is cached; the config has not changed, but the result needs updating.
                  _TraceLogger.logVerbose(
                    MySiteWebPartCacheDataProvider._logSource,
                    `Requested item ${webPartInstanceId} has newer results; updating.`);

                  this._saveRenderAndUpdateItem(
                    dataUpdated,
                    rerender,
                    response,
                    webPartInstanceId,
                    webPartAlias,
                    rawResponseString,
                    webPartHash,
                    cachedWebPart.id);
                  qosMonitor.writeSuccess(this._getWebPartIdLogObject(webPartAlias, webPartInstanceId));

                } else {
                  // This WebPart is cached; the result and config have not changed; No op.
                  _TraceLogger.logVerbose(
                    MySiteWebPartCacheDataProvider._logSource,
                    `Requested cached item ${webPartInstanceId} has matching result; No write back.`);
                  qosMonitor.writeSuccess(this._getWebPartIdLogObject(webPartAlias, webPartInstanceId));
                }
              })

              .catch((error: Error): void => {
              _TraceLogger.logError(
                MySiteWebPartCacheDataProvider._logSource,
                error,
                `getWebPartData: error while processing cached WebPart item ${webPartInstanceId}.`);
                qosMonitor.writeUnexpectedFailure(
                  `CachedItemUpdatedResultError`,
                  error,
                  this._getWebPartIdLogObject(webPartAlias, webPartInstanceId)
                );
              });

            } else {
              // This WebPart is cached; but the config seems to have changed; update the exisiting item.
              _TraceLogger.logVerbose(
                MySiteWebPartCacheDataProvider._logSource,
                `Requested cached item ${webPartInstanceId} has different config. Updating.`);

              requestData().then((rawResponse: IDataRow[]) => {
                response = rawResponse;
                rawResponseString = JSON.stringify(response);
                this._saveRenderAndUpdateItem(
                  dataUpdated,
                  rerender,
                  response,
                  webPartInstanceId,
                  webPartAlias,
                  rawResponseString,
                  webPartHash,
                  cachedWebPart.id);
                qosMonitor.writeSuccess(this._getWebPartIdLogObject(webPartAlias, webPartInstanceId));
              })

              .catch((error: Error): void => {
              _TraceLogger.logError(
                MySiteWebPartCacheDataProvider._logSource,
                error,
                `getWebPartData: error while processing cached WebPart item ${webPartInstanceId}.`);
                qosMonitor.writeUnexpectedFailure(
                  `CachedItemUpdatedConfigError`,
                  error,
                  this._getWebPartIdLogObject(webPartAlias, webPartInstanceId)
                );
              });
            }

          } catch (error) {
            _TraceLogger.logError(
              MySiteWebPartCacheDataProvider._logSource,
              error,
            `getWebPartData: error while processing cached WebPart item ${webPartInstanceId}.`);
            qosMonitor.writeUnexpectedFailure(
              `CachedItemUpdateError`,
              error,
              this._getWebPartIdLogObject(webPartAlias, webPartInstanceId)
            );

          } finally {
            // If the WebPart is cached, we should always at least return the cached result, regardless of error.
            _TraceLogger.logVerbose(
              MySiteWebPartCacheDataProvider._logSource,
              `Returning cached item result for ${webPartInstanceId}.`);
            _PerformanceLogger.markComponent(
              `WebPart.${webPartAlias}.${webPartInstanceId}`, 'CacheHit');
            _PerformanceLogger.endMarkForComponent(`WebPart.${webPartAlias}.${webPartInstanceId}`);
            return Promise.resolve(JSON.parse(cachedWebPart.result));
          }
        }

      } else {
        // This WebPart is not cached; create a new cached item.
        _TraceLogger.logVerbose(
          MySiteWebPartCacheDataProvider._logSource,
          `Requested cached WebPart item ${webPartInstanceId} doesn't exist. Creating.`);

        return requestData().then((rawResponse: IDataRow[]) => {
          response = rawResponse;
          rawResponseString = JSON.stringify(response);

          this._async.setTimeout(() => {
            this._tryUpdateCache(
              webPartInstanceId,
              webPartAlias,
              rawResponseString,
              webPartHash);
          },
          DELAY);

          qosMonitor.writeSuccess(
            {
              'alias': webPartAlias,
              'instanceId': webPartInstanceId
            });
          _EngagementLogger.logEvent('MySiteWebPartCacheDataProvider.CacheMiss');
          _PerformanceLogger.markComponent(
              `WebPart.${webPartAlias}.${webPartInstanceId}`, 'CacheMiss');
          _PerformanceLogger.endMarkForComponent(`WebPart.${webPartAlias}.${webPartInstanceId}`);
        })

        .catch((error: Error) => { // tslint:disable-line:typedef
          // This will tell us if requestData() or dataUpdated fails, specifically.
          _TraceLogger.logError(
            MySiteWebPartCacheDataProvider._logSource,
            error,
            `getWebPartData: error while processing non-cached WebPart item ${webPartInstanceId}.`);
            qosMonitor.writeUnexpectedFailure(
              `NewItemRequestSaveRenderError`,
              error,
              this._getWebPartIdLogObject(webPartAlias, webPartInstanceId)
            );
          if (DEBUG) {
            throw error;
          }
        })

        .then(() => {
          return response;
        });
      }
    }

  /**
   * Saves, renders, and updates cache data.
   *
   * @param dataUpdated - A function that saves the response to the WebPart and then renders.
   * @param rerender - Controls whether or not to re-render on new data after rendering cached data.
   *                   Default: Always re-render.
   * @param response - The response from a requestData() call.
   * @param webPartInstanceId - The WebPart instanceId.
   * @param webPartAlias - The WebPart alias (manifest alias).
   * @param rawResponseString - The raw response string from a requestData() call.
   *                            We need the raw string before saving/transformation for consistency.
   * @param webPartHash - The hash of the WebPart config to write back to the cache.
   * @param cacheItemId - The ID (SharePointListItem ID) of the item to update.
   */
  private _saveRenderAndUpdateItem(
    dataUpdated: (response: IDataRow[]) => void,
    rerender: boolean,
    response: IDataRow[],
    webPartInstanceId: string,
    webPartAlias: string,
    rawResponseString: string,
    webPartHash: string,
    cacheItemId: number): void {
      Validate.isNotNullOrUndefined(dataUpdated, 'dataUpdated');
      Validate.isNotNullOrUndefined(rerender, 'rerender');
      Validate.isNotNullOrUndefined(response, 'response');
      Validate.isNonemptyString(webPartInstanceId, 'webPartInstanceId');
      Validate.isNonemptyString(webPartAlias, 'webPartAlias');
      Validate.isNonemptyString(rawResponseString, 'rawResponseString');
      Validate.isNonemptyString(webPartHash, 'webPartHash');
      Validate.isNotNullOrUndefined(cacheItemId, 'cacheItemId');

      if (rerender) {
        dataUpdated(response);
      }
      this._async.setTimeout(() => {
        this._updateCache(webPartInstanceId, webPartAlias, rawResponseString, webPartHash, cacheItemId);
      },
      DELAY);
  }

  /**
   * Initialize the MySiteWebPartCacheDataProvider.
   */
  private _initialize(): void {
    const qosSuffix: string = `_initialize`;
    const qosMonitor: _QosMonitor = new _QosMonitor(QOS_PREFIX + qosSuffix);
    const spPageContextInfoError: string = `Page ${this._currentPageCacheFolderKey} is missing _spPageContextInfo.`;
    const spWebPartDataError: string = `Page ${this._currentPageCacheFolderKey} is missing _pageContext.`;

    if (this._pageContext) {
      this._getPrefetchedPageContextProperties(this._pageContext.legacyPageContext);

      if (window[PREFETCHED_WEBPARTDATA]) {
        this._getPrefectedWebPartDataProperties(window[PREFETCHED_WEBPARTDATA].Properties);
        this._getPrefetchedWebPartDataItems(window[PREFETCHED_WEBPARTDATA].Items);
        qosMonitor.writeSuccess();

      } else {
        _TraceLogger.logError(
          MySiteWebPartCacheDataProvider._logSource,
          new Error(),
          spWebPartDataError);
        qosMonitor.writeUnexpectedFailure(`SPWebPartDataError`, new Error());
      }

    } else {
      _TraceLogger.logError(
        MySiteWebPartCacheDataProvider._logSource,
        new Error(),
        spPageContextInfoError);
      qosMonitor.writeUnexpectedFailure(`SPPageContextInfoError`, new Error());
    }
  }

  /**
   * Get the flushed WebPartData Properties.
   */
  private _getPrefectedWebPartDataProperties(data: any): void { // tslint:disable-line:no-any
    Validate.isNotNullOrUndefined(data, 'data');

    this._mySiteAbsoluteUrl = data.MySiteUrl;
    this._currentPageCacheFolderKey = data.PageCacheFolderKey;
    this._webPartCachePath = data.WebPartCachePath;
    this._shouldCreatePageCacheFolder = data.CreatePageCacheFolder as boolean;
    this._shouldCreateWebPartCacheFolder = data.CreateWebPartCacheFolder as boolean;
    if (data.UtcTimeStamp) {
      this._utcTimeStamp = MySiteWebPartCacheHelper.dateTimeStringToJsDate(data.UtcTimeStamp);
    }
  }

  /**
   * Get the flushed Page Context Properties.
   */
  private _getPrefetchedPageContextProperties(data: any): void { // tslint:disable-line:no-any
    Validate.isNotNullOrUndefined(data, 'data');
    Validate.isNonemptyString(data.webAbsoluteUrl, 'data.webAbsoluteUrl');

    this._currentWebAbsoluteUrl = data.webAbsoluteUrl;
  }

  /**
   * Get the flushed WebPartData Items.
   */
  private _getPrefetchedWebPartDataItems(data: any): void { // tslint:disable-line:no-any
    Validate.isNotNullOrUndefined(data, 'data');

    Object.keys(data).forEach((key) => {
      const cacheItem: IMySiteWebPartCacheDataItem = {
        webPartInstanceId: key,
        result: data[key].Value,
        hash: data[key].HashCode,
        id: Number(data[key].ID),
        uniqueId: Guid.parse(data[key].UniqueId),
        modified: MySiteWebPartCacheHelper.dateTimeStringToJsDate(data[key].Modified),
        cacheVersion: Number(data[key].CacheVersion)
      };
      this._cacheItemsMap.set(key, cacheItem);
    });
    _TraceLogger.logVerbose(
      MySiteWebPartCacheDataProvider._logSource,
      `_cacheItemsMap Size: ${this._cacheItemsMap.size}.`);
  }

  /**
   * Create a remote subfolder.
   */
  private _createRemoteSubfolder(
    relativeRemoteFolderPath: string,
    folderTitle: string,
    shouldCreate: boolean): Promise<{}> {
      Validate.isNonemptyString(relativeRemoteFolderPath, 'relativeRemoteFolderPath');
      Validate.isNonemptyString(folderTitle, 'folderTitle');

      if (!shouldCreate) {
        _TraceLogger.logVerbose(
          MySiteWebPartCacheDataProvider._logSource,
          `Folder: ${folderTitle} already created; no op.`);
        return Promise.resolve();
      }

      _TraceLogger.logVerbose(
        MySiteWebPartCacheDataProvider._logSource,
        `Creating folder: ${folderTitle}.`);

      const url: string = MySiteWebPartCacheHelper.getRemoteSubfolderCreateUrl(
        this._currentWebAbsoluteUrl,
        this._mySiteAbsoluteUrl,
        relativeRemoteFolderPath,
        folderTitle);

      return MySiteWebPartCacheHelper.post(
        this._spHttpClient,
        url,
        {},
        MySiteWebPartCacheHelper.checkFolderStatus
      );
  }

  /**
   * Create a Remote item.
   */
  private _createCacheItem(
    webRelativeFolderPath: string,
    title: string,
    value: string,
    hash: string): Promise<Response> {
      Validate.isNonemptyString(webRelativeFolderPath, 'webRelativeFolderPath');
      Validate.isNonemptyString(title, 'title');
      Validate.isNonemptyString(value, 'value');
      Validate.isNonemptyString(hash, 'hash');

      const url: string = MySiteWebPartCacheHelper.getRemoteItemCreateUrl(
        this._currentWebAbsoluteUrl,
        this._mySiteAbsoluteUrl,
        this._webPartCachePath
      );

      const body: {} = MySiteWebPartCacheHelper.getCreateItemBody(
        webRelativeFolderPath,
        title,
        value,
        hash
      );

      return MySiteWebPartCacheHelper.post(
        this._spHttpClient,
        url,
        body,
        MySiteWebPartCacheHelper.checkStatus
      );
  }

  /**
   * Update a remote item.
   */
  private _updateCacheItem(value: string, hash: string, itemID: number): Promise<Response> {
    Validate.isNonemptyString(value, 'value');
    Validate.isNonemptyString(hash, 'hash');
    Validate.isNotNullOrUndefined(itemID, 'itemID');

    const url: string = MySiteWebPartCacheHelper.getRemoteItemUpdateUrl(
      this._currentWebAbsoluteUrl,
      this._mySiteAbsoluteUrl,
      this._webPartCachePath,
      itemID.toString()
    );

    const body: {} = MySiteWebPartCacheHelper.getUpdateItemBody(value, hash);

    return MySiteWebPartCacheHelper.post(
      this._spHttpClient,
      url,
      body,
      MySiteWebPartCacheHelper.checkStatus
    );
  }

  /**
   * Update the cache.
   * 1. Check and Create the WebPartCache folder.
   * 2. Check and Create the specific Page cache folder.
   * 3. If we already have the webPartUniqueID in the cachedItemMap, we update the existing item.
   * 4. If we haven't seen this item before, we create a new item.
   * Note: If your intention is not to block on this call, make sure you use an async setTimeout.
   */
  private _updateCache(
    webPartInstanceId: string,
    webPartAlias: string,
    result: string,
    hash: string,
    id?: number): void {
      const qosSuffix: string = `_updateCache`;
      const qosMonitor: _QosMonitor = new _QosMonitor(QOS_PREFIX + qosSuffix);
      // We update the error tag as we enter each chained Promise.
      let qosUpdateCacheError: string = `UnknownError`;

      try {
        Validate.isNonemptyString(webPartInstanceId, 'webPartInstanceId');

        const webPartCachePath: string = MySiteWebPartCacheHelper.getUrlWithoutLastSegment(this._webPartCachePath);
        const webPartCacheFolderName: string = MySiteWebPartCacheHelper.getUrlLastSegment(this._webPartCachePath);

        // If we need to create the single WebPartCache folder, create it.
        qosUpdateCacheError = `WebPartCacheFolderCreateError`;
        this._createRemoteSubfolder( // tslint:disable-line:no-unused-expression
          webPartCachePath,
          webPartCacheFolderName,
          this._shouldCreateWebPartCacheFolder

        ).then( () => {
          // If we need to create the specific Page cache folder inside the WebPartCache folder, create it.
          qosUpdateCacheError = `PageCacheFolderCreateError`;
          return this._createRemoteSubfolder(
            this._webPartCachePath,
            this._currentPageCacheFolderKey,
            this._shouldCreatePageCacheFolder);

        }).then( () => {
          // If the map contains a local cache entry, we update the server cache item.
          if (this._cacheItemsMap.has(webPartInstanceId)) {
            _TraceLogger.logVerbose(
              MySiteWebPartCacheDataProvider._logSource,
              `Updating Cache Item ${webPartInstanceId}.`);

            qosUpdateCacheError = `UpdateCacheItemError`;
            return this._updateCacheItem(
              result,
              hash,
              id).then((response: Response) => {
                return this._updateCacheItemInMap(response, webPartInstanceId, webPartAlias);
              });

          // Else this must be a new entry, create the server cache item.
          } else {
            const folderPath: string =
              Uri.concatenate(
                this._webPartCachePath,
                this._currentPageCacheFolderKey);

            _TraceLogger.logVerbose(
              MySiteWebPartCacheDataProvider._logSource,
              `Creating Cache Item ${webPartInstanceId}.`);

            qosUpdateCacheError = `CreateCacheItemError`;
            return this._createCacheItem(
              folderPath,
              webPartInstanceId,
              result,
              hash).then((response: Response) => {
                return this._createCacheItemInMap(response);
              });
          }

        }).then( () => {
          qosMonitor.writeSuccess(this._getWebPartIdLogObject(webPartAlias, webPartInstanceId));

        }).catch((error: Error): void => {
          _TraceLogger.logError(
            MySiteWebPartCacheDataProvider._logSource,
            error,
            /* tslint:disable-next-line:max-line-length */
            `_updateCache: ${qosUpdateCacheError} WebPart: ${webPartInstanceId} Page: ${this._currentPageCacheFolderKey}`);
          qosMonitor.writeUnexpectedFailure(
            qosUpdateCacheError,
            error,
            this._getWebPartIdLogObject(webPartAlias, webPartInstanceId)
          );
        });
      } catch (error) {
        qosMonitor.writeUnexpectedFailure(
          qosUpdateCacheError,
          error,
          this._getWebPartIdLogObject(webPartAlias, webPartInstanceId)
        );
      }
  }

  /**
   * Update the cache without throwing an exception.
   */
  private _tryUpdateCache(
    webPartInstanceId: string,
    webPartAlias: string,
    result: string,
    hash: string,
    id?: number): void {
      try {
          this._updateCache(webPartInstanceId, webPartAlias, result, hash, id);
      } catch (error) {
        _TraceLogger.logError(
          MySiteWebPartCacheDataProvider._logSource,
          error,
        `getWebPartData: error while processing item ${webPartInstanceId}.`);
      }
  }

  /**
   * Adds an item (i.e., read from the prefetched content) to the client cache map.
   */
  private _createCacheItemInMap(response: Response): Promise<Response> {
    return this._writeToCacheItemMap(response);
  }

  /**
   * Updates an item (i.e., read from the prefetched content) to the client cache map.
   */
  private _updateCacheItemInMap(
    response: Response,
    webPartInstanceId: string,
    webPartAlias: string): Promise<Response> {
      return this._writeToCacheItemMap(response, webPartInstanceId);
    }

  /**
   * Writes or updates items in the client cache map depending on input.
   * On update, the response doesn't contain the webpartinstanceid so we need to pass it in as a key.
   */
  private _writeToCacheItemMap(
    response: Response,
    webPartInstanceId?: string): Promise<Response> {
      Validate.isNotNullOrUndefined(response, 'response');

      const qosSuffix: string = `_writeToCacheItemMap`;
      const qosMonitor: _QosMonitor = new _QosMonitor(QOS_PREFIX + qosSuffix);
      let qosWriteToCacheItemMapError: string = `UnknownError`;
      return response.json().then((data: any) => { // tslint:disable-line:no-any
        Validate.isNotNullOrUndefined(data, 'data');
        Validate.isNotNullOrUndefined(data.value, 'data.value');

        let cacheItemKey: string;
        // If this is an update response, we need to pass these in to get a complete update.
        if (webPartInstanceId) {
          cacheItemKey = webPartInstanceId;
        }
        let cacheItemId: number;
        let cacheItemHash: string;
        let cacheItemVersion: number;
        let cacheItemResult: string;
        const cacheItemModified: Date = MySiteWebPartCacheHelper.getDateNowAsUTC();
        // Parse Create and Update responses.
        qosWriteToCacheItemMapError = `ResponseParseError`;
        Object.keys(data.value).forEach((key) => {
          const fieldName: string = data.value[key][FIELDNAME];
          const fieldValue: string = data.value[key][FIELDVALUE];
          switch (fieldName) {
            // These cases are only in the create response.
            case SPBUILTINFIELDID_ID:
              cacheItemId = Number(fieldValue);
              break;
            case SPBUILTINFIELDID_TITLE:
              cacheItemKey = fieldValue;
              break;
            // These cases are in both update and create responses.
            case SPHOMECACHELIST_HASHCODE_FIELDNAME:
              cacheItemHash = fieldValue;
              break;
            case SPHOMECACHELIST_CACHEVERSION_FIELDNAME:
              cacheItemVersion = Number(fieldValue);
              break;
            case SPHOMECACHELIST_VALUE_FIELDNAME:
              cacheItemResult = fieldValue;
              break;
            default:
              break;
          }
        });
        if (this._cacheItemsMap.has(cacheItemKey)) {
          // Update an existing item with an item update response.
          qosWriteToCacheItemMapError = `UpdateCacheItemMapError`;
          const existingCacheItem: IMySiteWebPartCacheDataItem = this._cacheItemsMap.get(cacheItemKey);
          existingCacheItem.result = cacheItemResult;
          existingCacheItem.hash = cacheItemHash;
          existingCacheItem.modified = cacheItemModified;
        } else {
          // Add a new item from an item creation response.
          qosWriteToCacheItemMapError = `CreateCacheItemMapError`;
          const cacheItem: IMySiteWebPartCacheDataItem = {
            webPartInstanceId: cacheItemKey,
            result: cacheItemResult,
            hash: cacheItemHash,
            id: cacheItemId,
            modified: cacheItemModified,
            cacheVersion: cacheItemVersion
          };
          this._cacheItemsMap.set(cacheItemKey, cacheItem);
        }
        _TraceLogger.logVerbose(
          MySiteWebPartCacheDataProvider._logSource,
          `_cacheItemsMap Size: ${this._cacheItemsMap.size}.`);
        qosMonitor.writeSuccess();
        return response;
      })

      .catch((error: Error) => {
        _TraceLogger.logError(
          MySiteWebPartCacheDataProvider._logSource,
          error,
          `_writeToCacheItemMap: error client caching response for WebPart: ${webPartInstanceId}`);
        qosMonitor.writeUnexpectedFailure(
          qosWriteToCacheItemMapError,
          error);
        return Promise.resolve<Response>(undefined);
      });
    }

    // WORKAROUND: sboyle: remove once proper in-place navigation is implemented (VSO:305861)
    @autobind
    private _onReceivePageNavMessage(message: MessageEvent): void {
      if (message && message.data && message.origin &&
        message.origin === window.location.origin &&
        typeof message.data === 'string' &&
        message.data.indexOf(SITEPAGE_LOADED) === 0) {
          this._currentPageCacheFolderKey = message.data.substring((SITEPAGE_LOADED).length);
          _EngagementLogger.logEvent('MySiteWebPartCacheDataProvider.InPageNavTriggered');
        }
    }

    private _getWebPartIdLogObject(webPartAlias: string, webPartInstanceId: string): {} {
      return { 'alias': webPartAlias, 'instanceId': webPartInstanceId };
    }
}