/**
 * @file MySiteWebPartCacheHelper.ts
 * @copyright Microsoft Corporation. All rights reserved.
 *
 * @remark Helper class for the MySiteWebPartCache.
 */

import { Uri, UriEncoding } from '@ms/odsp-utilities-bundle';
import { Validate } from '@microsoft/sp-core-library';
import { SPHttpClient } from '@microsoft/sp-http';

import { Md5Hash } from '@ms/sp-component-utilities';

// Current MySite Cache Schema Version
const CURRENT_CACHE_VERSION: string = `1.0`;
// Entity Types
const SPHOME_CACHELISTITEM_ENTITYTYPE: string = `SP.Data.SharePointHomeCacheListListItem`;
const LISTITEMCREATIONINFORMATIONUSINGPATH_ENTITYTYPE: string = `SP.ListItemCreationInformationUsingPath`;
const RESOURCEPATH_ENTITYTYPE: string = `SP.ResourcePath`;
// Property and Field titles
const SPBUILTINFIELDID_TITLE: string = `Title`;
const SPHOMECACHELIST_HASHCODE_FIELDNAME: string = `HashCode`;
const SPHOMECACHELIST_CACHEVERSION_FIELDNAME: string = `CacheVersion`;
const SPHOMECACHELIST_VALUE_FIELDNAME: string = `Value`;
// REST API Segments
const FOLDERS_ADDSUBFOLDER_API_SEGMENT: string = `/AddSubFolder(@subfolderTitle)`;
const ITEMS_VALIDATEUPDATELISTITEM: string = `/items(@itemID)/ValidateUpdateListItem`;
const REMOTEWEB_API_ENDPOINT: string = `/_api/SP.RemoteWeb(@remoteWeb)`;
const WEB_GETLIST_SEGMENT: string = `/Web/GetList(@webRelativeListPath)`;
const WEB_ADDVALIDATEUPDATEITEMUSINGPATH: string = `/AddValidateUpdateItemUsingPath`;
const WEB_GETFOLDERBYSERVERRELATIVEPATH_API_SEGMENT: string =
  `/Web/GetFolderByServerRelativePath(DecodedUrl=@parentFolderRelativeUrl)`;

export class MySiteWebPartCacheHelper {

  /**
   * Creates a REST url for creating a subfolder at the specified Remote folder location.
   *
   * @param currentWebAbsoluteUrl - Absolute web url of the current context.
   * @param remoteWebAbsoluteUrl - Absolute web url of the target remote web.
   * @param remoteParentFolderRelativeUrl - Relative url of the folder to create the subfolder in.
   * @param remoteFolderName - Title of the remote subfolder to be created.
   *
   * @returns - REST url and query string to create a subfolder at a Remote folder location.
   */
  public static getRemoteSubfolderCreateUrl(
    currentWebAbsoluteUrl: string,
    remoteWebAbsoluteUrl: string,
    remoteParentFolderRelativeUrl: string,
    remoteFolderName: string): string {
      Validate.isNonemptyString(currentWebAbsoluteUrl, 'currentWebAbsoluteUrl');
      Validate.isNonemptyString(remoteWebAbsoluteUrl, 'remoteWebAbsoluteUrl');
      Validate.isNonemptyString(remoteParentFolderRelativeUrl, 'remoteParentFolderRelativeUrl');
      Validate.isNonemptyString(remoteFolderName, 'remoteFolderName');

      const encodedRemoteWebAbsoluteUrl: string = UriEncoding.encodeRestUriStringToken(remoteWebAbsoluteUrl);
      const encodedRemoteParentFolderRelativeUrl: string = UriEncoding.encodeRestUriStringToken(
        remoteParentFolderRelativeUrl);
      const encodedRemoteFolderName: string = UriEncoding.encodeRestUriStringToken(remoteFolderName);

      const url: string =
      Uri.concatenate(
        Uri.concatenate(
          Uri.concatenate(
            currentWebAbsoluteUrl,
            REMOTEWEB_API_ENDPOINT),
            WEB_GETFOLDERBYSERVERRELATIVEPATH_API_SEGMENT),
            FOLDERS_ADDSUBFOLDER_API_SEGMENT) +
            `?@remoteWeb='${encodedRemoteWebAbsoluteUrl}'` +
            `&@parentFolderRelativeUrl='${encodedRemoteParentFolderRelativeUrl}'` +
            `&@subfolderTitle='${encodedRemoteFolderName}'`;
      return url;
    }

  /**
   * Retrieves the last uri segment.
   *
   * @param url - The target url.
   *
   * @returns - The last segment of the given url.
   */
  public static getUrlLastSegment(url: string): string {
    Validate.isNonemptyString(url, 'url');
    const uri: Uri = new Uri(url);
    return uri.getLastPathSegment();
  }

  /**
   * Retrieves the url, minus the last segment.
   *
   * @param url - The target url.
   *
   * @returns - The url, minus the last segment.
   */
  public static getUrlWithoutLastSegment(url: string): string {
    Validate.isNonemptyString(url, 'url');
    const lastSlashIndex: number = url.lastIndexOf('/');
    const urlWithoutLastSegment: string = (lastSlashIndex > -1) ? url.substring(0, lastSlashIndex) : url;
    return urlWithoutLastSegment;
  }

  /**
   * Creates a REST url for creating an item at the specified Remote folder location.
   *
   * @param currentWebAbsoluteUrl - Absolute web url of the current context.
   * @param remoteWebAbsoluteUrl - Absolute web url of the target remote web.
   * @param webRelativeListPath - Relative url of the folder to create the item in.
   *
   * @returns - REST url and query string to create an item at a Remote folder location.
   */
  public static getRemoteItemCreateUrl(
    currentWebAbsoluteUrl: string,
    remoteWebAbsoluteUrl: string,
    webRelativeListPath: string): string {
      Validate.isNonemptyString(currentWebAbsoluteUrl, 'currentWebAbsoluteUrl');
      Validate.isNonemptyString(remoteWebAbsoluteUrl, 'remoteWebAbsoluteUrl');
      Validate.isNonemptyString(webRelativeListPath, 'webRelativeListPath');

      const encodedRemoteWebAbsoluteUrl: string = UriEncoding.encodeRestUriStringToken(remoteWebAbsoluteUrl);
      const encodedWebRelativeListPath: string = UriEncoding.encodeRestUriStringToken(webRelativeListPath);

      const url: string =
      Uri.concatenate(
        Uri.concatenate(
          Uri.concatenate(
            currentWebAbsoluteUrl,
            REMOTEWEB_API_ENDPOINT),
            WEB_GETLIST_SEGMENT),
            WEB_ADDVALIDATEUPDATEITEMUSINGPATH) +
            `?@remoteWeb='${encodedRemoteWebAbsoluteUrl}'` +
            `&@webRelativeListPath='${encodedWebRelativeListPath}'`;
      return url;
    }

  /**
   * Creates a REST url for updating a Remote item.
   *
   * @param currentWebAbsoluteUrl - Absolute web url of the current context.
   * @param remoteWebAbsoluteUrl - Absolute web url of the target remote web.
   * @param webRelativeListPath - Relative url of the folder to create the item in.
   * @param itemID - The target SharePoint item ID.
   *
   * @returns - REST url and query string to update an item at a Remote folder location.
   */
  public static getRemoteItemUpdateUrl(
    currentWebAbsoluteUrl: string,
    remoteWebAbsoluteUrl: string,
    webRelativeListPath: string,
    itemID: string): string {
      Validate.isNonemptyString(currentWebAbsoluteUrl, 'currentWebAbsoluteUrl');
      Validate.isNonemptyString(remoteWebAbsoluteUrl, 'remoteWebAbsoluteUrl');
      Validate.isNonemptyString(webRelativeListPath, 'webRelativeListPath');
      Validate.isNonemptyString(itemID, 'itemID');

      const encodedRemoteWebAbsoluteUrl: string = UriEncoding.encodeRestUriStringToken(remoteWebAbsoluteUrl);
      const encodedWebRelativeListPath: string = UriEncoding.encodeRestUriStringToken(webRelativeListPath);

      const url: string =
      Uri.concatenate(
        Uri.concatenate(
          Uri.concatenate(
            currentWebAbsoluteUrl,
            REMOTEWEB_API_ENDPOINT),
            WEB_GETLIST_SEGMENT),
            ITEMS_VALIDATEUPDATELISTITEM) +
            `?@remoteWeb='${encodedRemoteWebAbsoluteUrl}'` +
            `&@webRelativeListPath='${encodedWebRelativeListPath}'` +
            `&@itemID='${itemID}'`;
      return url;
    }

  /**
   * Convert date string format like 2016-08-24T23:13:18.97-07:00 to Date object.
   *
   * @param dateTime - A date string like 2016-08-24T23:13:18.97-07:00
   *
   * @returns - A Date object with the datetime of the input string.
   */
  public static dateTimeStringToJsDate(dateTime: string): Date {
    Validate.isNonemptyString(dateTime, 'dateTime');

    const timeValues: string[] = dateTime.split(/[^0-9]/);
    if (timeValues.length < 6) {
      return; // error
    } else {
      return new Date(Date.UTC(
        Number(timeValues[0]),
        Number(timeValues[1]) - 1,
        Number(timeValues[2]),
        Number(timeValues[3]),
        Number(timeValues[4]),
        Number(timeValues[5])));
    }
  }

  /**
   * Checks the status of a response and resolves/rejects the Promise.
   *
   * @param Response - The response to check.
   *
   * @returns - A resolved/rejected Promise object.
   */
  public static checkStatus(response: Response): Promise<Response> {
    Validate.isNotNullOrUndefined(response, 'response');

    if ((response.status >= 200 && response.status < 300)) {
      return Promise.resolve(response);
    } else {
      return Promise.reject(new Error(JSON.stringify(response)));
    }
  }

  /**
   * Checks the status of a response and resolves/rejects the Promise.
   * Note: Because of a folder creation bug, 500 is a valid response.
   *
   * @param Response - The response to check.
   *
   * @returns - A resolved/rejected Promise object.
   */
  public static checkFolderStatus(response: Response): Promise<Response> {
    Validate.isNotNullOrUndefined(response, 'response');

    if ((response.status >= 200 && response.status < 300) || response.status === 500) {
      return Promise.resolve(response);
    } else {
      return Promise.reject(new Error(JSON.stringify(response)));
    }
  }

  /**
   * Creates a JSON body for creating a WebPart Cache Item.
   *
   * @param webRelativeFolderPath - Web relative folder path for the cache item.
   * @param title - Title of the cache item.
   * @param value - Value of the result to be cached.
   * @param hash - Hash of the config to be cached.
   *
   * @returns - JSON body for creating a WebPart Cache Item.
   */
  public static getCreateItemBody(
    webRelativeFolderPath: string,
    title: string,
    value: string,
    hash: string): {} {
      Validate.isNonemptyString(webRelativeFolderPath, 'webRelativeFolderPath');
      Validate.isNonemptyString(title, 'title');
      Validate.isNonemptyString(value, 'value');
      Validate.isNonemptyString(hash, 'hash');

      const body: {} = {
        'listItemCreateInfo': {
            '__metadata': {
              'type': `${LISTITEMCREATIONINFORMATIONUSINGPATH_ENTITYTYPE}`
            },
            'FolderPath': {
              '__metadata': {
                'type': `${RESOURCEPATH_ENTITYTYPE}`
              },
              'DecodedUrl': `${webRelativeFolderPath}`
            }
        },
        'formValues': [
          {
            'FieldName': `${SPBUILTINFIELDID_TITLE}`,
            'FieldValue': `${title}`,
            'HasException': false,
            'ErrorMessage': null // tslint:disable-line:no-null-keyword
          },
          {
            'FieldName': `${SPHOMECACHELIST_VALUE_FIELDNAME}`,
            'FieldValue': `${value}`,
            'HasException': false,
            'ErrorMessage': null // tslint:disable-line:no-null-keyword
          },
          {
            'FieldName': `${SPHOMECACHELIST_CACHEVERSION_FIELDNAME}`,
            'FieldValue': `${CURRENT_CACHE_VERSION}`,
            'HasException': false,
            'ErrorMessage': null // tslint:disable-line:no-null-keyword
          },
          {
            'FieldName': `${SPHOMECACHELIST_HASHCODE_FIELDNAME}`,
            'FieldValue': `${hash}`,
            'HasException': false,
            'ErrorMessage': null // tslint:disable-line:no-null-keyword
          },
          {
            'FieldName': `ContentType`,
            'FieldValue': `${SPHOME_CACHELISTITEM_ENTITYTYPE}`,
            'HasException': false,
            'ErrorMessage': null // tslint:disable-line:no-null-keyword
          }
        ],
        'bNewDocumentUpdate': false,
        'checkInComment': null // tslint:disable-line:no-null-keyword
      };
      return body;
  }

  /**
   * Creates a JSON body for updating a WebPart Cache Item.
   *
   * @param value - Value of the result to be cached.
   * @param hash - Hash of the config to be cached.
   *
   * @returns - JSON body for updating a WebPart Cache Item.
   */
  public static getUpdateItemBody(
    value: string,
    hash: string): {} {
      Validate.isNonemptyString(value, 'value');
      Validate.isNonemptyString(hash, 'hash');

      const body: {} = {
        'formValues': [
          {
            'FieldName': `${SPHOMECACHELIST_VALUE_FIELDNAME}`,
            'FieldValue': `${value}`,
            'HasException': false,
            'ErrorMessage': null // tslint:disable-line:no-null-keyword
          },
          {
            'FieldName': `${SPHOMECACHELIST_HASHCODE_FIELDNAME}`,
            'FieldValue': `${hash}`,
            'HasException': false,
            'ErrorMessage': null // tslint:disable-line:no-null-keyword
          },
          {
            'FieldName': `${SPHOMECACHELIST_CACHEVERSION_FIELDNAME}`,
            'FieldValue': `${CURRENT_CACHE_VERSION}`,
            'HasException': false,
            'ErrorMessage': null // tslint:disable-line:no-null-keyword
          }
        ],
        'bNewDocumentUpdate': false,
        'checkInComment': null // tslint:disable-line:no-null-keyword
      };
      return body;
  }

  /**
   * Posts to SPHTTPClient
   *
   * @param spHttpClient - The SPHttpClient object to post with.
   * @param url - The url to post to.
   * @param body - The body to post with.
   *
   * @returns - JSON body for updating a WebPart Cache Item.
   */
  public static post(
    spHttpClient: SPHttpClient,
    url: string,
    body: {},
    checkStatusFunction: (response: Response) => Promise<Response>): Promise<Response> {
      Validate.isNotNullOrUndefined(spHttpClient, 'spHttpClient');
      Validate.isNonemptyString(url, 'url');
      Validate.isNotNullOrUndefined(body, 'body');

      return spHttpClient.post(
        url,
        SPHttpClient.configurations.v1,
        {
          body: JSON.stringify(body),
          headers: { 'OData-Version': '3.0' } // we will convert to 4 when there are docs with examples :)
        }
      ).then(checkStatusFunction);
  }

  /**
   * Create an MD5 hash from an input string.
   *
   * @param stringTohash - The string to hash.
   *
   * @returns - A hashed string.
   */
  public static generateHash(stringToHash: string): string {
    return Md5Hash(stringToHash);
  }

  /**
   * Create an UTC Date object (time -> Now).
   *
   * @returns - A UTC Date object.
   */
  public static getDateNowAsUTC(): Date {
    const date: Date = new Date();
    return new Date(
      Date.UTC(
        date.getFullYear(),
        date.getMonth(),
        date.getDate(),
        date.getHours(),
        date.getMinutes(),
        date.getSeconds()));
  }

  /**
   * Get a Date with seconds added.
   *
   * @param date - The Date to add seconds to.
   * @param seconds - The seconds to add to the Date.
   *
   * @returns - A Date with seconds added.
   */
  public static addSecondsToDate(date: Date, seconds: number): Date {
    Validate.isNotNullOrUndefined(date, 'date');
    Validate.isNotNullOrUndefined(seconds, 'seconds');
    const datePlusSeconds: Date = new Date(date.toString());
    datePlusSeconds.setTime(datePlusSeconds.getTime() + (seconds * 1000));
    return datePlusSeconds;
  }

  /**
   * Compare a date with a date plus seconds added.
   *
   * @param referenceDate - The Date to add seconds to.
   * @param dateToCompare - The Date to compare to reference you will add seconds to.
   * @param secondsToAdd - A number of seconds to add to dateToCompare.
   *
   * @returns - True (boolean) if the referenceDate is greater than the dateToCompare plus secondsToAdd.
   */
  public static isExpired(
    referenceDate: Date,
    dateToCompare: Date,
    secondsToAdd: number): boolean {
      return referenceDate &&
        dateToCompare &&
        secondsToAdd &&
        secondsToAdd > 0 &&
        referenceDate > MySiteWebPartCacheHelper.addSecondsToDate(dateToCompare, secondsToAdd);
  }
}