/// <reference types="mocha" />

import { expect } from 'chai';
import { Uri, UriEncoding } from '@ms/odsp-utilities-bundle';
import { MySiteWebPartCacheHelper } from './../MySiteWebPartCacheHelper';

/* tslint:disable:no-unused-expression */

describe('MySiteWebPartCacheHelper tests', () => {

  // Current MySite Cache Schema Version
  const CURRENT_CACHE_VERSION: string = `1.0`;
  // Entity Types
  const SPHOME_CACHELISTITEM_ENTITYTYPE: string = `SP.Data.SharePointHomeCacheListListItem`;
  const LISTITEMCREATIONINFORMATIONUSINGPATH_ENTITYTYPE: string = `SP.ListItemCreationInformationUsingPath`;
  const RESOURCEPATH_ENTITYTYPE: string = `SP.ResourcePath`;
  // Property and Field titles
  const SPBUILTINFIELDID_TITLE: string = `Title`;
  const SPHOMECACHELIST_HASHCODE_FIELDNAME: string = `HashCode`;
  const SPHOMECACHELIST_CACHEVERSION_FIELDNAME: string = `CacheVersion`;
  const SPHOMECACHELIST_VALUE_FIELDNAME: string = `Value`;
  // REST API Segments
  const FOLDERS_ADDSUBFOLDER_API_SEGMENT: string = `/AddSubFolder(@subfolderTitle)`;
  const ITEMS_VALIDATEUPDATELISTITEM: string = `/items(@itemID)/ValidateUpdateListItem`;
  const REMOTEWEB_API_ENDPOINT: string = `/_api/SP.RemoteWeb(@remoteWeb)`;
  const WEB_GETLIST_SEGMENT: string = `/Web/GetList(@webRelativeListPath)`;
  const WEB_ADDVALIDATEUPDATEITEMUSINGPATH: string = `/AddValidateUpdateItemUsingPath`;
  const WEB_GETFOLDERBYSERVERRELATIVEPATH_API_SEGMENT: string =
    `/Web/GetFolderByServerRelativePath(DecodedUrl=@parentFolderRelativeUrl)`;

  const title: string = `216592E5-3AA5-4DA5-A2B3-BE427DC8FE7B`;
  const value: string = `randomvalue`;
  const hash: string = `randomhash`;
  const webRelativeFolderPath: string = `/personal/sboyle_microsoft_com/Lists/SharePointHomeCacheList/WebPartCache`;
  const currentWebAbsoluteUrl: string = `https://msft.spoppe.com/sites/test`;
  const remoteWebAbsoluteUrl: string = `https://msft-my.spoppe.com/personal/sboyle_microsoft_com`;
  const remoteFolderName: string = `random folder name`;
  const itemId: string = `1234`;
  const urlWithoutLastSegment: string = `https://msft-my.spoppe.com/personal`;
  const lastSegment: string = `sboyle_microsoft_com`;
  const requestDateOffset: number = 30;
  const secondsToExpire: number = 10;

  const updateItemBody: {} = {
    'formValues': [
      {
        'FieldName': `${SPHOMECACHELIST_VALUE_FIELDNAME}`,
        'FieldValue': `${value}`,
        'HasException': false,
        'ErrorMessage': null // tslint:disable-line:no-null-keyword
      },
      {
        'FieldName': `${SPHOMECACHELIST_HASHCODE_FIELDNAME}`,
        'FieldValue': `${hash}`,
        'HasException': false,
        'ErrorMessage': null // tslint:disable-line:no-null-keyword
      },
      {
        'FieldName': `${SPHOMECACHELIST_CACHEVERSION_FIELDNAME}`,
        'FieldValue': `${CURRENT_CACHE_VERSION}`,
        'HasException': false,
        'ErrorMessage': null // tslint:disable-line:no-null-keyword
      }
    ],
    'bNewDocumentUpdate': false,
    'checkInComment': null // tslint:disable-line:no-null-keyword
  };

  const createItemBody: {} = {
    'listItemCreateInfo': {
        '__metadata': {
          'type': `${LISTITEMCREATIONINFORMATIONUSINGPATH_ENTITYTYPE}`
        },
        'FolderPath': {
          '__metadata': {
            'type': `${RESOURCEPATH_ENTITYTYPE}`
          },
          'DecodedUrl': `${webRelativeFolderPath}`
        }
    },
    'formValues': [
      {
        'FieldName': `${SPBUILTINFIELDID_TITLE}`,
        'FieldValue': `${title}`,
        'HasException': false,
        'ErrorMessage': null // tslint:disable-line:no-null-keyword
      },
      {
        'FieldName': `${SPHOMECACHELIST_VALUE_FIELDNAME}`,
        'FieldValue': `${value}`,
        'HasException': false,
        'ErrorMessage': null // tslint:disable-line:no-null-keyword
      },
      {
        'FieldName': `${SPHOMECACHELIST_CACHEVERSION_FIELDNAME}`,
        'FieldValue': `${CURRENT_CACHE_VERSION}`,
        'HasException': false,
        'ErrorMessage': null // tslint:disable-line:no-null-keyword
      },
      {
        'FieldName': `${SPHOMECACHELIST_HASHCODE_FIELDNAME}`,
        'FieldValue': `${hash}`,
        'HasException': false,
        'ErrorMessage': null // tslint:disable-line:no-null-keyword
      },
      {
        'FieldName': `ContentType`,
        'FieldValue': `${SPHOME_CACHELISTITEM_ENTITYTYPE}`,
        'HasException': false,
        'ErrorMessage': null // tslint:disable-line:no-null-keyword
      }
    ],
    'bNewDocumentUpdate': false,
    'checkInComment': null // tslint:disable-line:no-null-keyword
  };

  const preGeneratedRemoteItemCreateUrl: string =
    Uri.concatenate(
      Uri.concatenate(
        Uri.concatenate(
          currentWebAbsoluteUrl,
          REMOTEWEB_API_ENDPOINT),
          WEB_GETLIST_SEGMENT),
          WEB_ADDVALIDATEUPDATEITEMUSINGPATH) +
          `?@remoteWeb='${UriEncoding.encodeRestUriStringToken(remoteWebAbsoluteUrl)}'` +
          `&@webRelativeListPath='${UriEncoding.encodeRestUriStringToken(webRelativeFolderPath)}'`;

  const preGeneratedRemoteSubfolderCreateUrl: string =
    Uri.concatenate(
      Uri.concatenate(
        Uri.concatenate(
          currentWebAbsoluteUrl,
          REMOTEWEB_API_ENDPOINT),
          WEB_GETFOLDERBYSERVERRELATIVEPATH_API_SEGMENT),
          FOLDERS_ADDSUBFOLDER_API_SEGMENT) +
          `?@remoteWeb='${UriEncoding.encodeRestUriStringToken(remoteWebAbsoluteUrl)}'` +
          `&@parentFolderRelativeUrl='${UriEncoding.encodeRestUriStringToken(webRelativeFolderPath)}'` +
          `&@subfolderTitle='${UriEncoding.encodeRestUriStringToken(remoteFolderName)}'`;

  const preGeneratedRemoteItemUpdateUrl: string =
    Uri.concatenate(
      Uri.concatenate(
        Uri.concatenate(
          currentWebAbsoluteUrl,
          REMOTEWEB_API_ENDPOINT),
          WEB_GETLIST_SEGMENT),
          ITEMS_VALIDATEUPDATELISTITEM) +
          `?@remoteWeb='${UriEncoding.encodeRestUriStringToken(remoteWebAbsoluteUrl)}'` +
          `&@webRelativeListPath='${UriEncoding.encodeRestUriStringToken(webRelativeFolderPath)}'` +
          `&@itemID='${UriEncoding.encodeRestUriStringToken(itemId)}'`;

  it('should correctly generate remoteItemUpdateUrl', () => {
    expect(MySiteWebPartCacheHelper.getRemoteItemUpdateUrl(
      currentWebAbsoluteUrl,
      remoteWebAbsoluteUrl,
      webRelativeFolderPath,
      itemId)).to.be.equal(preGeneratedRemoteItemUpdateUrl);
  });

  it('should correctly generate remoteSubfolderCreateUrl', () => {
    expect(MySiteWebPartCacheHelper.getRemoteSubfolderCreateUrl(
      currentWebAbsoluteUrl,
      remoteWebAbsoluteUrl,
      webRelativeFolderPath,
      remoteFolderName)).to.be.equal(preGeneratedRemoteSubfolderCreateUrl);
  });

  it('should correctly generate remoteItemCreateUrl', () => {
    expect(MySiteWebPartCacheHelper.getRemoteItemCreateUrl(
      currentWebAbsoluteUrl,
      remoteWebAbsoluteUrl,
      webRelativeFolderPath)).to.be.equal(preGeneratedRemoteItemCreateUrl);
  });

  it('should correctly create an update Item body', () => {
    expect(MySiteWebPartCacheHelper.getUpdateItemBody(
      value,
      hash)).to.be.deep.equal(updateItemBody);
  });

  it('should correctly create a create Item body', () => {
    expect(MySiteWebPartCacheHelper.getCreateItemBody(
      webRelativeFolderPath,
      title,
      value,
      hash)).to.be.deep.equal(createItemBody);
  });

  it('should correctly get the url without the last segment', () => {
    expect(MySiteWebPartCacheHelper.getUrlWithoutLastSegment(remoteWebAbsoluteUrl)).to.be.equal(urlWithoutLastSegment);
  });

  it('should correctly get the last segment of a url', () => {
    expect(MySiteWebPartCacheHelper.getUrlLastSegment(remoteWebAbsoluteUrl)).to.be.equal(lastSegment);
  });

  it('should correctly interpret an expired Date', () => {
    const lastModifiedDate: Date = new Date();
    const requestDate: Date = MySiteWebPartCacheHelper.addSecondsToDate(lastModifiedDate, requestDateOffset);
    const isExpired: boolean = MySiteWebPartCacheHelper.isExpired(requestDate, lastModifiedDate, secondsToExpire);
    expect(isExpired).to.be.true;
  });
});