/**
 * @Copyright (c) Microsoft Corporation.  All rights reserved.
 *
 * @file Common component used to display the error placeholder in case of OAuth Error.
 *
 */

import * as React from 'react';

import { autobind } from 'office-ui-fabric-react';
import { sealed } from '@microsoft/decorators';
import {
  _EngagementLogger,
  _ILogEntry,
  _LogEntry,
  _LogType
} from '@ms/sp-telemetry';

import {
  IPlaceholderProps,
  Placeholder
} from '../placeHolder';
import { IOAuthErrorRendererProps } from './IOAuthErrorRendererProps';
import strings from './loc/Strings.resx';

/**
 * Common component used to display the error placeholder in case of OAuth Error.
 *
 * @internal
 */
@sealed
export default class OAuthErrorRenderer extends React.Component<IOAuthErrorRendererProps, {}> {
  /**
   * Render the error place holder user experience.
   */
  public render(): React.ReactElement<IPlaceholderProps> {
    const isMultiFactor: boolean = !!this.props.error.redirectUrl;
    const friendlyErrorMessage: string = isMultiFactor ? strings.MultiFactorAuthMessage :
      this.props.customErrorMessage || this.props.error.message;
    return (
      <div className={ this.props.contentClassName }>
        <Placeholder
          icon={ this.props.webpartIcon || 'Permissions' }
          iconText={ this.props.webpartName }
          buttonLabel={ isMultiFactor && strings.LoginButtonLabel }
          description={ friendlyErrorMessage }
          onAdd={ isMultiFactor && this._onLogin }>
        </Placeholder>
      </div>
    );
  }

  // Should know how often multi-factor auth is occurring
  public componentDidMount(): void {
    if (this.props.error.redirectUrl) {
      this._logAction('Render');
    }
  }

  @autobind
  private _onLogin(): void {
    this._logAction('Click');

    window.location.href = this.props.error.redirectUrl;
  }

  @autobind
  private _logAction(action: string): void {
    const logEntry: _ILogEntry = new _LogEntry(
      'OAuthErrorRenderer',
      `MultiFactorAuthentication.${action}`,
      _LogType.Event,
      {
        'webpartAlias': this.props.webpartAlias
      }
    );
    _EngagementLogger.logEventWithLogEntry(logEntry);
  }
}