﻿/**
 * @copyright Microsoft Corporation. All rights reserved.
 *
 * @file Pagination.tsx
 */

import * as React from 'react';
import { autobind, css } from 'office-ui-fabric-react';

import { IPaginationProps } from './IPaginationProps';
import { IPageProps, PageType } from './IPageProps';
import Page from './Page';

import styles from './Pagination.module.scss';
import strings from './Strings.resx';

export default class Pagination extends React.PureComponent<IPaginationProps, {}> {

  public render(): JSX.Element {
    const paginationClassName: string = css(
      styles.Pagination,
      { [styles.noPageNum]: !this.props.showPageNum },
      this.props.className
    );

    return (
      this.props.itemsCount === 0 ? undefined : (
        <div className={paginationClassName}>
          {this.pages}
        </div>
      )
    );
  }

  private get pages(): React.ReactElement<IPageProps>[] {
    const pages: React.ReactElement<IPageProps>[] = [];
    const totalPages: number = Math.ceil(this.props.itemsCount / this.props.itemsPerPage);
    const pageLength: number = this.props.pageLength;
    let currentPage: number = this.props.currentPage;
    if (currentPage < 1) { currentPage = 1; }
    if (currentPage > totalPages) { currentPage = totalPages; }

    let startPage: number = Math.max(1, currentPage - Math.floor(pageLength / 2));
    let endPage: number = Math.min(totalPages, currentPage + Math.floor(pageLength / 2));
    if (endPage - startPage + 1 < pageLength) {
      if (currentPage < (totalPages / 2)) {
        endPage = Math.min(totalPages, pageLength + startPage);
      } else {
        startPage = Math.max(1, endPage - pageLength);
      }
    }
    if (endPage - startPage + 1 > pageLength) {
      if (currentPage > (totalPages / 2)) {
        startPage++;
      } else {
        endPage--;
      }
    }
    const prevPageNum: number = currentPage - 1 > 0 ? currentPage - 1 : currentPage;
    const nextPageNum: number = currentPage === endPage ? endPage : currentPage + 1;

    // Build Prev Page, disable it if current page number is 1
    pages.push(
      <Page
        className={this.props.pageClassName}
        key={strings.PaginationPrev + prevPageNum}
        text={strings.PaginationPrev}
        pageType={PageType.Prev}
        pageNumber={this.props.showPageNum ? prevPageNum : undefined}
        selected={false}
        disabled={this.props.showPageNum ? currentPage === 1 : !this.props.hasPrevPage}
        onClick={this._handleClick}
      />
    );

    // Build Default Page
    if (this.props.showPageNum) {
      for (let pageNum: number = startPage; pageNum <= endPage; pageNum++) {
        const pageTxt: string = `${pageNum}`;
        pages.push(
          <Page
            key={pageTxt + pageNum}
            text={pageTxt}
            pageType={PageType.Default}
            pageNumber={pageNum}
            selected={currentPage === pageNum}
            disabled={false}
            onClick={this._handleClick}
          />
        );
      }
    }

    // Build Next Page, disable it if current page number is the endPage
    pages.push(
      <Page
        key={strings.PaginationNext + nextPageNum}
        text={strings.PaginationNext}
        pageType={PageType.Next}
        pageNumber={this.props.showPageNum ? nextPageNum : undefined}
        selected={false}
        disabled={this.props.showPageNum ? currentPage === endPage : !this.props.hasNextPage}
        onClick={this._handleClick}
      />
    );

    return pages;
  }

  @autobind
  private _handleClick(pageNumber: number, pageType: PageType): void {
    this.props.onPageChange(pageNumber, pageType);
  }
}
