/**
 * @copyright Microsoft Corporation. All rights reserved.
 *
 * @file Parallax effect utility.  Accepts a target element and sets a translate3d
 * on the target element once the container is scrolled beyond the top bound of the
 * scrollable parent viewport.
 */

import { Validate } from '@microsoft/sp-core-library';
import { autobind } from 'office-ui-fabric-react';
import { _ScrollEventHandler } from '@microsoft/sp-webpart-base';

export default class Parallax {
  /**
   * Singleton instance of Parallax utility
   */
  private static _instance: Parallax;

  private readonly _parallaxElements: Set<HTMLElement> = new Set<HTMLElement>();
  private readonly _parallaxFactor: number = 1.81; // Parallax ratio, taken from example parallax site

  public static get instance(): Parallax {
    if (Parallax._instance === undefined) {
      Parallax._instance = new Parallax();
    }

    return Parallax._instance;
  }

  /**
   * Applies parallax effect onto the targetElement
   * @param targetElement
   */
  public addParallax(targetElement: HTMLElement): void {
    Validate.isNotNullOrUndefined(targetElement, 'targetElement');

    _ScrollEventHandler.instance.register(targetElement, this._onScroll);
    this._parallaxElements.add(targetElement);
    targetElement.style.setProperty('will-change', 'transform');
  }

  /**
   * Removes parallax effect from the targetElement
   * @param targetElement
   */
  public removeParallax(targetElement: HTMLElement): void {
    Validate.isNotNullOrUndefined(targetElement, 'targetElement');
    if (this._parallaxElements.has(targetElement)) {
      this._parallaxElements.delete(targetElement);
      this._setTranslate(targetElement, 0);
      targetElement.style.removeProperty('will-change');
    }

    if (this._parallaxElements.size === 0) {
      _ScrollEventHandler.instance.unregister(targetElement, this._onScroll);
    }
  }

  private constructor() {
  }

  @autobind
  private _onScroll(event: Event, clientRect: ClientRect): void {
    window.requestAnimationFrame(this._updateParallax.bind(this, clientRect));
  }

  @autobind
  private _updateParallax(clientRect: ClientRect): void {
    const scrollRegionTop: number = clientRect.top;
    this._parallaxElements.forEach((ele) => {
      if (ele.parentElement) {
        const { top, bottom }: { top: number, bottom: number } = ele.parentElement.getBoundingClientRect();
        if (top < scrollRegionTop && bottom > scrollRegionTop) {
          const parallaxDistance: number = Math.round((scrollRegionTop - top) / this._parallaxFactor);
          this._setTranslate(ele, parallaxDistance);
        } else if (top >= scrollRegionTop) {
          this._setTranslate(ele, 0);
        }
      }
    });
  }

  private _setTranslate(elm: HTMLElement, topPosition: number): void {
    const translate: string = 'translate3d(0px,' + topPosition + 'px, 0px)';
    elm.style.transform = translate;
  }
}