/**
 * @Copyright (c) Microsoft Corporation.  All rights reserved.
 *
 * @file Common web part placeholder component.
 *
 */

import { sealed } from '@microsoft/decorators';
import * as React from 'react';
import { IPlaceholderButtonProps, IPlaceholderProps } from './IPlaceholderProps';
import { autobind, BaseComponent, Icon } from 'office-ui-fabric-react';
import { IButton, PrimaryButton } from 'office-ui-fabric-react/lib/Button';
import { css } from 'office-ui-fabric-react/lib/Utilities';
import { BrowserUtilities } from '@ms/sp-component-utilities';

import { default as TextTruncate } from './../text/TextTruncate';
import './Placeholder.scss';

const MAX_DESCRIPTION_LINES: number = 3;

/**
 * Component used to place holder user experience in the web part display area.
 *
 * @internal
 */
/* tslint:disable:no-any */
@sealed
export default class Placeholder extends BaseComponent<IPlaceholderProps, {}> {
/* tslint:enable:no-any */
  private _placeholderButton: IButton;
  private _descriptionId: string;
  private _placeholderDescription?: HTMLElement;

  @autobind
  public componentWillMount(): void {
    this._processLines = this._async.throttle(this._processLines, 500);
  }

  public componentDidMount(): void {
    this._processLines();
    window.addEventListener('resize', this._handleWindowResize);
  }

  public componentDidUpdate(prevProps: IPlaceholderProps): void {
    // Since truncation can be lengthy, instead of complex shouldComponentUpdate
    // just simplify to width change
    if (this.props.description && this.props.webpartWidth !== prevProps.webpartWidth) {
      this._processLines();
    }
  }

  public componentWillUnmount(): void {
    window.removeEventListener('resize', this._handleWindowResize);
  }

  /**
   * Render the place holder user experience.
   */
  public render(): React.ReactElement<IPlaceholderProps> {
    const { children, contentClassName, description, icon, iconText } = this.props;
    this._descriptionId = `description-${Math.random().toString()}`;

    return (
      <div className={ css('Placeholder', contentClassName,
      { 'Placeholder--autoHeight': !!children }) }>
        <div className='Placeholder-container ms-Grid'>
          <div className='Placeholder-head ms-Grid-row'>
            <div className='Placeholder-headContainer'>
              <Icon
                className='Placeholder-icon'
                iconName={ icon }
              />
              <span className='Placeholder-text'>{ iconText }</span>
            </div>
          </div>
          <div className='Placeholder-description ms-Grid-row'>
            <span className='Placeholder-descriptionText'
              id={ this._descriptionId }
              ref={ this._resolveRef('_placeholderDescription') }>
              { description }
            </span>
          </div>
          { children }
          <div className='Placeholder-description ms-Grid-row'>
            { this.buttonElement }
            { this._extraButtonElements }
          </div>
        </div>
      </div>
    );
  }

  /**
   * Render the button element.
   */
  public get buttonElement(): React.ReactElement<React.HTMLProps<HTMLDivElement>>{
    return this.props.buttonLabel
      && this._renderButtonElement({
        description: this.props.description,
        label: this.props.buttonLabel,
        onClick: this.props.onAdd
      }, true);
  }

  /**
   * Focus the button element.
   */
  public focusOnPlaceHolderButton(): void {
    if (this._placeholderButton) {
      this._placeholderButton.focus();
    }
  }

  /**
   * If the placeholder contains multiple buttons, render these buttons element.
   */
  private get _extraButtonElements(): React.ReactElement<React.HTMLProps<HTMLDivElement>>[] {
    return this.props.extraButtons
      && this.props.extraButtons.map((buttonProps, index) => this._renderButtonElement(buttonProps, false));
  }

  @autobind
  private _handleWindowResize(): void {
    this._processLines();
  }

  private _processLines(): void {
    if (
      this._placeholderDescription &&
      // Don't truncate in the native app which has less affordances for seeing full instructions
      !BrowserUtilities.isWebViewHosted()) {
      TextTruncate.truncateLines(
        this._placeholderDescription,
        this.props.description,
        MAX_DESCRIPTION_LINES
      );
    }
  }

  private _renderButtonElement(buttonProps: IPlaceholderButtonProps, isDefaultButton: boolean):
    React.ReactElement<React.HTMLProps<HTMLDivElement>> {
    return (
      <PrimaryButton
        componentRef={ isDefaultButton && this._resolveRef('_placeholderButton') }
        onClick={ buttonProps.onClick }
        ariaDescription={ buttonProps.description }>
        { buttonProps.label }
      </PrimaryButton>
    );
  }
}