/**
 * @file ResponsiveUI.ts
 * @Copyright (c) Microsoft Corporation.  All rights reserved.
 */

export enum BREAKPOINT_SIZE {
    SM,
    MD,
    LG,
    XL,
    XXL,
    XXXL
}

export class ResponsiveUI {
  /**
   * This is used to pass in the screen or container size breakpoints fot responsive mode.
   * The entries should always be set in ascending order.
   * Example:
   *
   * let responsiveUI: ResponsiveUI = new ResponsiveUI();
   * responsiveUI.screenBreakpoints:  = {
   *  [480]: BREAKPOINT_SIZE.SM,
   *  [640]: BREAKPOINT_SIZE.MD,
   *  [1024]: BREAKPOINT_SIZE.LG,
   *  [1366]: BREAKPOINT_SIZE.XL,
   *  [1920]: BREAKPOINT_SIZE.XXL,
   *  [99999]: BREAKPOINT_SIZE.XXXL
   * };
  */
  public screenBreakpoints: { [key: number]: BREAKPOINT_SIZE } = undefined;

  /**
   * Returns the current responsive mode screen size based on the screen width.
  */
  public getCurrentScreenBreakpoint(): BREAKPOINT_SIZE {
    const _width: number = window.innerWidth;

    return this.getCurrentElementBreakpoint(_width);
  }

  /**
   * Returns the responsive mode element size based on the given width.
  */
  public getCurrentElementBreakpoint(width: number): BREAKPOINT_SIZE {
    if (this.screenBreakpoints !== undefined) {
      for (const key in this.screenBreakpoints) {
        if (parseInt(key, 10) >= width) {
          return this.screenBreakpoints[key];
        }
      }
    }

    return BREAKPOINT_SIZE.SM;
  }

}