import * as React from 'react';
import { _QosMonitor } from '@ms/sp-telemetry';
import { css, autobind, BaseComponent, IBaseProps, Icon } from 'office-ui-fabric-react';
import { ServiceScope } from '@microsoft/sp-core-library';
import { PageContext } from '@microsoft/sp-page-context';
import styles from './LocalFilePreview.module.scss';
import { spConstants } from '@ms/sp-component-utilities';
import { Uri } from '@ms/odsp-utilities-bundle';
import SPFileHandler, { IFileUploadResponse } from './SPFileHandler';
import strings from './loc/Strings.resx';

export interface ILocalFilePreviewProps extends IBaseProps {
  serviceScope: ServiceScope;
  pageContext: PageContext;
  localFile: File | Blob;
  QosTag: string;
  uploadingMessage: string;
  isBackgroundImage?: boolean;
  updateImageSource?: (newImageSource: string) => void;
  onUploadComplete: (response: IFileUploadResponse, fileName?: string) => void;
  onUploadError: (error: Error) => void;
  externalFile?: IFilePickerExternalImagesToUpload;
}

// copy from IFilePickerResponse (todo: take this out, when uploading logic moves into file picker anyway)
export interface IFilePickerExternalImagesToUpload {
  src: string;
  id: string;
  name: string;
}

export interface ILocalFilePreviewState {
  imageDataString: string;
}

export default class LocalFilePreview extends BaseComponent<ILocalFilePreviewProps, ILocalFilePreviewState> {
  private _spFileHandler: SPFileHandler;

  constructor(props: ILocalFilePreviewProps) {
    super(props);
    this._spFileHandler = new SPFileHandler(props.serviceScope);
    this.state = {
      imageDataString: undefined
    };
  }

  public componentWillMount(): void {
    if (this.props.externalFile) {
      this._startUpload(this.props.externalFile);
    } else if (this.props.localFile) {
      this._startUpload(this.props.localFile);
    }
  }

  public componentDidMount(): void {
    this._setStateForLocalFile();
  }

  public render(): JSX.Element {
    const imagePreview: JSX.Element =
      <img className={styles.imagePreview}
        alt={strings.PreviewAltTextWhileUploading}
        src={this.state.imageDataString}
        ref={this._resolveRef('_imagePreview')}
      />;
    const isBackground: boolean = this.state.imageDataString && this.props.isBackgroundImage;
    const isInline: boolean = this.state.imageDataString && !this.props.isBackgroundImage;

    return (
      <div className={styles.localFilePreview}
        style={isBackground ? {
          backgroundImage: `url(${this.state.imageDataString})`
        } : undefined}
        ref={this._resolveRef('_root')}>
        {isInline && imagePreview}
        <div className={styles.uploadingIcon}>
          <Icon iconName='Sync' />
          <span className={css('ms-font-xl', 'ms-fontWeight-semilight', 'ms-fontColor-white')}
            aria-live='assertive'
            aria-atomic={true}>
            {this.props.uploadingMessage}
          </span>
        </div>
      </div>
    );
  }

  private _setStateForLocalFile(): void {
    if (this.props.localFile) {
      const readingPromise: Promise<string> = this._spFileHandler.readFile(this.props.localFile);
      readingPromise.then((src: string) => {
        this.setState({
          imageDataString: src
        });
      });
    } else if (this.props.externalFile) {
      this.setState({
        imageDataString: this.props.externalFile.src
      });
    }
  }

  private _composeImageSourceUrl(fileName: string, webServerRelativeUrl: string, overridenPageName?: string): string {
    const pageName: string = this._spFileHandler.getPageName(
      overridenPageName || SPFileHandler.pagePathName);
    // i.e. "teams/clientsidewebparts/SiteAssets/SitePages/workbench-cm/40031-.pixeltheaussie.JPG";
    return Uri.concatenate(webServerRelativeUrl, `/SiteAssets/SitePages/${pageName}/${fileName}`);
  }

  /* tslint:disable:no-any */
  /* We need this 'any' because Event.returnValue is defined */
  /* as a boolean in TypeScript, but for the beforeUnload    */
  /* event this is defined as a string, which we are using   */
  /* here. We won't change TypeScript, so we do this.        */
  private _promptUploadingWarning(e: any): string {
    /* tslint:enable:no-any */
    const confirmationMessage: string = strings.UploadWarningMessage;
    e.returnValue = confirmationMessage;     // Gecko, Trident, Chrome 34+
    return confirmationMessage;              // Gecko, WebKit, Chrome <34
  }

  /**
   * Defines what is supposed to happen before the upload start, and set the callback
   * when upload process is finished
   */
  @autobind
  private _startUpload(file: File | Blob | IFilePickerExternalImagesToUpload): void {
    const imageUploadMonitor: _QosMonitor = new _QosMonitor(this.props.QosTag);
    try {
      let isValidFileType: boolean = false;
      let fileName: string;
      const prefix: string = Math.floor(Math.random() * 100000).toString();
      const isFileExternalImage: boolean = 'src' in file;

      if (file instanceof File &&
        this._spFileHandler.validateFileType(file, spConstants.imageExtensions)) {
        // Validate the file type and use file.name as fileName
        isValidFileType = true;
        fileName = `${prefix}-${SPFileHandler.replaceSpecialCharacters(file.name)}`;
      } else if (file instanceof Blob &&
        this._spFileHandler.validateMimeType(file, spConstants.imageExtensions)) {
        // Validate the mime type and use file.type and create a new filename
        isValidFileType = true;
        fileName = `${prefix}.${file.type.split('/').pop()}`;
      } else if (isFileExternalImage && this._spFileHandler.validateFileType(
        (file as IFilePickerExternalImagesToUpload).src, spConstants.imageExtensions)) {
        file = file as IFilePickerExternalImagesToUpload;
        isValidFileType = true;
        fileName = `${prefix}-${SPFileHandler.replaceSpecialCharacters(file.name)}`;
      }

      if (!isValidFileType) {
        this._onUploadError(new Error('fileTypeError'));
        imageUploadMonitor.writeExpectedFailure('InvalidFileType');
        return;
      }

      // pre-set the image source url once upload starts
      // in case user switch to preview mode before upload finishes
      const composedImageSource: string = this._composeImageSourceUrl(
        fileName,
        (this.props.pageContext && this.props.pageContext.web && this.props.pageContext.web.serverRelativeUrl) || '/'
      );

      if (this.props.updateImageSource) {
        this.props.updateImageSource(composedImageSource);
      }

      window.addEventListener('beforeunload', this._promptUploadingWarning);

      this._spFileHandler.uploadFile(
        file instanceof File ? file as File : (file instanceof Blob ? file as Blob : undefined) /* file */,
        isFileExternalImage ? (file as IFilePickerExternalImagesToUpload).src : undefined /* fileSrc */,
        fileName /* overriddenFileName */)
      .then((response: IFileUploadResponse) => {
        this._onUploadComplete(response, (file instanceof File ) && file.name);
        imageUploadMonitor.writeSuccess();
      }).catch((error: Error) => {
        imageUploadMonitor.writeUnexpectedFailure('UploadPromise', error);
        this._onUploadError(error);
      });
    } catch (exception) {
      imageUploadMonitor.writeUnexpectedFailure('CatchAll', exception);
    }
  }

  /**
   *  Defines what is supposed to happen when the upload process is complete,
   *  including clearing the mask/monitor/error msg
   */
  @autobind
  private _onUploadComplete(response: IFileUploadResponse, fileName?: string): void {
    window.removeEventListener('beforeunload', this._promptUploadingWarning);
    this.props.onUploadComplete(response, fileName);
  }

  @autobind
  private _onUploadError(error: Error): void {
    window.removeEventListener('beforeunload', this._promptUploadingWarning);
    if (this.props.updateImageSource) {
      this.props.updateImageSource(undefined);
    }
    this.props.onUploadError(error);
  }

}