﻿/**
 * @Copyright (c) Microsoft Corporation.  All rights reserved.
 *
 * @file SPFileHandler.ts
 */

import {
  _LogSource,
  _TraceLogger
} from '@ms/sp-telemetry';
import {
  ServiceScope,
  UrlUtilities
} from '@microsoft/sp-core-library';
import {
  SPHttpClient
} from '@microsoft/sp-http';
import { PageContext } from '@microsoft/sp-page-context';
import { Uri, UriEncoding } from '@ms/odsp-utilities-bundle';

export interface IFileUploadResponse {
  /**
   * File upload post request response with selected fields
   * necessary for storing the file and showing the thumbnail
   */
  ListId: string; // the main field where $select is required. Update getEndPoint for more fields
  ServerRelativeUrl: string;
  Title: string;
  UniqueId: string;
  // There's also a possible 'odata.error' for errors
}

export default class SPFileHandler {
  private static _logSource: _LogSource = new _LogSource('PublishingWebPartSPFileHandler');
  private _containerUrl: string;
  private _httpClient: SPHttpClient;
  private _pageContext: PageContext;
  private _reader: FileReader;

  /**
   * Leaf URL portion of the page as serverRequestPath doesn't change within the session.
   * Note that page renames don't chang the pathname after the page title was first entered.
   */
  public static get pagePathName(): string {
    const isNewPage: boolean = window.location.pathname.toUpperCase().indexOf('/NEWPAGE.ASPX') !== -1;
    if (isNewPage) {
      _TraceLogger.logError(this._logSource, new Error('Page media upload before page titled.'));
    }
    return window.location.pathname;
  }

  /**
   * Get a Url safe file name without incompatible special characters
   */
  public static replaceSpecialCharacters(inputString: string): string {
    // TODO: #431082- @CMalonzo to cleanup.
    // This must NEVER include the dot or it will interfere with extension filenames!!!
    const specialChars: RegExp = new RegExp('[&\/\\# ,+()$~\'":*?<>{}|]|%2.', 'g');
    // If file is set by JS code for TAB test, normally this inputString is undefined.
    return inputString ? inputString.replace(specialChars, '-') : '';
  }

  /**
   * Without %27 and params encoded like this, query will fail on the server because of overencoding with slashes
   * disappearing
   * @param key
   * @param value
   */
  private static _encodeQuerySegment(key: string, value: string): string {
    return `${key}=%27${UriEncoding.encodeRestUriStringToken(value)}%27`;
  }

  /**
   * Change the container in the upload end point url
   */
  public set containerUrl(containerUrl: string) {
    this._containerUrl = containerUrl;
  }

  public constructor(serviceScope: ServiceScope, containerUrl?: string) {
    serviceScope.whenFinished(() => {
      this._httpClient = serviceScope.consume(SPHttpClient.serviceKey);
      this._pageContext = serviceScope.consume(PageContext.serviceKey);
    });

    this._containerUrl = containerUrl;
  }

  /**
   * Read a file as dataURL for display, then this method to define the action
   * on completion of the reading
   */
  public readFile(file: File | Blob): Promise<string> {
    if (!this._reader) {
      this._reader = new FileReader();
    }

    const readPromise: Promise<string> = new Promise<string>((resolve, reject) => {
      if (!file) {
        _TraceLogger.logVerbose(SPFileHandler._logSource, 'File is undefined');
        reject(new RangeError('FileUndefined'));
      }
      this._reader.onloadend = () => resolve(this._reader.result);
      this._reader.onerror = (e: ErrorEvent) => {
        _TraceLogger.logVerbose(SPFileHandler._logSource, 'Error occurred when reading the file');
        reject(e);
      };
      this._reader.readAsDataURL(file);
    });
    return readPromise;
  }

  /**
   * Upload a single file to pre-defined end point, with an optionally specified name
   * returns a promise to define the actions on completion of each file upload
   * @param {File} file - the single file to be uploaded
   * @param {string} overridenFileName - the name to override the original file name, if given
   * @param {string} overridenFileName - the name to override the page name, if given
   */
  public uploadFile(
    file: File | Blob,
    fileSrc?: string,
    overriddenFileName?: string,
    overriddenPageName?: string
  ): Promise<IFileUploadResponse> {
    const headers: Headers = new Headers();
    const webAbsoluteUrl: string = this._pageContext &&
      this._pageContext.web && this._pageContext.web.absoluteUrl || '';

    headers.append('accept', 'application/json; odata=nometadata');
    headers.append('OData-version', '3.0');

    const fileName: string | undefined = (file as any).name; // tslint:disable-line:no-any
    if (!overriddenFileName && fileName) {
      overriddenFileName = fileName;
    }

    const fullEndPoint: string = this._getEndPoint(overriddenFileName, fileSrc, overriddenPageName);
    const postUrl: string = Uri.concatenate(webAbsoluteUrl, fullEndPoint);

    return this._httpClient.post(
      postUrl,
      SPHttpClient.configurations.v1,
      file ? {
        headers: headers,
        body: file
      } : {
        headers: headers
      }
    ).then((response: Response) => response.json())
     .then((uploadResponse: IFileUploadResponse) => {
        const errorObjectKey: string = 'odata.error';
        if (uploadResponse[errorObjectKey]) {
          const error: any = uploadResponse[errorObjectKey]; // tslint:disable-line:no-any
          throw new Error(error.message && error.message.value || error.code);
        } else {
          return uploadResponse;
        }
      })
    .catch((error: Error) => {
      throw error;
    });
  }

  /**
   * Validate the file type with a list of extensions
   */

  public validateFileType(file: File | string, fileExtensions: ReadonlyArray<string>): boolean {
    const fileName: string = file instanceof File ? file.name : file;
    const dotIndex: number = fileName.lastIndexOf('.');
    let ext: string;
    if (dotIndex === -1) {
      return false;
    } else {
      ext = fileName.substring(dotIndex + 1, fileName.length);
    }
    return fileExtensions.indexOf(ext.toLowerCase()) > -1;
  }

  /**
   * Validate the mime type with a list of extensions
   */

  public validateMimeType(file: Blob, fileExtensions: ReadonlyArray<string>): boolean {
    if (file.type) {
      const mimeType: string = file.type.split('/').pop();
      if (mimeType) {
        return fileExtensions.indexOf(mimeType.toLowerCase()) > -1;
      }
    }
    return false;
  }

  /**
   * Get the current page name without extension and special characters
   */
  public getPageName(pathname: string): string {
    const pageTitle: string = this.removeExtension(pathname);
    const pageName: string = SPFileHandler.replaceSpecialCharacters(pageTitle);
    return pageName;
  }

  /**
   * Remove the extesion from a file name
   */
  public removeExtension(inputString: string): string {
    const lastSlashIndex: number = inputString.lastIndexOf('/');
    const rawTitle: string = lastSlashIndex === -1 ?
      inputString : inputString.substring(lastSlashIndex + 1);
    const lastDotIndex: number = rawTitle.lastIndexOf('.');
    return lastDotIndex === -1 ? rawTitle : rawTitle.substring(0, lastDotIndex);
  }

  /**
   * Get the full end point url using the file name and container
   */
  private _getEndPoint(
    fileName: string,
    fileSrc?: string,
    overriddenPageName?: string
  ): string {
    fileName = encodeURIComponent(fileName);
    const select: string = '$select=ListId,ServerRelativeUrl,Title,UniqueId';

    if (this._containerUrl) {
      const containerUrl: string = encodeURIComponent(this._containerUrl);
      return '/_api/web/GetFolderByServerRelativeUrl(' +
        UrlUtilities.convertToODataStringLiteral(containerUrl) + ')/Files/Add(url=' +
        UrlUtilities.convertToODataStringLiteral(fileName) + `, overwrite=false)?${select}`;
    }

    if (fileSrc) {
      return '/_api/sitepages/AddImageFromExternalUrl?' +
        this._createSitePageEndpoint(fileName, overriddenPageName) + '&' +
        SPFileHandler._encodeQuerySegment('externalUrl', fileSrc) + `&${select}`;
    }

    return '/_api/sitepages/AddImage?' + this._createSitePageEndpoint(fileName, overriddenPageName) + `&${select}`;
  }

  private _createSitePageEndpoint(encodedFileName: string, overriddenPageName: string): string {
    const pageName: string = overriddenPageName || this.getPageName(SPFileHandler.pagePathName);
    return SPFileHandler._encodeQuerySegment('imageFileName', encodedFileName) + '&' +
      SPFileHandler._encodeQuerySegment('pageName', pageName);
  }
}
