/**
 * @copyright Microsoft Corporation. All rights reserved.
 *
 * @file TextUtility. helps with multiline ellipsis
 */
import { Text } from '@microsoft/sp-core-library';
import strings from './loc/Strings.resx';

export default class TextTruncate {
  public static truncateLines(
    element: HTMLElement,
    origText: string,
    maxLines: number = 3
  ): void {

    if (!origText || !element) {
      return;
    }

    maxLines = Math.round(maxLines);

    element.textContent = origText;

    const lineHeight: string = window.getComputedStyle(element, undefined).getPropertyValue('line-height');
    const maxHeight: number = maxLines * Math.round(parseFloat(lineHeight));
    const origTextWords: string[] = origText.split(' ');
    let currHeight: number = this._getElementContentHeight(element);

    if (currHeight > maxHeight && maxLines > 0) {
      let newTextWords: string[] = [];
      let newText: string = '';

      for (let i: number = origTextWords.length; i >= 0 && currHeight > maxHeight; i--) {
        newTextWords = origTextWords.slice(0, i);
        newText = newTextWords.join(' ');

        if (!newText) {
          const breakLastWord: boolean = i === 0 && currHeight > maxHeight;
          if (breakLastWord) {
            TextTruncate._breakSingleWord(element, origTextWords[0], currHeight, maxHeight);
          }

          break;
        }

        TextTruncate._localizeTruncation(element, newText);
        currHeight = this._getElementContentHeight(element);
      }

      const addTooltip: boolean = !element.title && newText !== element.textContent;
      if (addTooltip) {
        element.title = origText;
      }
    }
  }

  public static numberOfLines(element: HTMLElement): number {
    const currentRectHeight: number = this._getElementContentHeight(element);
    const currentComputedStyle: CSSStyleDeclaration = document.defaultView.getComputedStyle(element, undefined);

    return Math.round(currentRectHeight / parseInt(currentComputedStyle.lineHeight, 10));
  }

  private static _getElementContentHeight(element: HTMLElement): number {
    const styles: CSSStyleDeclaration = window.getComputedStyle(element);
    const padding: number = parseFloat(styles.paddingTop) + parseFloat(styles.paddingBottom);

    return element.offsetHeight - padding;
  }

  private static _breakSingleWord(
    element: HTMLElement,
    word: string,
    currHeight: number,
    maxHeight: number
    ): void {

    let newText: string = '';
    element.style.wordWrap = 'break-word'; // this property is also known as 'overflow-wrap'

    for (let i: number = word.length; i >= 0 && currHeight > maxHeight; i--) {
      newText = word.slice(0, i);

      if (!newText) {
        break;
      }

      TextTruncate._localizeTruncation(element, newText);
      currHeight = this._getElementContentHeight(element);
    }
  }

  private static _localizeTruncation(element: HTMLElement, shortText: string): void {
    element.textContent = Text.format(strings.TruncationTemplate, shortText);
  }
}