import * as React from 'react';

export enum SwipeDirection {
  Up,
  Down,
  Left,
  Right
}

export interface ITouchAreaProps extends React.Props<TouchArea> {
  onSwipe?: (direction: SwipeDirection) => void;
  preventDefault?: boolean;
}

export default class TouchArea extends React.Component<ITouchAreaProps, {}> {

  private _xDown: number;
  private _yDown: number;

  public constructor() {
    super();
    this._onTouchStart = this._onTouchStart.bind(this);
    this._onTouchMove = this._onTouchMove.bind(this);
    this._onMouseDown = this._onMouseDown.bind(this);
    this._onMouseUp = this._onMouseUp.bind(this);
    this._preventDefault = this._preventDefault.bind(this);
    this._calculateSwipeEvent = this._calculateSwipeEvent.bind(this);
  }

  public render(): JSX.Element {
    return (
      /* tslint:disable-next-line:react-a11y-event-has-role */
      <div
        onTouchStart={ this._onTouchStart }
        onMouseDown={ this._onMouseDown }
        onMouseUp={ this._onMouseUp }
        onTouchMove={ this._onTouchMove }>
        { this.props.children }
      </div>);
  }

  /* tslint:disable-next-line:no-any */
  private _onMouseDown(evt: React.MouseEvent<any>): void {
    this._preventDefault(evt);
    this._xDown = evt.clientX;
    this._yDown = evt.clientY;
  }

  /* tslint:disable-next-line:no-any */
  private _onTouchStart(evt: React.TouchEvent<any>): void {
    this._preventDefault(evt);
    if (evt.touches.length < 1) {
      return;
    }

    this._xDown = evt.touches[0].clientX;
    this._yDown = evt.touches[0].clientY;
  }

  /* tslint:disable-next-line:no-any */
  private _onMouseUp(evt: React.MouseEvent<any>): void {
    this._preventDefault(evt);
    this._calculateSwipeEvent(evt.clientX, evt.clientY);
  }

  /* tslint:disable-next-line:no-any */
  private _onTouchMove(evt: React.TouchEvent<any>): void {
    this._preventDefault(evt);
    if (evt.touches.length < 1) {
      return;
    }

    this._calculateSwipeEvent(evt.touches[0].clientX, evt.touches[0].clientY);
  }

  /* tslint:disable-next-line:no-any */
  private _preventDefault(evt: React.SyntheticEvent<any>): void {
    if (this.props.preventDefault === true) {
      evt.preventDefault();
    }
  }

  private _calculateSwipeEvent(clientX: number, clientY: number): void {
    if (!this._xDown || !this._yDown) {
      return;
    }

    if (!this.props.onSwipe) {
      return;
    }

    const xDiff: number = this._xDown - clientX;
    const yDiff: number = this._yDown - clientY;

    if (Math.abs(xDiff) > Math.abs(yDiff)) {
      if (xDiff > 0) {
        this.props.onSwipe(SwipeDirection.Right);
      } else {
        this.props.onSwipe(SwipeDirection.Left);
      }
    } else {
      if (yDiff > 0) {
        this.props.onSwipe(SwipeDirection.Up);
      } else {
        this.props.onSwipe(SwipeDirection.Down);
      }
    }

    this._xDown = undefined;
    this._yDown = undefined;
  }

}