﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Net;
using System.Security.Cryptography;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using System.Xml.Linq;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;

namespace ClientAppProvisioner
{
    public partial class MainForm : Form
    {
        WorkbenchSettings workbenchSettings = new WorkbenchSettings();

        public MainForm()
        {
            InitializeComponent();
            UpdateUI();
        }

        protected override void OnShown(EventArgs e)
        {
            base.OnShown(e);

            try
            {
                this.LoadSettings();
            }
            catch (Exception ex)
            {
                Program.ShowApplicationException(ex);
            }
        }

        private void btnProvisionSite_Click(object sender, EventArgs e)
        {
            this.SaveSettings();

            string clientSideFieldName = chkLegacyMode.Checked ? "ClientAppPageId" : "ClientSideApplicationId";
            string clientSidePageClass = chkLegacyMode.Checked ? "ClientRenderedPage" : "ClientSidePage";

            txtResult.Text = "";
            try
            {
                LogResult("Checking " + clientSideFieldName + " column on Pages library...");

                string checkUrl = "/_api/web/lists/GetByTitle('Pages')/Fields"
                    + "?$select=Id,Title,InternalName"
                    + "&$filter=InternalName eq '" + clientSideFieldName + "'";

                string checkResultJson = GetRequest(checkUrl);
                JObject checkResult = JsonConvert.DeserializeObject<JObject>(checkResultJson);
                JArray fieldValueArray = (JArray)checkResult["value"];
                if (fieldValueArray.Count == 1)
                {
                    LogResult("The column already exists.");
                }
                else
                {
                    LogResult("Creating " + clientSideFieldName + " column on Pages library...");

                    // http://sharepoint.stackexchange.com/questions/115958/is-it-possible-to-create-field-with-defined-internal-name-using-rest-api
                    var createFieldBody = new CreateFieldBodyJson();
                    createFieldBody.FieldTypeKind = CreateFieldBodyJson.FieldType.Text;
                    createFieldBody.Title = clientSideFieldName;
                    createFieldBody.SchemaXml = String.Format(@"
<Field
    Type=""Text""
    Name=""{0}""
    Description=""The Client Rendered App Page definition""
    DisplayName=""{0}""
    StaticName=""{0}""
    InternalName=""{0}""
    Hidden=""FALSE""
    Required=""TRUE""
    Sealed=""FALSE""
    ShowInDisplayForm=""TRUE""
    ShowInEditForm=""TRUE""
    ShowInListSettings=""TRUE""
    ShowInNewForm=""TRUE"">
</Field>
", clientSideFieldName);

                    string createFieldBodyJson = JsonConvert.SerializeObject(createFieldBody, Formatting.Indented);
                    Debug.WriteLine(createFieldBodyJson);

                    PostRequest("/_api/web/lists/GetByTitle('Pages')/fields", createFieldBodyJson);

                    LogResult("Successfully created " + clientSideFieldName + " column.");
                }

            }
            catch (Exception ex)
            {
                LogResult("ERROR: " + ex.Message + "\r\n");
            }

            try
            {
                string starterPageUrl = txtSiteUrl.Text + "/Pages/ClientApp.aspx";
                LogResult("\r\nUploading starter page:\r\n    " + starterPageUrl + "\r\n");

                string starterPageContent = @"<%@ Page language=""C#"" Inherits=""Microsoft.SharePoint.WebControls."
                    + clientSidePageClass
                    + @", Microsoft.SharePoint, Version=16.0.0.0, Culture=neutral, PublicKeyToken=71e9bce111e9429c"" %>";

                string addPageUrl = "/_api/web/lists/GetByTitle('Pages')/RootFolder/Files/add(url='"
                    + "ClientApp.aspx" + "',overwrite=true)?$expand=ListItemAllFields";

                string addPageResultJson = PostRequest(addPageUrl, starterPageContent);

                JObject addPageResult = JsonConvert.DeserializeObject<JObject>(addPageResultJson);
                JToken fields = addPageResult["ListItemAllFields"];
                int listItemId = fields["ID"].Value<int>();

                LogResult("Successfully created with ID=" + listItemId + "\r\n");

                LogResult("Setting page contents...");

                string setPropertiesUrl = "/_api/web/lists/GetByTitle('Pages')/Items(" + listItemId + ")";

                string appId = Guid.Parse(txtAppId.Text).ToString();

                var setPropertiersBodyMetadata = new JObject();
                setPropertiersBodyMetadata["type"] = "SP.Data.PagesItem";

                var setPropertiesBody = new JObject();
                setPropertiesBody["__metadata"] = setPropertiersBodyMetadata;
                setPropertiesBody[clientSideFieldName] = appId;
                setPropertiesBody["Title"] = "Welcome to the world of client side applications!";
                setPropertiesBody["PublishingPageContent"] = @"(Start writing your rich text here.)";

                string setPropertiesBodyJson = JsonConvert.SerializeObject(setPropertiesBody, Formatting.Indented);

                var headers = new Dictionary<string,string>();
                headers.Add("X-HTTP-Method", "MERGE");
                headers.Add("If-Match", "*");

                PostRequest(setPropertiesUrl, setPropertiesBodyJson, headers);

                LogResult("\r\nSuccess!");
                MessageBox.Show("The operation completed successfully.", "Client App Provisioner",
                    MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            catch (Exception ex)
            {
                LogResult("ERROR: " + ex.Message + "\r\n");
                Program.ShowApplicationException(ex);
            }
        }

        void LogResult(string message)
        {
            txtResult.Text += message + "\r\n";
            txtResult.Refresh();
            txtResult.SelectionStart = txtResult.Text.Length;
            txtResult.ScrollToCaret();
            this.Refresh();
        }


        /// <summary>
        /// Get form digest from Sharepoint Server
        /// </summary>
        /// <param name="uri"></param>
        /// <param name="credential"></param>
        /// <returns>Server Digest</returns>
        private static string GetFormDigest(string siteAbsoluteUrl, ICredentials credential)
        {
            string url = siteAbsoluteUrl + "/_api/contextinfo";
            HttpWebRequest requestPost = (HttpWebRequest)WebRequest.Create(url);
            requestPost.Method = "POST";
            requestPost.ContentLength = 0;
            requestPost.Accept = "application/xml;odata=verbose";
            requestPost.Headers.Add("OData-Version", "3.0");
            requestPost.Credentials = credential;
            var responsePost = (HttpWebResponse)requestPost.GetResponse();
            XDocument doc = XDocument.Load(responsePost.GetResponseStream());
            XNamespace ns_dataservices = "http://schemas.microsoft.com/ado/2007/08/dataservices";
            return doc.Descendants(ns_dataservices + "FormDigestValue").First().Value;
        }

        private string GetAutoIndented(string text)
        {
            if (jsonRegex.IsMatch(text))
            {
                try
                {
                    dynamic parsedJson = JsonConvert.DeserializeObject(text);
                    return JsonConvert.SerializeObject(parsedJson, Formatting.Indented);
                }
                catch (Exception ex)
                {
                    Debug.WriteLine("Failed to parse JSON: " + ex.Message);
                }
            }
            return text; // the input text is not JSON
        }

        static readonly Regex jsonRegex = new Regex(@"^\s*[\[{\""]", RegexOptions.Compiled);

        private void chkSpecifyCredentials_CheckedChanged(object sender, EventArgs e)
        {
            UpdateUI();
        }

        void UpdateUI()
        {
            txtLoginUser.Enabled = chkSpecifyCredentials.Checked;
            txtPassword.Enabled = chkSpecifyCredentials.Checked;
            txtResult.WordWrap = chkResponseWordWrap.Checked;
        }

        private void chkResponseWordWrap_CheckedChanged(object sender, EventArgs e)
        {
            UpdateUI();
        }

        #region Load/Save Settings

        void SaveSettings()
        {
            string filename = GetSettingsFilename();

            ReadSettingsFromControls();

            var settings = new JsonSerializerSettings();
            settings.NullValueHandling = NullValueHandling.Ignore;
            string data = JsonConvert.SerializeObject(workbenchSettings, Formatting.Indented, settings);
            using (var writer = new StreamWriter(filename))
            {
                writer.WriteLine(data);
            }
        }

        void LoadSettings()
        {
            string filename = GetSettingsFilename();

            if (!File.Exists(filename))
            {
                string templateFilename = filename + ".example";
                if (File.Exists(templateFilename))
                {
                    File.Copy(templateFilename, filename);
                }
            }

            string json;
            using (var reader = new StreamReader(filename))
            {
                json = reader.ReadToEnd();
            }
            workbenchSettings = JsonConvert.DeserializeObject<WorkbenchSettings>(json);
            this.WriteSettingsToControls();
        }

        void ReadSettingsFromControls()
        {
            workbenchSettings.SiteUrl = txtSiteUrl.Text;
            workbenchSettings.SpecifyCredentials = chkSpecifyCredentials.Checked;
            workbenchSettings.LoginUser = txtLoginUser.Text;

            workbenchSettings.EncryptedPassword = "";
            if (this.txtPassword.Text != "")
            {
                byte[] bytes0 = Encoding.UTF8.GetBytes(this.txtPassword.Text);
                byte[] bytes1 = ProtectedData.Protect(bytes0, null, DataProtectionScope.CurrentUser);
                byte[] bytes2 = ProtectedData.Protect(bytes1, null, DataProtectionScope.LocalMachine);
                workbenchSettings.EncryptedPassword = Convert.ToBase64String(bytes2);
            }

            workbenchSettings.AppId = txtAppId.Text;
        }

        void WriteSettingsToControls()
        {
            txtSiteUrl.Text = workbenchSettings.SiteUrl;
            chkSpecifyCredentials.Checked = workbenchSettings.SpecifyCredentials;
            txtLoginUser.Text = workbenchSettings.LoginUser;

            this.txtPassword.Text = "";
            if (!string.IsNullOrWhiteSpace(workbenchSettings.EncryptedPassword))
            {
                try
                {
                    byte[] bytes2 = Convert.FromBase64String(workbenchSettings.EncryptedPassword);
                    byte[] bytes1 = ProtectedData.Unprotect(bytes2, null, DataProtectionScope.LocalMachine);
                    byte[] bytes0 = ProtectedData.Unprotect(bytes1, null, DataProtectionScope.CurrentUser);
                    this.txtPassword.Text = Encoding.UTF8.GetString(bytes0);
                }
                catch (Exception ex)
                {
                    Debug.WriteLine("Failed to decrypt password: " + ex.Message);
                }
            }

            txtAppId.Text = workbenchSettings.AppId;
        }

        private static string GetSettingsFilename()
        {
            string filename = Path.Combine(Path.GetDirectoryName(Application.ExecutablePath), "WorkbenchSettings.json");
            return filename;
        }

        #endregion

        private void TextBoxSelectAllBehavior_KeyUp(object sender, KeyEventArgs e)
        {
            TextBox textBox = sender as TextBox;
            if (textBox != null)
            {
                if (e.Modifiers == Keys.Control && e.KeyCode == Keys.A)
                {
                    e.Handled = true;
                    textBox.SelectAll();
                }
            }
        }

        string PostRequest(string commandSuffix, string postBody, Dictionary<string,string> extraHeaders = null)
        {
            return PerformRequest("POST", commandSuffix,postBody, extraHeaders);
        }
        string GetRequest(string commandSuffix, Dictionary<string,string> extraHeaders = null)
        {
            return PerformRequest("GET", commandSuffix, null, extraHeaders);
        }

        string PerformRequest(string method, string commandSuffix, string postBody, Dictionary<string, string> extraHeaders)
        {
            try
            {
                Cursor.Current = Cursors.WaitCursor;

                // Normalize the slashes
                txtSiteUrl.Text = txtSiteUrl.Text.TrimEnd('/');
                string siteAbsoluteUrl = txtSiteUrl.Text;

                NetworkCredential credential = null;
                if (chkSpecifyCredentials.Checked)
                {
                    credential = new NetworkCredential(txtLoginUser.Text, txtPassword.Text);
                }

                string deployUrl = siteAbsoluteUrl + commandSuffix;

                var headers = new Dictionary<string,string>();

                if (method != "GET")
                {
                    string digest = GetFormDigest(siteAbsoluteUrl, credential);
                    headers.Add("X-RequestDigest", digest);
                }

                headers.Add("OData-Version", "3.0");
                if (extraHeaders != null)
                {
                    foreach (var pair in extraHeaders)
                    {
                        headers[pair.Key] = pair.Value;
                    }
                }


                HttpWebRequest request = (HttpWebRequest)WebRequest.Create(deployUrl);
                request.Method = method;
                request.ContentType = "application/json;odata=verbose;charset=utf-8";
                request.Accept = "application/json";

                foreach (var pair in headers)
                {
                    request.Headers.Add(pair.Key, pair.Value);
                }

                if (credential != null)
                {
                    request.Credentials = credential;
                }
                else
                {
                    request.UseDefaultCredentials = true;
                }

                if (postBody != null)
                {
                    byte[] content = Encoding.UTF8.GetBytes(postBody);
                    request.ContentLength = content.Length;
                    using (Stream reqStream = request.GetRequestStream())
                    {
                        reqStream.Write(content, 0, content.Length);
                    }
                }

                using (var response = (HttpWebResponse)request.GetResponse())
                {
                    var responseValue = string.Empty;

                    if (response.StatusCode == HttpStatusCode.NoContent)
                        return "";

                    if (response.StatusCode != HttpStatusCode.OK
                        && response.StatusCode != HttpStatusCode.Created)
                    {
                        var message = String.Format("Request failed. Received HTTP {0}", response.StatusCode);
                        throw new ApplicationException(message);
                    }

                    using (var responseStream = response.GetResponseStream())
                    {
                        if (responseStream != null)
                        {
                            using (var reader = new StreamReader(responseStream))
                            {
                                responseValue = reader.ReadToEnd();
                            }
                        }
                    }

                    return responseValue;
                }
            }
            catch (WebException ex)
            {
                string message = ex.Message;
                if (ex.Response == null)
                {
                    if (ex.InnerException != null)
                    {
                        throw ex.InnerException;
                    }
                    else
                    {
                        throw;
                    }
                }
                using (var responseStream = ex.Response.GetResponseStream())
                {
                    if (responseStream != null)
                        using (var reader = new StreamReader(responseStream))
                        {
                            string errorDetail = (reader.ReadToEnd() ?? "").Trim();
                            if (!string.IsNullOrEmpty(errorDetail))
                            {
                                message += "\r\n\r\n" + errorDetail;
                            }
                        }
                }
                throw new Exception(message, ex);
            }
            finally
            {
                Cursor.Current = Cursors.Default;
            }
        }

    }
}
