﻿using SifterKit.Controls;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace SPClientPackageExplorer
{
    public partial class MainForm : Form
    {
        DataFile dataFile;
        ListViewSorter ctlRootListSorter;
        ListViewSorter ctlTrackedListSorter;

        public MainForm()
        {
            InitializeComponent();

            ctlRootListSorter = new ListViewSorter(ctlRootList);
            ctlTrackedListSorter = new ListViewSorter(ctlTrackedList);

            this.dataFile = new DataFile();
            this.dataFile.Filename = Path.Combine(Path.GetDirectoryName(Application.ExecutablePath), "PackageData.json");

            try {
                this.dataFile.LoadFromFile();
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error loading data file:\r\n\r\n" + this.dataFile.Filename + "\r\n\r\n" + ex.Message,
                    this.Text, MessageBoxButtons.OK, MessageBoxIcon.Warning);
            }

            ReloadRootList();
            ReloadTrackedPackages();
        }

        #region Root List

        private void ctlRootList_DragOver(object sender, DragEventArgs e)
        {
            e.Effect = DragDropEffects.Copy;
        }

        private void ctlRootList_DragDrop(object sender, DragEventArgs e)
        {
            // Can only drop files, so check
            if (!e.Data.GetDataPresent(DataFormats.FileDrop))
            {
                return;
            }

            var rootFolders = this.dataFile.RootFolders;

            try
            {
                string[] folders = (string[])e.Data.GetData(DataFormats.FileDrop);
                foreach (string folder in folders)
                {
                    if (!Directory.Exists(folder))
                        throw new Exception("Expecting a folder:\r\n" + folder);
                    {
                        string packageFilename = Path.Combine(folder, "package.json");
                        if (!File.Exists(packageFilename))
                            throw new Exception("The folder does not contain a package.json file:\r\n" + folder);

                        if (!rootFolders.Any(x => x.ToUpper() == packageFilename.ToUpper())) {
                            rootFolders.Add(folder);
                        }
                    }
                }
                ReloadRootList();
            }
            catch (Exception ex)
            {
                MessageBox.Show(this, "Error: " + ex.Message, this.Text, MessageBoxButtons.OK,
                  MessageBoxIcon.Stop);
            }
        }

        private void ReloadRootList()
        {
            ctlRootList.Items.Clear();
            foreach (string rootFolder in this.dataFile.RootFolders)
            {
                var listViewItem = new ListViewItem(rootFolder);
                ctlRootList.Items.Add(listViewItem);
            }
        }

        private void ctlRootList_SelectedIndexChanged(object sender, EventArgs e)
        {
            ReloadDependencyTree();
        }

        private string GetRootListSelection()
        {
            if (ctlRootList.SelectedItems.Count >= 1)
            {
                return ctlRootList.SelectedItems[0].Text;
            }
            return null;
        }

        #endregion

        #region Dependency Tree

        private void ReloadDependencyTree()
        {
            string rootFolder = GetRootListSelection();
            ctlDependencyTree.Nodes.Clear();

            if (rootFolder != null)
            {
                PackageJsonTree packageJsonTree = new PackageJsonTree();
                packageJsonTree.LoadTree(Path.Combine(rootFolder, "package.json"));

                ReloadDependencyTreeNodes(ctlDependencyTree.Nodes, packageJsonTree.RootPackage);
            }

            UpdateDependencyTree();
        }

        private void ReloadDependencyTreeNodes(TreeNodeCollection parentNodeCollection, PackageJson packageJson)
        {
            var treeNode = new TreeNode(packageJson.Name + "@" + packageJson.Version);
            treeNode.Tag = packageJson;

            TrackedPackage trackedPackage;
            if (this.dataFile.TrackedPackageSet.TrackedPackagesByName.TryGetValue(packageJson.Name, out trackedPackage))
            {
                treeNode.ImageKey = this.GetImageKeyForPublisher(trackedPackage.Publisher);
                treeNode.SelectedImageKey = treeNode.ImageKey;

                // Is this version being tracked?
                if (!trackedPackage.UsedVersions.Contains(packageJson.Version))
                {
                    treeNode.Text = "[" + treeNode.Text + "]";
                }

                // Is it starred?
                if (trackedPackage.Starred)
                {
                    treeNode.Text = "* " + treeNode.Text;
                }
            }

            parentNodeCollection.Add(treeNode);

            foreach (var dependency in packageJson.Dependencies)
            {
                if (dependency.Resolution != null)
                {
                    ReloadDependencyTreeNodes(treeNode.Nodes, dependency.Resolution);
                }
            }
        }

        private void UpdateDependencyTree()
        {
            try
            {
                ctlDependencyTree.BeginUpdate();
                UpdateDependencyTreeInternal(ctlDependencyTree.Nodes);
            }
            finally
            {
                ctlDependencyTree.EndUpdate();
            }
        }

        private void UpdateDependencyTreeInternal(TreeNodeCollection treeNodes)
        {
            foreach (TreeNode treeNode in treeNodes)
            {
                PackageJson packageJson = treeNode.Tag as PackageJson;
                if (packageJson == null)
                    return;

                TrackedPackage trackedPackage;
                if (this.dataFile.TrackedPackageSet.TrackedPackagesByName.TryGetValue(packageJson.Name, out trackedPackage))
                {
                    treeNode.ImageKey = this.GetImageKeyForPublisher(trackedPackage.Publisher);
                    treeNode.SelectedImageKey = treeNode.ImageKey;

                    treeNode.Text = packageJson.Name;

                    // Is this version being tracked?
                    if (!trackedPackage.UsedVersions.Contains(packageJson.Version))
                    {
                        treeNode.Text = "[" + treeNode.Text + "]";
                    }

                    // Is it starred?
                    if (trackedPackage.Starred)
                    {
                        treeNode.Text = "* " + treeNode.Text;
                    }
                }

                UpdateDependencyTreeInternal(treeNode.Nodes);
            }
        }

        private PackageJson GetDependencyTreeSelection()
        {
            if (ctlDependencyTree.SelectedNode != null)
            {
                return ctlDependencyTree.SelectedNode.Tag as PackageJson;
            }
            return null;
        }

        private void ctlDependencyTree_AfterSelect(object sender, TreeViewEventArgs e)
        {
            ReloadTrackedPackages();
        }

        #endregion

        #region Tracked Packages List

        private void ReloadTrackedPackages()
        {

            var trackedPackageSet = this.dataFile.TrackedPackageSet;

            ctlTrackedList.Items.Clear();
            foreach (TrackedPackage trackedPackage in trackedPackageSet.TrackedPackages)
            {
                var listViewItem = new ListViewItem();
                listViewItem.Tag = trackedPackage;
                ctlTrackedList.Items.Add(listViewItem);
            }

            UpdateTrackedPackages();
        }

        private void UpdateTrackedPackages()
        {
            try
            {
                ctlTrackedList.BeginUpdate();

                foreach (ListViewItem listViewItem in ctlTrackedList.Items)
                {
                    TrackedPackage trackedPackage = listViewItem.Tag as TrackedPackage;
                    if (trackedPackage != null)
                    {
                        listViewItem.SubItems.Clear();

                        listViewItem.Text = trackedPackage.Name;
                        listViewItem.ImageKey = GetImageKeyForPublisher(trackedPackage.Publisher);

                        listViewItem.SubItems.Add(trackedPackage.Starred ? "*" : "");
                        listViewItem.SubItems.Add(trackedPackage.Publisher.ToString());
                        listViewItem.SubItems.Add(trackedPackage.UsedInBrowser ? "YES" : "");
                        listViewItem.SubItems.Add(trackedPackage.MaxVersion.ToString());
                        listViewItem.SubItems.Add(trackedPackage.UsedVersions.Count.ToString());
                        listViewItem.SubItems.Add(trackedPackage.ApiContract.ToString());
                        listViewItem.SubItems.Add(trackedPackage.Note);
                    }
                }
            }
            finally
            {
                ctlTrackedList.EndUpdate();
            }
        }

        #endregion

        private void btnAdd_Click(object sender, EventArgs e)
        {
            var trackedPackageSet = this.dataFile.TrackedPackageSet;

            PackageJson selectedPackageJson = GetDependencyTreeSelection();
            if (selectedPackageJson != null)
            {
                trackedPackageSet.AddPackage(selectedPackageJson);
                UpdateDependencyTree(); // because the version warning can be cleared
                ReloadTrackedPackages();
            }
        }

        private void btnSave_Click(object sender, EventArgs e)
        {
            this.dataFile.SaveToFile();
            MessageBox.Show("Saved.", this.Text, MessageBoxButtons.OK, MessageBoxIcon.Information);
        }

        private void btnLoad_Click(object sender, EventArgs e)
        {
            this.dataFile.LoadFromFile();
            this.ReloadRootList();
            this.ReloadDependencyTree();
            this.ReloadTrackedPackages();
        }

        private string GetImageKeyForPublisher(TrackedPackagePublisher publisher)
        {
            switch (publisher)
            {
                case TrackedPackagePublisher.NotApplicable: return "NotApplicable";
                case TrackedPackagePublisher.External: return "External";
                case TrackedPackagePublisher.Microsoft: return "Microsoft";
                case TrackedPackagePublisher.SharePoint: return "SharePoint";
            }
            return "Unknown";
        }

        private void mnuContext_Opening(object sender, CancelEventArgs e)
        {
            bool enabled = ctlTrackedList.SelectedItems.Count > 0;
            mnuContextUnknown.Enabled = enabled;
            mnuContextNotApplicable.Enabled = enabled;
            mnuContextExternal.Enabled = enabled;
            mnuContextMicrosoft.Enabled = enabled;
            mnuContextSharePoint.Enabled = enabled;
        }

        private void mnuContext_Publisher_Click(object sender, EventArgs e)
        {
            TrackedPackagePublisher publisher = (TrackedPackagePublisher)Enum.Parse(typeof(TrackedPackagePublisher), ((ToolStripMenuItem)sender).Tag as string);
            foreach (ListViewItem selectedItem in ctlTrackedList.SelectedItems)
            {
                TrackedPackage trackedPackage = selectedItem.Tag as TrackedPackage;
                if (trackedPackage != null)
                {
                    trackedPackage.Publisher = publisher;
                }
            }
            this.UpdateDependencyTree();
            this.UpdateTrackedPackages();
        }

        private void mnuContext_Api_Click(object sender, EventArgs e)
        {
            TrackedPackageApiContract apiContract = (TrackedPackageApiContract)Enum.Parse(typeof(TrackedPackageApiContract), ((ToolStripMenuItem)sender).Tag as string);
            foreach (ListViewItem selectedItem in ctlTrackedList.SelectedItems)
            {
                TrackedPackage trackedPackage = selectedItem.Tag as TrackedPackage;
                if (trackedPackage != null)
                {
                    trackedPackage.ApiContract = apiContract;
                }
            }
            this.UpdateTrackedPackages();
        }

        private void mnuContextBrowser_Click(object sender, EventArgs e)
        {
            bool? newState = null;
            foreach (ListViewItem selectedItem in ctlTrackedList.SelectedItems)
            {
                TrackedPackage trackedPackage = selectedItem.Tag as TrackedPackage;
                if (trackedPackage != null)
                {
                    if (newState != null)
                    {
                        trackedPackage.UsedInBrowser = newState.Value;
                    }
                    else
                    {
                        trackedPackage.UsedInBrowser = !trackedPackage.UsedInBrowser;
                        newState = trackedPackage.UsedInBrowser;
                    }
                }
            }
            this.UpdateTrackedPackages();
        }

        private void btnCSV_Click(object sender, EventArgs e)
        {
            string csvFilename = Path.Combine(Path.GetDirectoryName(Application.ExecutablePath), "Report.csv");
            using (StreamWriter writer = new StreamWriter(csvFilename))
            {
                writer.WriteLine(GetCsvLine("Name", "Publisher", "Browser?", "API Contract",
                    "Max Version", "# Versions", "Versions", "Note"));
                foreach (TrackedPackage trackedPackage in this.dataFile.TrackedPackageSet.TrackedPackages)
                {
                    var values = new List<string>();
                    values.Add(trackedPackage.Name);
                    values.Add(trackedPackage.Publisher.ToString());
                    values.Add(trackedPackage.UsedInBrowser ? "Y" : "N");
                    values.Add(trackedPackage.ApiContract.ToString());
                    values.Add(trackedPackage.MaxVersion.ToString());
                    values.Add(trackedPackage.UsedVersions.Count.ToString());
                    values.Add(string.Join(",", trackedPackage.UsedVersions.OrderBy(x => x)));
                    values.Add(string.IsNullOrWhiteSpace(trackedPackage.Note) ? " " : trackedPackage.Note);
                    writer.WriteLine(GetCsvLine(values.ToArray()));
                }
            }
            MessageBox.Show("Wrote CSV", this.Text, MessageBoxButtons.OK, MessageBoxIcon.Information);
        }

        private string GetCsvLine(params string[] values)
        {
            string line = "";
            foreach (string value in values)
            {
                if (line.Length > 0)
                    line += ",";
                line += "\"" + value.Replace("\"", "\"\"") + "\"";
            }
            return line;
        }

        private void ctlDependencyTree_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == 13)
            {
                e.Handled = true;
                this.btnAdd_Click(sender, EventArgs.Empty);
                return;
            }

            if (e.KeyChar == ' ')
            {
                // Toggle star
                var trackedPackageSet = this.dataFile.TrackedPackageSet;

                PackageJson selectedPackageJson = GetDependencyTreeSelection();
                if (selectedPackageJson != null)
                {
                    TrackedPackage trackedPackage;
                    if (trackedPackageSet.TrackedPackagesByName.TryGetValue(selectedPackageJson.Name, out trackedPackage))
                    {
                        trackedPackage.Starred = !trackedPackage.Starred;
                        UpdateTrackedPackages();
                        UpdateDependencyTree();
                    }
                    else
                    {
                        throw new Exception("This package \"" + selectedPackageJson.Name
                            + "\" cannot be starred because it has not been added");
                    }
                }
            }
        }

        private void ctlTrackedList_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == ' ')
            {
                // Toggle star
                bool? newState = null;

                foreach (ListViewItem selectedItem in ctlTrackedList.SelectedItems)
                {
                    TrackedPackage trackedPackage = selectedItem.Tag as TrackedPackage;
                    if (trackedPackage != null)
                    {
                        if (newState == null)
                            newState = !trackedPackage.Starred;
                        trackedPackage.Starred = newState.Value;
                    }
                }

                this.UpdateTrackedPackages();
                this.UpdateDependencyTree();
            }
        }

        private void mnuContextDelete_Click(object sender, EventArgs e)
        {
            foreach (ListViewItem selectedItem in ctlTrackedList.SelectedItems)
            {
                TrackedPackage trackedPackage = selectedItem.Tag as TrackedPackage;
                if (trackedPackage != null)
                {
                    this.dataFile.TrackedPackageSet.RemovePackage(trackedPackage.Name);
                }
            }
            this.ReloadTrackedPackages();
        }

        private void mnuContextNote_Click(object sender, EventArgs e)
        {
            if (ctlTrackedList.SelectedItems.Count != 1)
                return;

            ListViewItem selectedItem = ctlTrackedList.SelectedItems[0];
            TrackedPackage trackedPackage = selectedItem.Tag as TrackedPackage;
            if (trackedPackage == null)
                return;

            using (var form = new EditNoteDialog())
            {
                form.txtPackageName.Text = trackedPackage.Name;
                form.txtNote.Text = trackedPackage.Note;
                if (form.ShowDialog() == DialogResult.OK)
                {
                    trackedPackage.Note = form.txtNote.Text;
                    UpdateTrackedPackages();
                }
            }
        }

        private void ctlTrackedList_DoubleClick(object sender, EventArgs e)
        {
            mnuContextNote_Click(sender, e);
        }
    }

}
