﻿using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace SPClientPackageExplorer
{
    class Dependency
    {
        public readonly string Name;
        public readonly string VersionPattern;

        public PackageJson Resolution = null;

        public Dependency(string name, string versionPattern) {
            this.Name = name;
            this.VersionPattern = versionPattern;
        }

        public override string ToString()
        {
            return (Name ?? "?") + "@" + (VersionPattern ?? "?");
        }
    }

    class PackageJson
    {
        public PackageJsonTree Tree;
        public string FilePath;
        public string Name;
        public string Version;
        public readonly List<Dependency> Dependencies = new List<Dependency>();
        public PackageReport PackageReport = null;

        private PackageJson(PackageJsonTree tree, string filePath, string name, string version)
        {
            this.Tree = tree;
            this.FilePath = filePath;
            this.Name = name;
            this.Version = version;
        }

        internal static PackageJson Load(string packageJsonFilename, PackageJsonTree packageJsonTree)
        {

            string json = File.ReadAllText(packageJsonFilename);
            dynamic jsonObject = JsonConvert.DeserializeObject<JObject>(json);

            var packageJson = new PackageJson(packageJsonTree, packageJsonFilename,
                jsonObject.name.ToString(), jsonObject.version.ToString());

            foreach (var dependencyPair in (JObject)jsonObject.dependencies)
            {
                var dependency = new Dependency(dependencyPair.Key, dependencyPair.Value.ToString());
                packageJson.Dependencies.Add(dependency);
            }

            return packageJson;
        }
    }

    class PackageReport
    {
        public string Name;
        public int NumberOfInstalls;
        public HashSet<string> InstalledVersions = new HashSet<string>();

        public PackageReport(string name)
        {
            this.Name = name;
        }
    }

    class PackageJsonTree
    {
        // Cache for ResolveModuleFilename():  (dependencyName, startingFolder) --> result
        readonly Dictionary<Tuple<string, string>, string> resolvedModuleFilenames = new Dictionary<Tuple<string, string>, string>();
        // (packageName, version)
        readonly HashSet<Tuple<string, string>> alreadyLoadedModules = new HashSet<Tuple<string, string>>();
        readonly Dictionary<string, PackageReport> packageReportsByName = new Dictionary<string, PackageReport>();

        public PackageJson RootPackage { get; private set; }

        public ReadOnlyDictionary<string, PackageReport> PackageReportsByName
        {
            get
            {
                return new ReadOnlyDictionary<string, PackageReport>(this.packageReportsByName);
            }
        }

        public void Clear()
        {
            this.RootPackage = null;
            this.resolvedModuleFilenames.Clear();
            this.packageReportsByName.Clear();
        }

        class QueuedItem
        {
            public Dependency Dependency;
            public PackageJson Parent;
            public QueuedItem(Dependency dependency, PackageJson parent)
            {
                this.Dependency = dependency;
                this.Parent = parent;
            }
        }

        public void LoadTree(string rootPackageJsonFilename)
        {
            this.Clear();

            var queue = new Queue<QueuedItem>();
            this.RootPackage = this.LoadPackage(Path.GetFullPath(rootPackageJsonFilename), null, queue);

            while (queue.Count > 0)
            {
                QueuedItem queuedItem = queue.Dequeue();
                Dependency dependency = queuedItem.Dependency;
                string parentFolder = Path.GetDirectoryName(queuedItem.Parent.FilePath);

                string dependencyPackageJsonFilename = ResolveModuleFilename(dependency.Name, parentFolder);
                if (dependencyPackageJsonFilename == null)
                {
                    Debug.WriteLine("ERROR: Unable to resolve dependency \"" + dependency.Name + "\" for this package:");
                    Debug.WriteLine(parentFolder);
                }
                else
                {
                    dependency.Resolution = this.LoadPackage(dependencyPackageJsonFilename, queuedItem.Parent, queue);
                }
            }
        }

        private PackageJson LoadPackage(string packageJsonFilename, PackageJson parent, Queue<QueuedItem> queue)
        {
            PackageJson packageJson = PackageJson.Load(packageJsonFilename, this);

            // Update the statistics
            PackageReport packageReport;
            if (!packageReportsByName.TryGetValue(packageJson.Name, out packageReport))
            {
                packageReport = new PackageReport(packageJson.Name);
                packageReportsByName.Add(packageJson.Name, packageReport);
            }
            ++packageReport.NumberOfInstalls;
            packageReport.InstalledVersions.Add(packageJson.Version);
            packageJson.PackageReport = packageReport;

            // If this dependency package has been traversed before, we do add it to the tree,
            // but we won't follow any of its children.  This way the tree shows the full set
            // of dependencies for a package version exactly once.
            if (alreadyLoadedModules.Add(Tuple.Create(packageJson.Name, packageJson.Version)))
            {
                // Queue the dependencies
                foreach (Dependency dependency in packageJson.Dependencies)
                {
                    queue.Enqueue(new QueuedItem(dependency, packageJson));
                }
            }

            return packageJson;
        }

        // Searches for the given dependency, starting from the specified folder, and returns the absolute path
        // of the matching package.json file.
        private string ResolveModuleFilename(string dependencyName, string startingFolder)
        {
            string currentFolder = startingFolder;

            for (;;)
            {
                string normalized = Utilities.GetJunctionPointTarget(currentFolder);
                if (normalized != currentFolder)
                {
                    currentFolder = normalized;
                }

                var dictionaryKey = Tuple.Create(dependencyName, currentFolder);
                string resultFilename;
                if (this.resolvedModuleFilenames.TryGetValue(dictionaryKey, out resultFilename))
                    return resultFilename;

                // First look in the local node_modules folder
                string localMatchFolder = Path.Combine(currentFolder, "node_modules", dependencyName);
                if (Directory.Exists(localMatchFolder))
                {
                    string matchedPackageJsonFilename = Path.Combine(localMatchFolder, "package.json");
                    if (File.Exists(matchedPackageJsonFilename))
                    {
                        // SUCCESS.  Update the cache
                        this.resolvedModuleFilenames.Add(dictionaryKey, matchedPackageJsonFilename);
                        return matchedPackageJsonFilename;
                    }
                    else
                    {
                        Debug.WriteLine("WARNING: This folder looked like a match, but is missing package.json:");
                        Debug.WriteLine(localMatchFolder);
                    }
                }

                // Can we go up from "X/node_modules/Y" to "X"?
                string parentFolder = Path.GetDirectoryName(currentFolder);
                if (parentFolder == null || parentFolder == Path.GetPathRoot(startingFolder))
                    return null; // reached root folder

                currentFolder = parentFolder;
            }
        }
    }
}
