﻿using System;
using System.Diagnostics;
using System.Windows.Forms;

namespace SifterKit.Controls {

    /// <summary>
    /// Allow a Windows Forms <see cref="ListView"/> control to automatically sort itself
    /// when you click on a column.  If you click the same column a second time, it sorts
    /// in descending order.
    /// </summary>
    /// <example>
    /// Typical usage:
    /// <code>
    /// public partial class Form1 : Form
    ///     ListViewSorter MyListViewSorter;
    ///
    ///     public Form1()
    ///     {
    ///         // Defines "ListView MyListView;"
    ///         InitializeComponent();
    ///
    ///         MyListViewSorter = new ListViewSorter(MyListView);
    ///     }
    /// }
    /// </code>
    /// </example>
    /// <owner alias="PGonzal" />
    public class ListViewSorter : System.Collections.IComparer
    {
        ListView Control;

        public ListViewSorter(ListView control)
        {
            Control = control;
            Control.ListViewItemSorter = this;
            Control.ColumnClick += new ColumnClickEventHandler(Control_ColumnClick);
        }

        public int ColumnIndex { get; set; }

        void Control_ColumnClick(object sender, ColumnClickEventArgs e)
        {
            Debug.Assert(Control.ListViewItemSorter == this);

            if (ColumnIndex == e.Column)
            {
                if (Control.Sorting == SortOrder.Ascending)
                    Control.Sorting = SortOrder.Descending;
                else
                    Control.Sorting = SortOrder.Ascending;
            }
            else
            {
                Control.Sorting = SortOrder.Ascending;
            }

            ColumnIndex = e.Column;

            Control.Sort();
        }

        protected virtual int CompareItemTags(object tag1, object tag2)
        {
            return 0;
        }

        protected virtual int CompareAscending(ListViewItem item1, ListViewItem item2)
        {
            int result;  // strcmp() style result

            // If the child class can interpret the tags, that takes precedence
            result = CompareItemTags(item1.Tag, item2.Tag);
            if (result != 0) return result;

            // Get the column text
            string str1 = null;
            if (ColumnIndex < item1.SubItems.Count)
                str1 = item1.SubItems[ColumnIndex].Text;
            string str2 = null;
            if (ColumnIndex < item2.SubItems.Count)
                str2 = item2.SubItems[ColumnIndex].Text;

            // can they be interpreted as numbers?
            double double1;
            double double2;

            if (double.TryParse(str1, out double1) && double.TryParse(str2, out double2))
            {
                if (Math.Abs(double1 - double2) < 1E-10)
                    result = 0;
                else
                    result = double1 < double2 ? -1 : 1;
            }
            else
            {
                // Interpret them as strings:
                result = String.Compare(str1, str2);
            }

            return result;
        }

        public int Compare(object obj1, object obj2)
        {
            ListViewItem item1 = obj1 as ListViewItem;
            ListViewItem item2 = obj2 as ListViewItem;

            if (obj1 == null || obj2 == null)
                return 0;

            int result = CompareAscending(item1, item2);

            // Debug.WriteLine(result + ": " + item1.Text + " <--> " + item2.Text);

            if (Control.Sorting == SortOrder.Descending)
                result = -result;

            return result;
        }
    }

}