﻿using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace SPClientPackageExplorer
{
    enum TrackedPackagePublisher
    {
        Unknown,
        SharePoint,
        Microsoft,
        External,
        NotApplicable
    }

    enum TrackedPackageApiContract
    {
        Unknown,
        Internal,
        PublicStatic,
        PublicDynamic
    }

    class TrackedPackage
    {
        public string Name { get; private set; }
        public TrackedPackagePublisher Publisher { get; set; }
        public readonly HashSet<string> UsedVersions = new HashSet<string>();
        public bool UsedInBrowser = false;
        public TrackedPackageApiContract ApiContract = TrackedPackageApiContract.Unknown;
        public bool Starred = false;
        public string Note = "";

        public TrackedPackage(string name)
        {
            this.Name = name;
            this.Publisher = TrackedPackagePublisher.Unknown;
        }

        public string MaxVersion
        {
            get
            {
                List<Version> versions = new List<Version>(10);
                foreach (string versionText in this.UsedVersions)
                {
                    Version version;
                    if (!Version.TryParse(versionText, out version))
                    {
                        // Some packages have syntactically invalid SemVer strings;
                        // In this case, just sort alphabetically and hope for the best
                        return this.UsedVersions.Max();
                    }
                    versions.Add(version);
                }
                return versions.Max().ToString();
            }
        }

        internal void WriteJson(JsonWriter jsonWriter)
        {
            jsonWriter.WriteStartObject();
            jsonWriter.WritePropertyName("Name");
            jsonWriter.WriteValue(this.Name);

            jsonWriter.WritePropertyName("Publisher");
            jsonWriter.WriteValue((int)this.Publisher);

            jsonWriter.WritePropertyName("UsedInBrowser");
            jsonWriter.WriteValue(this.UsedInBrowser);

            jsonWriter.WritePropertyName("ApiContract");
            jsonWriter.WriteValue(this.ApiContract);

            jsonWriter.WritePropertyName("Starred");
            jsonWriter.WriteValue(this.Starred);

            if (!string.IsNullOrWhiteSpace(this.Note))
            {
                jsonWriter.WritePropertyName("Note");
                jsonWriter.WriteValue(this.Note);
            }

            jsonWriter.WritePropertyName("UsedVersions");
            jsonWriter.WriteStartArray();
            foreach (string usedVersion in this.UsedVersions)
            {
                jsonWriter.WriteValue(usedVersion);
            }
            jsonWriter.WriteEndArray();

            jsonWriter.WriteEndObject();
        }

        internal void ReadJson(dynamic trackedPackageJson, DataFile dataFile)
        {
            this.Publisher = (TrackedPackagePublisher)trackedPackageJson.Publisher;
            this.UsedInBrowser = trackedPackageJson.UsedInBrowser;
            this.ApiContract = (TrackedPackageApiContract)trackedPackageJson.ApiContract;

            if (dataFile.Version > DataFileVersion.V1000)
            {
                this.Starred = trackedPackageJson.Starred;
            }

            this.Note = "";
            if (trackedPackageJson["Note"] != null)
            {
                this.Note = trackedPackageJson.Note;
            }

            foreach (string usedVersion in trackedPackageJson.UsedVersions)
            {
                this.UsedVersions.Add(usedVersion);
            }
        }
    }

    class TrackedPackageSet
    {
        private readonly Dictionary<string, TrackedPackage> trackedPackagesByName = new Dictionary<string, TrackedPackage>();

        public ReadOnlyDictionary<string, TrackedPackage> TrackedPackagesByName
        {
            get
            {
                return new ReadOnlyDictionary<string, TrackedPackage>(this.trackedPackagesByName);
            }
        }

        public ReadOnlyCollection<TrackedPackage> TrackedPackages
        {
            get
            {
                return new ReadOnlyCollection<TrackedPackage>(this.trackedPackagesByName.Values.OrderBy(x => x.Name).ToArray());
            }
        }

        internal void ReadJson(dynamic trackedPackageSetJson, DataFile dataFile)
        {
            this.Clear();
            foreach (dynamic trackedPackageJson in trackedPackageSetJson.Packages)
            {
                var trackedPackage = new TrackedPackage(trackedPackageJson.Name.ToString());
                trackedPackage.ReadJson(trackedPackageJson, dataFile);
                this.trackedPackagesByName.Add(trackedPackage.Name, trackedPackage);
            }
        }

        internal void WriteJson(JsonWriter jsonWriter)
        {
            jsonWriter.WriteStartObject();
            jsonWriter.WritePropertyName("Packages");
            jsonWriter.WriteStartArray();

            foreach (var trackedPackage in this.TrackedPackages)
            {
                trackedPackage.WriteJson(jsonWriter);
            }

            jsonWriter.WriteEndArray();

            jsonWriter.WriteEndObject();
        }

        public void AddPackage(PackageJson packageJson)
        {
            TrackedPackage trackedPackage;
            if (!trackedPackagesByName.TryGetValue(packageJson.Name, out trackedPackage))
            {
                trackedPackage = new TrackedPackage(packageJson.Name);
                trackedPackagesByName.Add(packageJson.Name, trackedPackage);
            }

            trackedPackage.UsedVersions.Add(packageJson.Version);
        }

        public void Clear()
        {
            this.trackedPackagesByName.Clear();
        }

        public bool RemovePackage(string nameToRemove)
        {
            return this.trackedPackagesByName.Remove(nameToRemove);
        }
    }
}
