﻿using Microsoft.Win32.SafeHandles;
using System;
using System.Runtime.InteropServices;
using System.Text;

namespace SPClientPackageExplorer
{
    public static class Utilities
    {
        /// <summary>
        /// If folderThatMightBeAJunctionPoint is an NTFS directory junction, then this returns
        /// the absolute path of the target physical folder.  Otherwise, it returns the
        /// input parameter.
        /// </summary>
        public static string GetJunctionPointTarget(string folderThatMightBeAJunctionPoint)
        {
            using (SafeFileHandle handle = GetReparsePointHandle(folderThatMightBeAJunctionPoint,
                EFileAccess.GenericRead))
            {
                string target = GetJunctionPointTarget(handle);

                if (target == null)
                    return folderThatMightBeAJunctionPoint;

                return target;
            }
        }

        private static string GetJunctionPointTarget(SafeFileHandle handle)
        {
            int bufferSize = Marshal.SizeOf(typeof(REPARSE_DATA_BUFFER));
            IntPtr buffer = Marshal.AllocHGlobal(bufferSize);

            try
            {
                int bytesReturned;
                bool result = DeviceIoControl(handle.DangerousGetHandle(), FSCTL_GET_REPARSE_POINT,
                    IntPtr.Zero, 0, buffer, bufferSize, out bytesReturned, IntPtr.Zero);

                if (!result)
                {
                    int error = Marshal.GetLastWin32Error();
                    if (error == ERROR_NOT_A_REPARSE_POINT)
                        return null;

                    ThrowLastError("DeviceIoControl failed for junction point");
                }

                REPARSE_DATA_BUFFER reparseDataBuffer = (REPARSE_DATA_BUFFER)
                    Marshal.PtrToStructure(buffer, typeof(REPARSE_DATA_BUFFER));

                if (reparseDataBuffer.ReparseTag != IO_REPARSE_TAG_MOUNT_POINT)
                    return null;

                string targetFolder = Encoding.Unicode.GetString(reparseDataBuffer.PathBuffer,
                    reparseDataBuffer.SubstituteNameOffset, reparseDataBuffer.SubstituteNameLength);

                /// Special prefix that indicates that the path is treated as a non-interpreted
                /// path in the virtual file system.
                const string prefix = @"\??\";

                if (targetFolder.StartsWith(prefix))
                    targetFolder = targetFolder.Substring(prefix.Length);

                return targetFolder;
            }
            finally
            {
                Marshal.FreeHGlobal(buffer);
            }
        }

        private static SafeFileHandle GetReparsePointHandle(string reparsePoint, EFileAccess accessMode)
        {
            IntPtr fileHandle = CreateFile(reparsePoint, accessMode, EFileShare.Read | EFileShare.Write | EFileShare.Delete,
                IntPtr.Zero, ECreationDisposition.OpenExisting,
                EFileAttributes.BackupSemantics | EFileAttributes.OpenReparsePoint, IntPtr.Zero);

            SafeFileHandle reparsePointHandle = new SafeFileHandle(fileHandle, true);

            if (Marshal.GetLastWin32Error() != 0)
                ThrowLastError("CreateFile failed for junction point");

            return reparsePointHandle;
        }

        private static void ThrowLastError(string message)
        {
            throw new Exception(message, Marshal.GetExceptionForHR(Marshal.GetHRForLastWin32Error()));
        }

        #region Win32 PInvoke

        private const int ERROR_NOT_A_REPARSE_POINT = 4390;
        private const int ERROR_REPARSE_ATTRIBUTE_CONFLICT = 4391;
        private const int ERROR_INVALID_REPARSE_DATA = 4392;
        private const int ERROR_REPARSE_TAG_INVALID = 4393;
        private const int ERROR_REPARSE_TAG_MISMATCH = 4394;
        private const int FSCTL_SET_REPARSE_POINT = 0x000900A4;
        private const int FSCTL_GET_REPARSE_POINT = 0x000900A8;
        private const int FSCTL_DELETE_REPARSE_POINT = 0x000900AC;
        private const uint IO_REPARSE_TAG_MOUNT_POINT = 0xA0000003;

        private enum ECreationDisposition : uint
        {
            New = 1,
            CreateAlways = 2,
            OpenExisting = 3,
            OpenAlways = 4,
            TruncateExisting = 5,
        }

        [Flags]
        private enum EFileAccess : uint
        {
            GenericRead = 0x80000000,
            GenericWrite = 0x40000000,
            GenericExecute = 0x20000000,
            GenericAll = 0x10000000,
        }

        [Flags]
        private enum EFileShare : uint
        {
            None = 0x00000000,
            Read = 0x00000001,
            Write = 0x00000002,
            Delete = 0x00000004,
        }

        [Flags]
        private enum EFileAttributes : uint
        {
            Readonly = 0x00000001,
            Hidden = 0x00000002,
            System = 0x00000004,
            Directory = 0x00000010,
            Archive = 0x00000020,
            Device = 0x00000040,
            Normal = 0x00000080,
            Temporary = 0x00000100,
            SparseFile = 0x00000200,
            ReparsePoint = 0x00000400,
            Compressed = 0x00000800,
            Offline = 0x00001000,
            NotContentIndexed = 0x00002000,
            Encrypted = 0x00004000,
            Write_Through = 0x80000000,
            Overlapped = 0x40000000,
            NoBuffering = 0x20000000,
            RandomAccess = 0x10000000,
            SequentialScan = 0x08000000,
            DeleteOnClose = 0x04000000,
            BackupSemantics = 0x02000000,
            PosixSemantics = 0x01000000,
            OpenReparsePoint = 0x00200000,
            OpenNoRecall = 0x00100000,
            FirstPipeInstance = 0x00080000
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct REPARSE_DATA_BUFFER
        {
            public uint ReparseTag;
            public ushort ReparseDataLength;
            public ushort Reserved;
            public ushort SubstituteNameOffset;
            public ushort SubstituteNameLength;
            public ushort PrintNameOffset;
            public ushort PrintNameLength;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 0x3FF0)]
            public byte[] PathBuffer;
        }

        [DllImport("kernel32.dll", SetLastError = true)]
        private static extern IntPtr CreateFile(
            string lpFileName,
            EFileAccess dwDesiredAccess,
            EFileShare dwShareMode,
            IntPtr lpSecurityAttributes,
            ECreationDisposition dwCreationDisposition,
            EFileAttributes dwFlagsAndAttributes,
            IntPtr hTemplateFile);

        [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern bool DeviceIoControl(IntPtr hDevice, uint dwIoControlCode,
            IntPtr InBuffer, int nInBufferSize,
            IntPtr OutBuffer, int nOutBufferSize,
            out int pBytesReturned, IntPtr lpOverlapped);

        #endregion
    }
}
